//<<<<<<<<<<<<<<<<<<<<<< makeExtruded QuadStripMesh >>>>>>>>>>>>>>>
void Mesh:: makeExtrudedQuadStrip(Point2 P[], int numPts, Vector3 w)
{// quad strip in list P[0],P[1], P[2], .., of points in xy-plane
  // extrusion is along vector w
  // Original quadstrip should be ordered CW: P0,P1,P3,P2 
  // as seen from pos. z-axis looking at xy-plane.
  int numSegments = numPts/2 - 1;
  numVerts = numPts * 2; // for all waists = quadrilaterals
  numFaces = numSegments * 4; // each segment has 4 faces
  numNorms = numFaces; // for visibly flat faces

  pt   = new Point3[numVerts];  assert(pt); // make space for the lists
  face = new Face[numFaces];    assert(face);
  norm = new Vector3[numNorms]; assert(norm);

  for(int s = 0; s <= numSegments; s++) // for each waist
  {
      int twoI = 2 * s;
      pt[4*s    ].set(P[twoI    ].x,P[twoI    ].y,0);
      pt[4*s + 1].set(P[twoI + 1].x,P[twoI + 1].y,0);
      pt[4*s + 2].set(P[twoI + 1].x + w.x,P[twoI + 1].y + w.y, w.z);
      pt[4*s + 3].set(P[twoI    ].x + w.x,P[twoI    ].y + w.y,w.z);
  }
  for(int i = 0; i < numSegments; i++)
    for(int j = 0; j < 4; j++)
    {
      int which = 4 * i + j; // which face
      int i0 = which, i1 = i0 + 4,  i3 = 4 * i + (j + 3)%4,  i2 = i3 + 4;
      norm[which] = newell4Pts(pt[i0],pt[i1],pt[i2], pt[i3]);
      norm[which].normalize();
      face[which].vert = new VertexID[4]; assert(face[which].vert != NULL);
      face[which].nVerts = 4;
      face[which].vert[0].vertIndex = i0;// same as norm index
      face[which].vert[0].normIndex = which;
      face[which].vert[1].vertIndex = i1; 
      face[which].vert[1].normIndex = which;
      face[which].vert[2].vertIndex = i2;
      face[which].vert[2].normIndex = which;
      face[which].vert[3].vertIndex = i3;
      face[which].vert[3].normIndex = which;
    }
} // end, makeExtrudedQuadStrip

//<<<<<<<<<<<<<<<<<<<<<< checkOrthogVects >>>>>>>>>>>>>>>>
void checkOrthogVects(Vector3 a, Vector3 b, Vector3 c)
{ // check that system is orthonormal
  double aDotb = fabs(dot3(a, b));
  double aDotc = fabs(dot3(a, c));
  double bDotc = fabs(dot3(b, c));
  if( aDotb > 0.000001 || aDotc > 0.000001 || bDotc > 0.000001)
    cout << " Bad!! vectors NOT orthogonal!\n";
  double A = fabs(dot3(a,a) - 1.0);
  double B = fabs(dot3(b,b) - 1.0);
  double C = fabs(dot3(c,c) - 1.0);
  if(A > 0.0000001)
    cout << "Bad!! first is not normalized!\n";
  if( B > 0.0000001)
    cout << "Bad!! second is not normalized!\n";
  if( C > 0.0000001)
    cout << "Bad!! third is not normalized!\n";
}


//######################### MAKE SMOOTH TUBE ######################
void Mesh:: makeSmoothTube(void)
{  // make tube: wrap polygon about spine
  int numSpinePts = 200;
  int numSlices = 12; // # vrtices in polygon
  float radius = 1.0;
  float fractForInnerOuterRadii = 0.4f; // use inner-outer radii n-gon for polygon
  //---------------------------- make space for lists -----------------
  numVerts = numSlices *  numSpinePts ; // total # vertices in mesh
  numFaces = numSlices * (numSpinePts -1);       // total # faces in mesh
  numNorms = numVerts;
  pt   = new Point3[numVerts];  assert(pt   != NULL); // make space for the lists
  face = new Face[numFaces];    assert(face != NULL);
  norm = new Vector3[numNorms]; assert(norm != NULL);
  //make polygon that wraps spine:
  Point2 *p = new Point2[numSlices];  assert(p);// place for polygon that wraps
  float ang = 0, delAng = TWOPI/numSlices;
  for(int j = 0; j < numSlices; j++, ang += delAng) // numSlices should be even
  { 
    float rad = radius;
    if(j%2) rad = fractForInnerOuterRadii * radius; //alternating radius
    p[j].x = rad * cos(ang);
    p[j].y = rad * sin(ang);
  }
  double t = 0, tMax, delT, eps = 0.0001;
  tMax = TWOPI;
  delT = tMax/(numSpinePts-1);
  int last = -1;
  // loop over values of t for different spine segments
  for(int k = 0; k < numSpinePts; k++, t += delT) // for each spine point..
  {
    Point3 spine(spineX(t), spineY(t), spineZ(t));
  //make Frenet frame at spine points:
    Vector3 tangent, spinePrimePrime,B,N;
    //do tangent with numerical derivative
    double xderiv = (spineX(t + eps) - spineX(t - eps))/(2*eps); 
    double yderiv = (spineY(t + eps) - spineY(t - eps))/(2*eps); 
    double zderiv = (spineZ(t + eps) - spineZ(t - eps))/(2*eps); 
    tangent.set(xderiv,yderiv,zderiv);tangent.normalize();
    // second differences for the acceleration
    double xDoublePrime = (spineX(t - eps) - 2 * spineX(t) + spineX(t + eps))/(eps*eps);
    double yDoublePrime = (spineY(t - eps) - 2 * spineY(t) + spineY(t + eps))/(eps*eps);
    double zDoublePrime = (spineZ(t - eps) - 2 * spineZ(t) + spineZ(t + eps))/(eps*eps);

    spinePrimePrime.set(xDoublePrime, yDoublePrime, zDoublePrime);
    B = tangent.cross(spinePrimePrime); B.normalize();
    N = B.cross(tangent);N.normalize();
    checkOrthogVects(tangent,B,N); // are tangent, B, and n orthonormal?
    //compute vertex list
    for(int j = 0; j < numSlices; j++)
    { 
      Vector3 V(p[j].x * B.x + p[j].y * N.x, // construct new vector
                p[j].x * B.y + p[j].y * N.y,
                p[j].x * B.z + p[j].y * N.z);
      pt[++last].set(spine.x + V.x, spine.y + V.y, spine.z + V.z);
      norm[last].set(V.x, V.y, V.z);
      norm[last].normalize();
    } // end: for each polygon vertex at this spine point
  } // end: for each spine segment
  // ----------------- make face lists --------------------
  int ind = -1;
  for(int spinePt = 0; spinePt < numSpinePts - 1; spinePt++)
  {
    for(int f = 0; f < numSlices; f++) // for each face in this segment
    {
      face[++ind].nVerts = 4;
      face[ind].vert = new VertexID[4]; assert(face[ind].vert != NULL);
      int i0 = spinePt * numSlices + f; // first index of the vert in this face
      int i1 = (f == numSlices - 1) ? spinePt * numSlices :(i0 + 1); 
      int i2 = i1 + numSlices;
      int i3 = i0 + numSlices;  // last index of the vert in this face
      face[ind].vert[0].vertIndex = face[ind].vert[0].normIndex = i0;
      face[ind].vert[1].vertIndex = face[ind].vert[1].normIndex = i1;
      face[ind].vert[2].vertIndex = face[ind].vert[2].normIndex = i2;
      face[ind].vert[3].vertIndex = face[ind].vert[3].normIndex = i3;       
    } // end: for each face in this segment
  } // end: for each spine segment - building face and normal lists
} // end: makeTube()

//MeshDec98.cpp     - Sandy, 12/12/98 - for ECE660
// This is NOT code that you can just drop into your programs... it is provided to
// show one way of working with texture, which you might find useful after, of course,
// thoroughly scrutinizing how it does it.
// This shows one way to enhance the Mesh class so textures can be drawn on the faces of a mesh.
//  An excerpt from mesh.h shows:
// Extra fields added for handling textures:
// numTxtrs;  - the number of different textures to be used by this mesh object
// numTxtrVerts; - the total number of different texture coordinate pairs used
// txtrName;  - array of texture ID's to be given by openGL through glGenTextures()
// RGBpixmap *pix;  - array of RGBpixmaps, one for each texture
// txtrVert; - array of texture coordinate pairs used in this mesh

// The face description has been enhanced as well by adding the fields:
// whichTxtr; - 0 if this face doesn't use a texture; else the texture ID given by OpenGL
// txtrIndex; - index telling which texture vertex is associated with this vertex in a faces

//Also enhanced are the methods:
// drawMesh() - draws each face as before, but renders texture on those for which whichTxtr > 0
// readMesh(char * fname) - reads a 3vt file to load data into a mesh object. 
// See example file, barn.3vt, below, that reads 3 texture files.
// makeExtrudedQuadStrip(Point2 P[], int numPts, Vector3 w,char *fname); - given an array of 2D points
// representing a quad-strip in xy-plane, it builds the extruded version as a mesh, and paints one
// texture on the top face. Takes additional parameter, fname, - name of BMP file for texture pixmap

// One additional method of the RGBpixmap class was discussed in class:it's setTexture(). You'll find it at the end of this file.

//@@@@@@@@@@@@@@ Mesh.h excerpt @@@@@@@@@@@@@@
ifndef _MeshHeader
#define _MeshHeader
#include <windows.h>
#include <assert.h>
#include <iostream.h>
#include <string.h>
#include <fstream.h>
#include <iomanip.h>
#include <math.h>
#include <gl/Gl.h>
#include <gl/Glu.h>
#include <gl/glut.h>
#define PI 3.14159265
#define TWOPI 6.2831853
//#include "hlGeom.h"   
#include "RGBpixmap.h"  //as already availabel on FTP site
//############ class VertexID ################
class VertexID{public: int vertIndex, normIndex, txtrIndex;};
//################ class FACE ############
class Face{
public:
        int nVerts;
        int whichTxtr;
        VertexID * vert; // array of vertex and normal indices
        Face(){ nVerts = 0; vert = NULL;}
};
//############## class Mesh ####################
class Mesh{
private:
        int numVerts;
        int numNorms;
        int numFaces;
        int numTxtrs;
        int numTxtrVerts;
        unsigned int *txtrName; // array of texture ID's
        RGBpixmap *pix; // array of RGBpixmaps for textures
        Point3 *pt; // array of points
        Vector3 *norm; // array of normals
        Point2 *txtrVert; // array of texture vertices
        Face *face; // array of faces
        int lastVertUsed;
        int lastNormUsed;
        int lastFaceUsed;
        int lastTextrVertUsed;
        void addPrism(Point2 poly[], int numVertsInPoly, float height);
public:
        void readMesh(char * fname);
        void writeMesh(char* fname);
        void printMesh();
        void drawMesh();
        void drawEdges();
        void freeMesh();
        int isEmpty() { return (numVerts == 0) || (numFaces == 0) || (numNorms == 0);}
        void makeEmpty() { numVerts = numFaces = numNorms = numTxtrVerts = 0;}
        Mesh(){
                numVerts = numFaces = numNorms = numTxtrVerts = 0; 
                pt = NULL; norm  =  NULL; face = NULL;
                lastVertUsed = lastNormUsed = lastFaceUsed = lastTextrVertUsed = -1;
        }
        void makeSmoothTube(void);
        Vector3 newell4(int indx[]);
       void makeExtrudedQuadStrip(Point2 P[], int numPts, Vector3 w,char *fname);
};
#endif

//@@@@@@@@@@@@ Mesh.cpp Excerpt @@@@@@@@@@@@@
// some .cpp files for the TexturedMesh class
// fsh - Nov. 17, 1998
#include "hlGeom.h" // needed in Mesh.h
#include "Mesh.h"
#include "RGBpixmap.h"
//<<<<<<<<<<<<<<<<<<<<<< drawMesh >>>>>>>>>>>>>>>>>>>>
void Mesh :: drawMesh()
{ // draw each face of this mesh using OpenGL: draw each polygon.
        if(isEmpty()) return; // mesh is empty
        for(int f = 0; f < numFaces; f++)
        {
                int n = face[f].nVerts;
                if(face[f].whichTxtr > 0) //enable texturing and bind proper texture to this polygon
                {
                         glEnable(GL_TEXTURE_2D);
                         glBindTexture(GL_TEXTURE_2D,face[f].whichTxtr);

                        glBegin(GL_POLYGON);
                        for(int v = 0; v < n; v++) // for each vertex
                        {
                                int in = face[f].vert[v].normIndex; // get normal index
                                   assert(in >= 0 && in < numNorms);
                                glNormal3f(norm[in].x, norm[in].y, norm[in].z); // set normal to vertex
    int it = face[f].vert[v].txtrIndex; // get texture coord index for this vertex
                                glTexCoord2f(txtrVert[it].x , txtrVert[it].y); // set texture coords
    int iv = face[f].vert[v].vertIndex; // get vertex index
                                assert(iv >= 0 && iv < numVerts);
                                glVertex3f(pt[iv].x, pt[iv].y, pt[iv].z);
                        }
                        glEnd();
                }
                else // texture not used for this face - normal rendering as before
                {
                        glDisable(GL_TEXTURE_2D);
                        glBegin(GL_POLYGON);
                        for(int v = 0; v < n; v++)
                        {
                                int in = face[f].vert[v].normIndex;
                                   assert(in >= 0 && in < numNorms);
                                glNormal3f(norm[in].x, norm[in].y, norm[in].z);
                                int iv = face[f].vert[v].vertIndex; 
                                   assert(iv >= 0 && iv < numVerts);
                                glVertex3f(pt[iv].x, pt[iv].y, pt[iv].z);
                        }
                        glEnd();
                }
        }
        glFlush();
}               
//<<<<<<<<<<<<<<<<<<<<<<<<<<<< readMesh >>>>>>>>>>>>>>>>>>>>>>>>
void Mesh:: readMesh(char * fname) //read a .3vt file that describes a mesh with texture
{
        fstream inStream;
        inStream.open(fname, ios ::in); // open the input stream
        if(inStream.fail()) {cout << "can't open file: " << fname << endl; return;}
        if(inStream.eof()){makeEmpty();return;}
        inStream >> numVerts >> numNorms >> numFaces >> 
numTxtrs >> numTxtrVerts; 
        pt = new Point3[numVerts];  assert(pt != NULL);   // make arrays for vertices and vertex normals
        norm = new Vector3[numNorms];  assert(norm != NULL);
        face = new Face[numFaces];    assert(face != NULL);
       txtrName = new unsigned int[numTxtrs]; assert(txtrName != NULL);// for texture ID's
       txtrVert = new Point2[numTxtrVerts]; assert(txtrVert != NULL); 
        pix = new RGBpixmap[numTxtrs]; assert( pix != NULL);

       glGenTextures(numTxtrs,txtrName); // let OpenGL set up ID's
      char tName[20]; // to hold .BMP file names of pixmaps
      for(int j = 0; j < numTxtrs; j++) // read texture files and make textures
        {
                inStream >> tName; // get next texture file name
                pix[j].readBMPFile(tName); 
                pix[j].setTexture(txtrName[j]);
        }
       for(int i = 0; i < numVerts; i++)     // read in the vertices
                inStream >> pt[i].x >> pt[i].y >> pt[i].z;
        for(int ii = 0; ii < numNorms; ii++)    // read in the normals
                inStream >> norm[ii].x >> norm[ii].y >> norm[ii].z;
        for(int mm = 0; mm < numTxtrVerts; mm++)
                inStream >> txtrVert[mm].x >> txtrVert[mm].y; // array of texture vertices
        for(int f = 0; f < numFaces; f++)   // read in face data
        {
                int thisTxtrNum;
                int n;
                inStream >> n >> thisTxtrNum;
                face[f].nVerts = n;

                if(thisTxtrNum == -1)
                        face[f].whichTxtr = 0;
                else
                        face[f].whichTxtr = txtrName[thisTxtrNum];
                face[f].vert = new VertexID[n]; assert(face[f].vert != NULL);
                for(int k = 0; k < n; k++)    // read vertex indices for this face
                        inStream >> face[f].vert[k].vertIndex;
                for(int kk = 0; kk < n; kk++)    // read normal indices for this face
                        inStream >> face[f].vert[kk].normIndex;
                if(face[f].whichTxtr > 0)
                        for(int mm = 0; mm < n; mm++)
                                inStream >> face[f].vert[mm].txtrIndex;

        }       inStream.close();
} // end of readMesh
//<<<<<<<<<<<<<<<<<<<<<< makeExtruded QuadStripMesh >>>>>>>>>>>>>>>
void Mesh:: makeExtrudedQuadStrip(Point2 P[], int numPts, Vector3 w, char* fname)
{// quad strip in list P[0],P[1], P[2], .., of points in xy-plane
        // extrusion is along vector w
        // Original quadstrip should be ordered CW: P0,P1,P3,P2 
        // as seen from pos. z-axis looking at xy-plane
        int numSegments = numPts/2 - 1;

        unsigned int ID = 201; // my ad hoc number for this texture

        numVerts = numPts * 2; // for all waists = quadrilaterals
        numFaces = numSegments * 4; // each segment has 4 faces
        numNorms = numFaces; // for visibly flat faces
        RGBpixmap pix;
        pix.readBMPFile(fname); // read .bmp file to make pixmap
        pix.setTexture(ID);

        //cout << " numVerts, Norms, Faces = " << numVerts << "," 
<< numNorms << "," << numFaces << endl;
        pt   = new Point3[numVerts];  assert(pt); // make space for the lists
        face = new Face[numFaces];    assert(face);
        norm = new Vector3[numNorms]; assert(norm);
        numTxtrs = 1;  // texture on top faces only
        numTxtrVerts = numPts; // one texture coord pair for quadstrip vertex
        txtrVert = new Point2[numTxtrVerts]; assert(txtrVert != NULL); 

        float left = P[0].x, bott = P[0].y, right = P[0].x, top = P[0].y; //search for bounding box
        for(int j = 1; j < numPts; j++)
        {
                if(P[j].x < left) left = P[j].x;
                if(P[j].x > right) right = P[j].x;
                if(P[j].y < bott) bott = P[j].y;
                if(P[j].y > top) top = P[j].y;
        }

        float wid = right - left, ht = top - bott; // dimensions of bounding box

       float numTilesHoriz = 13, numTilesVert = 4; // change this for your Mesh !!

        for(int s = 0; s <= numSegments; s++) // for each waist
        {
                        int twoI = 2 * s;
                        Point2 p1,p2;
                        p1.set(P[twoI].x,P[twoI].y);
                        p2.set(P[twoI + 1].x,P[twoI + 1].y);
                        pt[4*s    ].set(p1.x,p1.y,0);
                        pt[4*s + 1].set(p2.x,p2.y,0);
                        pt[4*s + 2].set(p2.x + w.x,p2.y + w.y, w.z);
                        pt[4*s + 3].set(p1.x + w.x,p1.y + w.y,w.z);
                        Point2 q1,q2;
                        q1.set((p1.x-left)*numTilesHoriz/wid,(p1.y-bott)*numTilesVert/ht);
                       q2.set((p2.x-left)*numTilesHoriz/wid,(p2.y-bott)*numTilesVert/ht);
                        txtrVert[twoI].set(q1.x,q1.y);
                        txtrVert[twoI +1].set(q2.x,q2.y);
        }
        for(int i = 0; i < numSegments; i++)
                for(int j = 0; j < 4; j++)
                {
                        int which = 4 * i + j; // which face
                        int i0 = which, i1 = i0 + 4,  i3 = 4 * i + (j + 3)%4,  i2 = i3 + 4;
                        norm[which] = newell4Pts(pt[i0],pt[i1],pt[i2], pt[i3]);
                        norm[which].normalize();
                        face[which].vert = new VertexID[4]; assert(face[which].vert != NULL);
                        face[which].nVerts = 4;
                        face[which].vert[0].vertIndex = i0;// same as norm index
                        face[which].vert[0].normIndex = which;
                        face[which].vert[1].vertIndex = i1; 
                        face[which].vert[1].normIndex = which;
                        face[which].vert[2].vertIndex = i2;
                        face[which].vert[2].normIndex = which;
                        face[which].vert[3].vertIndex = i3;
                        face[which].vert[3].normIndex = which;
                        if(j == 3)
                        {
                                face[which].whichTxtr = ID;
                                face[which].vert[0].txtrIndex = 2 * i;
                                face[which].vert[1].txtrIndex = 2 * i + 2;
                                face[which].vert[2].txtrIndex = 2 * i + 3;
                                face[which].vert[3].txtrIndex = 2 * i + 1;
                        }
                        else face[which].whichTxtr = 0;
    }
} // end, makeExtrudedQuadStrip

//@@@@@@@@@@@@@ barn.3vt @@@@@@@@@@@@@@@
10 7 7 3 11
fleur.bmp
lemon.bmp
mandrill.bmp
0 0 0
0 1 0
0 1 1
0 0.5 1.5
0 0 1
1 0 0
1 1 0
1 1 1
1 0.5 1.5
1 0 1
0 -1 0
0 -0.707107 0.707107
0 0.707107 0.707107
0 1 0
0 0 -1
1 0 0
-1 0 0
0 0
1 0
1 1
0 1
1 0.707
0 0.707
0.167 0
0.833 0
0.833 0.666
0.5 1
0.167 0.666
4 -1
0 5 9 4 
0 0 0 0 
4 -1
3 4 9 8 
1 1 1 1 
4 0
2 3 8 7 
2 2 2 2 
1 4 5 0
4 1
1 2 7 6 
3 3 3 3
0 1 2 3
4 -1
0 1 6 5 
4 4 4 4 
5 2
5 6 7 8 9 
5 5 5 5 5 
6 7 8 9 10
5 -1
0 4 3 2 1 
6 6 6 6 6 

//################### setTexture #####################
//<<<<<<<<<<<<<<<<<<<<< setTexture >>>>>>>>>>>>>>>>>>>>>>>
void RGBpixmap :: setTexture(unsigned int textureName)
{
  glBindTexture(GL_TEXTURE_2D,textureName);
  glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER,GL_NEAREST);
  glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER,GL_NEAREST);
  glTexImage2D(GL_TEXTURE_2D, 0, GL_RGB,nCols,nRows,0, GL_RGB,  GL_UNSIGNED_BYTE, pixel);
}

