// ------------------------------------------------
// Filename      : generate-moving-sphere-frame.cpp
// Author        : Sven Maerivoet
// Last modified : 28/08/2001
// Target        : C++ (OpenGL)
//
// All rights reserved.
// ------------------------------------------------

#include <iostream>
#include <math.h>
#include <string>
#include "opengl/math.h"

// string-streams are not always compatible
// => Microsoft Visual C++ 6.0 and UNIX/Linux are treated separately
#ifdef _WIN32
  #include <sstream>
#else
  #include <strstream>
  #define  stringstream strstream
#endif

using namespace std;

double getValue(string str)
{
  stringstream tmp;
  double number;

  tmp << str;
  if (!(tmp >> number)) {
    cerr << "getValue() -> error getting value [" << str << "]." << endl;
    exit(1);
  }
  else {
    char t;
    if (tmp >> t) {
      cerr << "getValue() -> bum chars in number [" << str << "]." << endl;
      exit(1);
    } 
  }

  return number;
}

void generateMovingSphereFrame(int part,
                               int framesPerPart,
                               int currentFrame)
{
  cout << "def Wall {" << endl;
  cout << "  push" << endl;
  cout << "    defaultMaterials" << endl;
  cout << "    use          ColorWhite" << endl;
  cout << "    imageTexture data/textures/other/zebra.bmp 0 1 1" << endl;
  cout << "    translate    0.0 0.0 -1.0" << endl;
  cout << "    scale        1.0 1.0  0.1" << endl;
  cout << "    cube" << endl;
  cout << "  pop" << endl;
  cout << "}" << endl;

  cout << endl;

  cout << "def GlassSphere {" << endl;
  cout << "  push" << endl;
  cout << "    defaultMaterials" << endl;
  cout << "    ambient                    0.1 0.1 0.1" << endl;
  cout << "    diffuse                    0.0 0.0 0.0" << endl;
  cout << "    specular                   0.7 0.7 0.7" << endl;
  cout << "    specularExponent         100" << endl;
  cout << "    transparency               0.8" << endl;
  cout << "    retainSpecularReflection   1" << endl;

  double minX, maxX, x;
  double minZ, maxZ, z;
  double minTS, maxTS, ts;
  double fraction = ((double) (currentFrame - 1)) / ((double) (framesPerPart - 1));

  switch (part) {
    case 1 :
      // move from left to right
      minX = -1.2;
      maxX = 1.2;
      x = Math::lerp(minX,maxX,fraction);
      cout << "    translate                  " << x << " 0.0 0.0" << endl;
      break;

    case 2 :
      // move from right to center
      minX = 1.2;
      maxX = 0.0;
      x = Math::lerp(minX,maxX,fraction);
      cout << "    translate                  " << x << " 0.0 0.0" << endl;
      break;

    case 3 :
      // move from center to back
      minZ = 0.0;
      maxZ = -1.4;
      z = Math::lerp(minZ,maxZ,fraction);
      cout << "    translate                  0.0 0.0 " << z << endl;
      break;

    case 4 :
      // move from back to front
      minZ = -1.4;
      maxZ = 1.4;
      z = Math::lerp(minZ,maxZ,fraction);
      cout << "    translate                  0.0 0.0 " << z << endl;
      break;

    case 5 :
      // move from front to center
      minZ = 1.4;
      maxZ = 0.0;
      z = Math::lerp(minZ,maxZ,fraction);
      cout << "    translate                  0.0 0.0 " << z << endl;
      break;

    case 6 :
      // increase translucency
      minTS = 0.0;
      maxTS = 0.15;
      ts = Math::lerp(minTS,maxTS,fraction);
      cout << "    disableRefraction          1" << endl;
      cout << "    translucencyStrength       " << ts << endl;
      break;
  }

  cout << "    scale                      0.5 0.5 0.5" << endl;
  cout << "    sphere" << endl;
  cout << "  pop" << endl;
  cout << "}" << endl;

  cout << endl;
  cout << "! ===============================" << endl;
  cout << endl;
  cout << "include data/sdl/include/colors.sdl" << endl;
  cout << endl;
  cout << "eyePoint  0.0 0.0 1.5" << endl;
  cout << "viewPoint 0.0 0.0 0.0" << endl;
  cout << endl;
  cout << "light 10.0 10.0 10.0 1.0 1.0 1.0" << endl;
  cout << endl;
  cout << "push use Wall        pop" << endl;
  cout << "push use GlassSphere pop" << endl;
}

void main(int argc, char* argv[])
{
  if (argc != (1 + 3)) {
    cout << endl;
    cout << "Usage : generate-moving-sphere-frame <part> <frames-per-part> <current-frame>" << endl;
    cout << endl;
    cout << "Part 1 moves from left to right, part 2 moves from right to center," << endl;
    cout << "part 3 moves from center to back, part 4 moves from back to front," << endl;
    cout << "part 5 moves from front to center and part 6 increases translucency." << endl;
    cout << endl;
    exit(0);
  }

  generateMovingSphereFrame((int) getValue(argv[1]),(int) getValue(argv[2]),(int) getValue(argv[3]));
}
