// ------------------------------
// Filename      : opengl.cpp
// Author        : Sven Maerivoet
// Last modified : 13/06/2001
// Target        : C++ (OpenGL)
//
// All rights reserved.
// ------------------------------

#ifdef _WIN32
  #include <windows.h>
#endif

#include "constants.h"
#include "chrono.h"
#include "math.h"
#include "opengl.h"
#include "ray.h"
#include "rng.h"
#include <GL/gl.h>
#include <GL/glu.h>
#include <GL/glut.h>
#include <math.h>
#include <iostream>
#include <fstream>

// string-streams are not always compatible
// => Microsoft Visual C++ 6.0 and UNIX/Linux are treated separately
#ifdef _WIN32
  #include <sstream>
#else
  #include <strstream>
  #define  ostringstream ostrstream
#endif

using namespace std;

// --------------------------------------------
// initialization of static const class members
// --------------------------------------------
const int  TOpenGLApp::kQuasiFullScreen           = -1;
const bool TOpenGLApp::kEnableDoubleBuffering     = true;
const bool TOpenGLApp::kDisableDoubleBuffering    = false;
const bool TOpenGLApp::kEnableDepthBuffer         = true;
const bool TOpenGLApp::kDisableDepthBuffer        = false;
const bool TOpenGLApp::kComputeShadows            = true;
const bool TOpenGLApp::kDontComputeShadows        = false;
const bool TOpenGLApp::kUseStochasticSampling     = true;
const bool TOpenGLApp::kDontUseStochasticSampling = false;
const bool TOpenGLApp::kNormalizeColors           = true;
const bool TOpenGLApp::kDontNormalizeColors       = false;
const bool TOpenGLApp::kUseLandscapeMaterials     = true;
const bool TOpenGLApp::kDontUseLandscapeMaterials = false;
const bool TOpenGLApp::kSmoothenVertexNormals     = true;
const bool TOpenGLApp::kDontSmoothenVertexNormals = false;
const bool TOpenGLApp::kApplyTextureMapping       = true;
const bool TOpenGLApp::kDontApplyTextureMapping   = false;

const Color TOpenGLApp::kBlack      = Color(0.00,0.00,0.00);
const Color TOpenGLApp::kBlue       = Color(0.00,0.00,1.00);
const Color TOpenGLApp::kGreen      = Color(0.00,1.00,0.00);
const Color TOpenGLApp::kCyan       = Color(0.00,1.00,1.00);
const Color TOpenGLApp::kRed        = Color(1.00,0.00,0.00);
const Color TOpenGLApp::kMagenta    = Color(1.00,0.00,1.00);
const Color TOpenGLApp::kDarkBrown  = Color(0.54,0.27,0.07);
const Color TOpenGLApp::kBrown      = Color(0.62,0.32,0.18);
const Color TOpenGLApp::kLightBrown = Color(0.80,0.52,0.25);
const Color TOpenGLApp::kDarkGray   = Color(0.25,0.25,0.25);
const Color TOpenGLApp::kGray       = Color(0.50,0.50,0.50);
const Color TOpenGLApp::kLightGray  = Color(0.75,0.75,0.75);
const Color TOpenGLApp::kYellow     = Color(1.00,1.00,0.00);
const Color TOpenGLApp::kWhite      = Color(1.00,1.00,1.00);

const int TOpenGLApp::kLight1 = GL_LIGHT0;
const int TOpenGLApp::kLight2 = GL_LIGHT1;
const int TOpenGLApp::kLight3 = GL_LIGHT2;
const int TOpenGLApp::kLight4 = GL_LIGHT3;
const int TOpenGLApp::kLight5 = GL_LIGHT4;
const int TOpenGLApp::kLight6 = GL_LIGHT5;
const int TOpenGLApp::kLight7 = GL_LIGHT6;
const int TOpenGLApp::kLight8 = GL_LIGHT7;

// -----------
// constructor
// -----------
TOpenGLApp::TOpenGLApp(void)
{
  // initialize the random-number-generator
  RNG::setSeedWithTime();
}

// ----------------
// copy-constructor
// ----------------
TOpenGLApp::TOpenGLApp(const TOpenGLApp& openGLApp)
{
}

// ----------
// destructor
// ----------
TOpenGLApp::~TOpenGLApp(void)
{
}

// --------------
// initialization
// --------------
void TOpenGLApp::initialize(int clientWindowXPos,
                            int clientWindowYPos,
                            int clientWindowXSize,
                            int clientWindowYSize,
                            string clientWindowTitle,
                            bool enableDoubleBuffering,
                            bool enableDepthBuffer)
{
  char* argv[1]; // dummy argument list for glutInit()
  char dummyString[8];
  argv[0] = dummyString; // hook up the pointer
  int argc = 1;
  glutInit(&argc,argv);

  fEnableDoubleBuffering = enableDoubleBuffering;
  fEnableDepthBuffer = enableDepthBuffer;
  disableAntiAliasing();

  // initialize the displaymode
  unsigned int displayMode = GLUT_RGBA;
  if (fEnableDoubleBuffering) {
    displayMode |= GLUT_DOUBLE;
  }
  else {
    displayMode |= GLUT_SINGLE;
  }
  if (fEnableDepthBuffer) {
    displayMode |= GLUT_DEPTH;
  }
  // always set the accumulation-buffer's bit
  displayMode |= GLUT_ACCUM;
  glutInitDisplayMode(displayMode);

  // setup the client-window
  fClientWindowXPos = clientWindowXPos;
  fClientWindowYPos = clientWindowYPos;
  if (clientWindowXSize == kQuasiFullScreen) {
    fClientWindowXSize = glutGet((GLenum) GLUT_SCREEN_WIDTH) - 7;
  }
  else {
    fClientWindowXSize = clientWindowXSize;
  }
  if (clientWindowYSize == kQuasiFullScreen) {
    fClientWindowYSize = glutGet((GLenum) GLUT_SCREEN_HEIGHT) - 60;
  }
  else {
    fClientWindowYSize = clientWindowYSize;
  }
  fClientWindowTitle = clientWindowTitle;
  fFullScreenMode = false;

  // set the client-window's position, size and title
  glutInitWindowPosition(fClientWindowXPos,fClientWindowXPos);
  glutInitWindowSize(fClientWindowXSize,fClientWindowYSize);
  glutCreateWindow(fClientWindowTitle.c_str());

  // set user's rendering-preferences
  if (fEnableDepthBuffer) {
    glEnable(GL_DEPTH_TEST);
    // automatically normalize all normal-vectors
    glEnable(GL_NORMALIZE);
  }

  // let textures be correctly displayed
  glHint(GL_PERSPECTIVE_CORRECTION_HINT,GL_NICEST);
  enableGoodTextureInterpolation();
  disableGlowingTextures();

  // make the world-window as large as the client-window in use
  setWorldWindow(0.0,(double) fClientWindowYSize,(double) fClientWindowXSize,0.0);

  // match the viewport to the world-window
  setViewport(0,fClientWindowYSize,fClientWindowXSize,0);

  fCurrentPosition.set(0.0,0.0,0.0);
  fCurrentDirection = 0.0;

  setCameraSlideStepSize(10.0);
  setCameraRotateStepSize(5.0);
  setCameraViewAngleStepSize(5.0);
  setCameraAspectRatioStepSize(0.1);
  setCameraAdvanceStepSize(10.0);
  disableContinuousCameraMovements();
  disableCameraMomentum();
  coupleCameraYawAndRoll();
  fCameraNrOfControlPointsOnPath = 0;

  fSnapshotSequenceNr = 0;

  fRays = new Ray[9];

  explicitInitialize();
}

// -----------------------------------------------
// explicit initialization (override if necessary)
// -----------------------------------------------
void TOpenGLApp::explicitInitialize(void)
{
}

// ----------------------
// callback registrations
// ----------------------
void TOpenGLApp::registerDisplayFunc(void (*displayFuncPtr) (void)) const
{
  glutDisplayFunc(displayFuncPtr);
}

void TOpenGLApp::registerReshapeFunc(void (*reshapeFuncPtr) (int width, int height)) const
{
  glutReshapeFunc(reshapeFuncPtr);
}

void TOpenGLApp::registerKeyboardFunc(void (*keyboardFuncPtr) (unsigned char key, int x, int y)) const
{
  glutKeyboardFunc(keyboardFuncPtr);
}

void TOpenGLApp::registerSpecialFunc(void (*specialFuncPtr) (int key, int x, int y)) const
{
  glutSpecialFunc(specialFuncPtr);
}

void TOpenGLApp::registerMouseFunc(void (*mouseFuncPtr) (int button, int state, int x, int y)) const
{
  glutMouseFunc(mouseFuncPtr);
}

void TOpenGLApp::registerMotionFunc(void (*motionFuncPtr) (int x, int y)) const
{
  glutMotionFunc(motionFuncPtr);
}

void TOpenGLApp::registerPassiveMotionFunc(void (*passiveMotionFuncPtr) (int x, int y)) const
{
  glutPassiveMotionFunc(passiveMotionFuncPtr);
}

void TOpenGLApp::registerIdleFunc(void (*idleFuncPtr) (void)) const
{
  glutIdleFunc(idleFuncPtr);
}

// ------------------------------------------
// callback-functions (override if necessary)
// ------------------------------------------
void TOpenGLApp::displayFunc(void)
{
  // show a white cross in a blue window
  setBackgroundColor(kBlue);
  clearScreen();

  setColor(kWhite);
  line (0.0,0.0,(double) fClientWindowXSize,(double) fClientWindowYSize);
  line (0.0,(double) fClientWindowXSize,(double) fClientWindowYSize,0.0);
}

void TOpenGLApp::reshapeFunc(int width, int height)
{
  double viewportAspectRatio = (double) width / (double) height;

  fClientWindowXSize = width;
  fClientWindowYSize = height;

  // set the largest matching viewport when resizing occurs
  // this maintains the aspect-ratio
  if (fWorldWindowAspectRatio > viewportAspectRatio) {
    setViewport(0,width / fWorldWindowAspectRatio,width,0);
  }
  else {
    setViewport(0,height,height * fWorldWindowAspectRatio,0);
  }
}

void TOpenGLApp::keyboardFunc(unsigned char key, int x, int y)
{
  fScheduleRedisplay = true;

  if (!fCameraHasMomentum) {
    // disable/enable continuous camera movements
    if (key == 'c') {
      disableContinuousCameraMovements();
    }
    else if (key == 'C') {
      enableContinuousCameraMovements();
    }
  }

  const char kEscapeKey = 0x1B;

  switch (key) {

    // change the camera's view-volume
    case 'v' :
      setCameraPerspectiveShape(getCameraViewAngle() - fCameraViewAngleStepSize,
                                getCameraAspectRatio(),
                                getCameraDistanceOfNearPlane(),
                                getCameraDistanceOfFarPlane());
      break;

    case 'V' :
      setCameraPerspectiveShape(getCameraViewAngle() + fCameraViewAngleStepSize,
                                getCameraAspectRatio(),
                                getCameraDistanceOfNearPlane(),
                                getCameraDistanceOfFarPlane());
      break;

    case 'a' :
      setCameraPerspectiveShape(getCameraViewAngle(),
                                getCameraAspectRatio() - fCameraAspectRatioStepSize,
                                getCameraDistanceOfNearPlane(),
                                getCameraDistanceOfFarPlane());
      break;

    case 'A' :
      setCameraPerspectiveShape(getCameraViewAngle(),
                                getCameraAspectRatio() + fCameraAspectRatioStepSize,
                                getCameraDistanceOfNearPlane(),
                                getCameraDistanceOfFarPlane());
      break;

    case 'n' :
      if ((getCameraDistanceOfNearPlane() - fCameraSlideStepSize) < 0.001) {
        setCameraPerspectiveShape(getCameraViewAngle(),
                                  getCameraAspectRatio(),
                                  0.001,
                                  getCameraDistanceOfFarPlane());
      }
      else {
        setCameraPerspectiveShape(getCameraViewAngle(),
                                  getCameraAspectRatio(),
                                  getCameraDistanceOfNearPlane() - fCameraSlideStepSize,
                                  getCameraDistanceOfFarPlane());
      }
      break;

    case 'N' :
      setCameraPerspectiveShape(getCameraViewAngle(),
                                getCameraAspectRatio(),
                                getCameraDistanceOfNearPlane() + fCameraSlideStepSize,
                                getCameraDistanceOfFarPlane());
      break;

    case 'f' :
      if ((getCameraDistanceOfFarPlane() - fCameraSlideStepSize) < getCameraDistanceOfNearPlane()) {
        setCameraPerspectiveShape(getCameraViewAngle(),
                                  getCameraAspectRatio(),
                                  getCameraDistanceOfNearPlane(),
                                  2.0 * getCameraDistanceOfNearPlane());
      }
      else {
        setCameraPerspectiveShape(getCameraViewAngle(),
                                  getCameraAspectRatio(),
                                  getCameraDistanceOfNearPlane(),
                                  getCameraDistanceOfFarPlane() - fCameraSlideStepSize);
      }
      break;

    case 'F' :
      setCameraPerspectiveShape(getCameraViewAngle(),
                                getCameraAspectRatio(),
                                getCameraDistanceOfNearPlane(),
                                getCameraDistanceOfFarPlane() + fCameraSlideStepSize);
      break;

    // lights and shading
    case 'l' :
      switchLightingOff();
      break;

    case 'L' :
      switchLightingOn();
      break;

    case 's' :
      enableFlatShading();
      break;

    case 'S' :
      enableSmoothShading();
      break;

    case 'e' :
      disableAntiAliasing();
      break;

    case 'E' :
      enableAntiAliasing(0.4,kJitterVectors4);
      break;

    // camera momentum
    case 'm' :
      disableCameraMomentum();
      break;

    case 'M' :
      enableCameraMomentum();
      break;

    case 'z' :
      fCameraSlideStepSize /= 2.0;
      fCameraRotateStepSize /= 2.0;
      fScheduleRedisplay = false;
      break;

    case 'Z' :
      fCameraSlideStepSize *= 2.0;
      fCameraRotateStepSize *= 2.0;
      fScheduleRedisplay = false;
      break;

    case 'y' :
      decoupleCameraYawAndRoll();
      fScheduleRedisplay = false;
      break;

    case 'Y' :
      coupleCameraYawAndRoll();
      fScheduleRedisplay = false;
      break;

    // full-screen toggle
    case 'x' : 
      if (fFullScreenMode) {
        fFullScreenMode = false;
        fClientWindowXPos = fOldClientWindowXPos;
        fClientWindowYPos = fOldClientWindowYPos;
        fClientWindowXSize = fOldClientWindowXSize;
        fClientWindowYSize = fOldClientWindowYSize;
        glutPositionWindow(fClientWindowXPos,fClientWindowYPos);
        glutReshapeWindow(fClientWindowXSize,fClientWindowYSize);
        redisplay();
      }
      break;

    case 'X' :
      if (!fFullScreenMode) {
        fOldClientWindowXPos = fClientWindowXPos;
        fOldClientWindowYPos = fClientWindowYPos;
        fOldClientWindowXSize = fClientWindowXSize;
        fOldClientWindowYSize = fClientWindowYSize;
        fFullScreenMode = true;
        glutFullScreen();
        redisplay();
      }
      break;

    // snapshot-taking
    case 't' : 
    case 'T' :
      takeSnapshot("snapshot" + convertDoubleToString((double) fSnapshotSequenceNr) + ".bmp");
      ++fSnapshotSequenceNr;
      fScheduleRedisplay = false;
      break;

    // program termination
    case kEscapeKey :
      exit(0);
      break;
  }

  if (fScheduleRedisplay) {
    redisplay();
  }
}

void TOpenGLApp::specialFunc(int key, int x, int y)
{
  switch (key) {

    // orient the camera in its own coordinate system
    case GLUT_KEY_LEFT :
      yawCamera(-fCameraRotateStepSize);
      if (fCameraYawAndRollCoupled) {
        rollCamera(-fCameraRotateStepSize);
      }
      fCurrentCameraMovement = kCameraYawing;
      fCameraYaw = kYawLeft;
      break;

    case GLUT_KEY_RIGHT :
      yawCamera(fCameraRotateStepSize);
      if (fCameraYawAndRollCoupled) {
        rollCamera(+fCameraRotateStepSize);
      }
      fCurrentCameraMovement = kCameraYawing;
      fCameraYaw = kYawRight;
      break;

    case GLUT_KEY_UP :
      pitchCamera(fCameraRotateStepSize);
      fCurrentCameraMovement = kCameraPitching;
      fCameraPitch = kPitchDown;
      break;

    case GLUT_KEY_DOWN :
      pitchCamera(-fCameraRotateStepSize);
      fCurrentCameraMovement = kCameraPitching;
      fCameraPitch = kPitchUp;
      break;

    case GLUT_KEY_PAGE_UP :
      rollCamera(-fCameraRotateStepSize);
      fCurrentCameraMovement = kCameraRotating;
      fCameraRotate = kRotateLeft;
      break;

    case GLUT_KEY_PAGE_DOWN :
      rollCamera(fCameraRotateStepSize);
      fCurrentCameraMovement = kCameraRotating;
      fCameraRotate = kRotateRight;
      break;
  }

  if (!fCameraHasMomentum) {

    Point3D cameraEyePoint;

    switch (key) {

      // slide the camera in the world-coordinate system
      case GLUT_KEY_F1 :
        cameraEyePoint = getCameraEyePoint();
        cameraEyePoint.fX -= fCameraSlideStepSize;
        setCamera(cameraEyePoint,getCameraViewPoint(),getCameraUpDirection());
        fCurrentCameraMovement = kCameraMoving;
        fCameraMove = kMoveXLeft;
        break;

      case GLUT_KEY_F2 :
        cameraEyePoint = getCameraEyePoint();
        cameraEyePoint.fX += fCameraSlideStepSize;
        setCamera(cameraEyePoint,getCameraViewPoint(),getCameraUpDirection());
        fCurrentCameraMovement = kCameraMoving;
        fCameraMove = kMoveXRight;
        break;

      case GLUT_KEY_F3 :
        cameraEyePoint = getCameraEyePoint();
        cameraEyePoint.fY -= fCameraSlideStepSize;
        setCamera(cameraEyePoint,getCameraViewPoint(),getCameraUpDirection());
        fCurrentCameraMovement = kCameraMoving;
        fCameraMove = kMoveYDown;
        break;

      case GLUT_KEY_F4 :
        cameraEyePoint = getCameraEyePoint();
        cameraEyePoint.fY += fCameraSlideStepSize;
        setCamera(cameraEyePoint,getCameraViewPoint(),getCameraUpDirection());
        fCurrentCameraMovement = kCameraMoving;
        fCameraMove = kMoveYUp;
        break;

      case GLUT_KEY_F5 :
        cameraEyePoint = getCameraEyePoint();
        cameraEyePoint.fZ -= fCameraSlideStepSize;
        setCamera(cameraEyePoint,getCameraViewPoint(),getCameraUpDirection());
        fCurrentCameraMovement = kCameraMoving;
        fCameraMove = kMoveZForward;
        break;

      case GLUT_KEY_F6 :
        cameraEyePoint = getCameraEyePoint();
        cameraEyePoint.fZ += fCameraSlideStepSize;
        setCamera(cameraEyePoint,getCameraViewPoint(),getCameraUpDirection());
        fCurrentCameraMovement = kCameraMoving;
        fCameraMove = kMoveZBackward;
        break;

      // slide the camera in its own coordinate system
      case GLUT_KEY_F7 :
        slideCamera(-fCameraSlideStepSize,0.0,0.0);
        fCurrentCameraMovement = kCameraSliding;
        fCameraSlide = kSlideXLeft;
        break;

      case GLUT_KEY_F8 :
        slideCamera(fCameraSlideStepSize,0.0,0.0);
        fCurrentCameraMovement = kCameraSliding;
        fCameraSlide = kSlideXRight;
        break;

      case GLUT_KEY_F9 :
        slideCamera(0.0,-fCameraSlideStepSize,0.0);
        fCurrentCameraMovement = kCameraSliding;
        fCameraSlide = kSlideYDown;
        break;

      case GLUT_KEY_F10 :
        slideCamera(0.0,fCameraSlideStepSize,0.0);
        fCurrentCameraMovement = kCameraSliding;
        fCameraSlide = kSlideYUp;
        break;

      case GLUT_KEY_F11 :
        slideCamera(0.0,0.0,-fCameraSlideStepSize);
        fCurrentCameraMovement = kCameraSliding;
        fCameraSlide = kSlideZForward;
        break;

      case GLUT_KEY_F12 :
        slideCamera(0.0,0.0,fCameraSlideStepSize);
        fCurrentCameraMovement = kCameraSliding;
        fCameraSlide = kSlideZBackward;
        break;
    }
  }

  redisplay();
}

void TOpenGLApp::mouseFunc(int button, int state, int x, int y)
{
}

void TOpenGLApp::motionFunc(int x, int y)
{
}

void TOpenGLApp::passiveMotionFunc(int x, int y)
{
}

void TOpenGLApp::idleFunc(void)
{
  if (fCameraHasMomentum) {
    // advance along its own negative Z-direction
    slideCamera(0.0,0.0,-fCameraSlideStepSize);
  }
  else {

    if ((!fContinuousCameraMovements) ||
        (fCurrentCameraMovement == kCameraStill)) {
      return;
    }

    Point3D cameraEyePoint;

    switch (fCurrentCameraMovement) {

      case kCameraMoving :
        switch (fCameraMove) {

          case kMoveXLeft :
            cameraEyePoint = getCameraEyePoint();
            cameraEyePoint.fX -= fCameraSlideStepSize;
            setCamera(cameraEyePoint,getCameraViewPoint(),getCameraUpDirection());
            break;

          case kMoveXRight :
            cameraEyePoint = getCameraEyePoint();
            cameraEyePoint.fX += fCameraSlideStepSize;
            setCamera(cameraEyePoint,getCameraViewPoint(),getCameraUpDirection());
            break;

          case kMoveYDown :
            cameraEyePoint = getCameraEyePoint();
            cameraEyePoint.fY -= fCameraSlideStepSize;
            setCamera(cameraEyePoint,getCameraViewPoint(),getCameraUpDirection());
            break;

          case kMoveYUp :
            cameraEyePoint = getCameraEyePoint();
            cameraEyePoint.fY += fCameraSlideStepSize;
            setCamera(cameraEyePoint,getCameraViewPoint(),getCameraUpDirection());
            break;

          case kMoveZForward :
            cameraEyePoint = getCameraEyePoint();
            cameraEyePoint.fZ -= fCameraSlideStepSize;
            setCamera(cameraEyePoint,getCameraViewPoint(),getCameraUpDirection());
            break;

          case kMoveZBackward :
            cameraEyePoint = getCameraEyePoint();
            cameraEyePoint.fZ += fCameraSlideStepSize;
            setCamera(cameraEyePoint,getCameraViewPoint(),getCameraUpDirection());
            break;
        }
        break;

      case kCameraSliding :
        switch (fCameraSlide) {

          case kSlideXLeft :
            slideCamera(-fCameraSlideStepSize,0.0,0.0);
            break;

          case kSlideXRight : 
            slideCamera(fCameraSlideStepSize,0.0,0.0);
            break;

          case kSlideYDown :
            slideCamera(0.0,-fCameraSlideStepSize,0.0);
            break;

          case kSlideYUp :
            slideCamera(0.0,fCameraSlideStepSize,0.0);
            break;

          case kSlideZForward :
            slideCamera(0.0,0.0,-fCameraSlideStepSize);
            break;

          case kSlideZBackward :
            slideCamera(0.0,0.0,fCameraSlideStepSize);
            break;
        }
        break;

      case kCameraRotating :
        switch (fCameraRotate) {

          case kRotateLeft :
            rollCamera(-fCameraRotateStepSize);
            break;

          case kRotateRight :
            rollCamera(fCameraRotateStepSize);
            break;
        }
        break;

      case kCameraPitching :
        switch (fCameraPitch) {

          case kPitchDown :
            pitchCamera(fCameraRotateStepSize);
            break;

          case kPitchUp :
            pitchCamera(-fCameraRotateStepSize);
            break;
        }
        break;

      case kCameraYawing :
        switch (fCameraYaw) {

          case kYawLeft :
            yawCamera(-fCameraRotateStepSize);
            if (fCameraYawAndRollCoupled) {
              rollCamera(-fCameraRotateStepSize);
            }
            break;

          case kYawRight :
            yawCamera(+fCameraRotateStepSize);
            if (fCameraYawAndRollCoupled) {
              rollCamera(fCameraRotateStepSize);
            }
            break;
        }
        break;
    }
  }

  redisplay();
}

// -----------------------------------
// activator for GLUT's event-handling
// -----------------------------------
void TOpenGLApp::activateEventHandling(void) const
{
  glutMainLoop();
}

// -------------------------------------------------
// the real display-callback (don't call directly !)
// -------------------------------------------------
void TOpenGLApp::antiAliasedDisplayFunc(void)
{
  // note that this method isn't declared protected or private but public
  // because a program needs to register it to the OpenGL-system from
  // outside the TOpenGLApp-class !

  if (fAntiAliasingEnabled) {
    glClear(GL_ACCUM_BUFFER_BIT);
    for (int jitterVector = 0; jitterVector < fNrOfJitterVectors; ++jitterVector) {
      slideCamera(fJitterFactor * fJitterVectors->getX(jitterVector),
                  fJitterFactor * fJitterVectors->getY(jitterVector),
                  0.0);
      displayFunc();
      glAccum(GL_ACCUM,(GLfloat) (1.0 / fNrOfJitterVectors));
    }
    glAccum(GL_RETURN,(GLfloat) 1.0);
  }
  else {
    displayFunc();
  }

  finishFrame();
}

// -------------
// miscellaneous
// -------------
void TOpenGLApp::enableAntiAliasing(double jitterFactor, EJitterVectors nrOfJitterVectors)
{
  fJitterFactor = jitterFactor;

  switch (nrOfJitterVectors) {

    default :

    case kJitterVectors2 :
      fNrOfJitterVectors = 2;
      break;

    case kJitterVectors3 :
      fNrOfJitterVectors = 3;
      break;

    case kJitterVectors4 :
      fNrOfJitterVectors = 4;
      break;

    case kJitterVectors8 :
      fNrOfJitterVectors = 8;
      break;

    case kJitterVectors15 :
      fNrOfJitterVectors = 15;
      break;

    case kJitterVectors24 :
      fNrOfJitterVectors = 24;
      break;

    case kJitterVectors66 :
      fNrOfJitterVectors = 66;
      break;
  }

  fJitterVectors = new JitterVectors(fNrOfJitterVectors);

  fAntiAliasingEnabled = true;
}

void TOpenGLApp::disableAntiAliasing(void)
{
  fAntiAliasingEnabled = false;
}

void TOpenGLApp::redisplay(void) const
{
  glutPostRedisplay();
}

int TOpenGLApp::getClientWindowXSize(void) const
{
  return fClientWindowXSize;
}

int TOpenGLApp::getClientWindowYSize(void) const
{
  return fClientWindowYSize;
}

void TOpenGLApp::takeSnapshot(string filename) const
{
  RGBApixmap rgbapixmap;
  rgbapixmap.copyFromScreen(0,0,fClientWindowXSize,fClientWindowYSize);
  rgbapixmap.writeToBMPFile(filename);
}

// ---------------------------------
// OpenGL window-to-viewport mapping
// ---------------------------------
void TOpenGLApp::setWorldWindow(double left, double top, double right, double bottom)
{
  glMatrixMode(GL_PROJECTION);
  glLoadIdentity();
  gluOrtho2D((GLdouble) left,(GLdouble) right,(GLdouble) bottom,(GLdouble) top);

  fWorldWindow.set(left,top,right,bottom);

  // determine the aspect-ratio
  if (top == bottom) {
    fWorldWindowAspectRatio = 1.0; // beware (prevent division by zero) !
  }
  else {
    fWorldWindowAspectRatio = (right - left) / (top - bottom);
  }

  glMatrixMode(GL_MODELVIEW);
}

void TOpenGLApp::setViewport(int left, int top, int right, int bottom)
{
  glViewport((GLint) left,(GLint) bottom,(GLint) (right - left),(GLint) (top - bottom));
  glMatrixMode(GL_MODELVIEW);
}

Rectangle2D TOpenGLApp::getWorldWindow(void) const
{
  return fWorldWindow;
}

double TOpenGLApp::getWorldWindowAspectRatio(void) const
{
  return fWorldWindowAspectRatio;
}

Rectangle2D TOpenGLApp::getViewport(void) const
{
  return fViewport;
}

// ----------------------------
// OpenGL projection management
// ----------------------------
void TOpenGLApp::setCamera(const Point3D& cameraEyePoint,
                           const Point3D& cameraViewPoint,
                           const Vector3D& cameraUpDirection)
{
  fCameraEyePoint = cameraEyePoint;
  fCameraViewPoint = cameraViewPoint;
  fCameraUpDirection = cameraUpDirection;

  fCameraN.fX = fCameraEyePoint.fX - fCameraViewPoint.fX;
  fCameraN.fY = fCameraEyePoint.fY - fCameraViewPoint.fY;
  fCameraN.fZ = fCameraEyePoint.fZ - fCameraViewPoint.fZ;

  fCameraU = fCameraUpDirection.outer(fCameraN);

  fCameraV = fCameraN.outer(fCameraU);

  fCameraN.normalize();
  fCameraU.normalize();
  fCameraV.normalize();

  setModelViewMatrix();
}

Point3D TOpenGLApp::getCameraEyePoint(void) const
{
  return fCameraEyePoint;
}

Point3D TOpenGLApp::getCameraViewPoint(void) const
{
  return fCameraViewPoint;
}

Vector3D TOpenGLApp::getCameraUpDirection(void) const
{
  return fCameraUpDirection;
}

void TOpenGLApp::slideCamera(double du, double dv, double dn)
{
  fCameraEyePoint.fX += (du * fCameraU.fX) + (dv * fCameraV.fX) + (dn * fCameraN.fX);
  fCameraEyePoint.fY += (du * fCameraU.fY) + (dv * fCameraV.fY) + (dn * fCameraN.fY);
  fCameraEyePoint.fZ += (du * fCameraU.fZ) + (dv * fCameraV.fZ) + (dn * fCameraN.fZ);

  setModelViewMatrix();
}

void TOpenGLApp::rollCamera(double degrees)
{
  double cosine = cos(Math::degToRad(degrees));
  double sine = sin(Math::degToRad(degrees));

  Vector3D t = fCameraU;
  fCameraU.fX = (cosine * t.fX) - (sine * fCameraV.fX);
  fCameraU.fY = (cosine * t.fY) - (sine * fCameraV.fY);
  fCameraU.fZ = (cosine * t.fZ) - (sine * fCameraV.fZ);
  fCameraV.fX = (sine * t.fX) + (cosine * fCameraV.fX);
  fCameraV.fY = (sine * t.fY) + (cosine * fCameraV.fY);
  fCameraV.fZ = (sine * t.fZ) + (cosine * fCameraV.fZ);

  /*
    ALTERNATIVE :
    -------------
    rotateCameraAroundAxes(fCameraU,fCameraV,-degrees);
  */

  setModelViewMatrix();
}

void TOpenGLApp::pitchCamera(double degrees)
{
  rotateCameraAroundAxes(fCameraN,fCameraV,degrees);
  setModelViewMatrix();
}

void TOpenGLApp::yawCamera(double degrees)
{
  rotateCameraAroundAxes(fCameraU,fCameraN,degrees);
  setModelViewMatrix();
}

void TOpenGLApp::setCameraPerspectiveShape(double cameraViewAngleInDegrees,
                                           double cameraAspectRatio,
                                           double cameraDistanceOfNearPlane,
                                           double cameraDistanceOfFarPlane)
{
  if (cameraViewAngleInDegrees < 0.1) {
    cameraViewAngleInDegrees = 0.1;
  }
  else if (cameraViewAngleInDegrees >= 180.0) {
    cameraViewAngleInDegrees = 179.0;
  }

  fCameraViewAngle = cameraViewAngleInDegrees;
  fCameraAspectRatio = cameraAspectRatio;
  fCameraDistanceOfNearPlane = cameraDistanceOfNearPlane;
  fCameraDistanceOfFarPlane = cameraDistanceOfFarPlane;

  // calculate the other camera's view-volume-planes
  double cameraHeight = fCameraDistanceOfNearPlane * tan(Math::degToRad(fCameraViewAngle / 2.0));
  double cameraWidth = fCameraAspectRatio * cameraHeight;

  fCameraTopPlane = (cameraHeight / 2.0);
  fCameraBottomPlane = -fCameraTopPlane;
  fCameraRightPlane = (cameraWidth / 2.0);
  fCameraLeftPlane = -fCameraRightPlane;

  glMatrixMode(GL_PROJECTION);
  glLoadIdentity();
  gluPerspective((GLdouble) fCameraViewAngle,
                 (GLdouble) fCameraAspectRatio,
                 (GLdouble) fCameraDistanceOfNearPlane,
                 (GLdouble) fCameraDistanceOfFarPlane);

  glMatrixMode(GL_MODELVIEW);
}

void TOpenGLApp::setCameraPerspectiveShape(double cameraLeftPlane,
                                           double cameraTopPlane,
                                           double cameraRightPlane,
                                           double cameraBottomPlane,
                                           double cameraDistanceOfNearPlane,
                                           double cameraDistanceOfFarPlane)
{
  fCameraLeftPlane = cameraLeftPlane;
  fCameraTopPlane = cameraTopPlane;
  fCameraRightPlane = cameraRightPlane;
  fCameraBottomPlane = cameraBottomPlane;
  fCameraDistanceOfNearPlane = cameraDistanceOfNearPlane;
  fCameraDistanceOfFarPlane = cameraDistanceOfFarPlane;

  double m[16];

  m[0]  = (2.0 * fCameraDistanceOfNearPlane) /
          (fCameraRightPlane - fCameraLeftPlane);
  m[1]  = 0.0;
  m[2]  = 0.0;
  m[3]  = 0.0;

  m[4]  = 0.0;
  m[5]  = (2.0 * fCameraDistanceOfNearPlane) /
          (fCameraTopPlane - fCameraBottomPlane);
  m[6]  = 0.0;
  m[7]  = 0.0;

  m[8]  = (fCameraRightPlane + fCameraLeftPlane) /
          (fCameraRightPlane - fCameraLeftPlane);
  m[9]  = (fCameraTopPlane + fCameraBottomPlane) /
          (fCameraTopPlane - fCameraBottomPlane);
  m[10] = -(fCameraDistanceOfFarPlane + fCameraDistanceOfNearPlane) /
           (fCameraDistanceOfFarPlane - fCameraDistanceOfNearPlane);
  m[11] = -1.0;

  m[12] = 0.0;
  m[13] = 0.0;
  m[14] = -(2.0 * fCameraDistanceOfFarPlane * fCameraDistanceOfNearPlane) /
           (fCameraDistanceOfFarPlane - fCameraDistanceOfNearPlane);
  m[15] = 0.0;

  glMatrixMode(GL_PROJECTION);
  glLoadMatrixd(m);

  /*
    ALTERNATIVE :
    -------------

    glFrustrum((GLdouble) fCameraLeftPlane,
               (GLdouble) fCameraRightPlane,
               (GLdouble) fCameraBottomPlane,
               (GLdouble) fCameraTopPlane,
               (GLdouble) fCameraDistanceOfNearPlane,
               (GLdouble) fCameraDistanceOfFarPlane);
  */

  // calculate the other camera's view-volume-properties
  fCameraViewAngle = 2.0 * Math::radToDeg(Math::arctan(fCameraTopPlane,fCameraDistanceOfNearPlane));
  fCameraAspectRatio = (fCameraRightPlane - fCameraLeftPlane) /
                       (fCameraTopPlane - fCameraBottomPlane);

  glMatrixMode(GL_MODELVIEW);
}

void TOpenGLApp::setCameraObliqueShape(const Vector3D& projectionDirection,
                                       double cameraLeftPlane,
                                       double cameraTopPlane,
                                       double cameraRightPlane,
                                       double cameraBottomPlane,
                                       double cameraDistanceOfNearPlane,
                                       double cameraDistanceOfFarPlane)
{
  fCameraLeftPlane = cameraLeftPlane;
  fCameraTopPlane = cameraTopPlane;
  fCameraRightPlane = cameraRightPlane;
  fCameraBottomPlane = cameraBottomPlane;
  fCameraDistanceOfNearPlane = cameraDistanceOfNearPlane;
  fCameraDistanceOfFarPlane = cameraDistanceOfFarPlane;

  glMatrixMode(GL_PROJECTION);
  glLoadIdentity();
  glOrtho((GLdouble) fCameraLeftPlane,
          (GLdouble) fCameraRightPlane,
          (GLdouble) fCameraBottomPlane,
          (GLdouble) fCameraTopPlane,
          (GLdouble) fCameraDistanceOfNearPlane,
          (GLdouble) fCameraDistanceOfNearPlane);

  if (projectionDirection.fZ == 0.0) {
    return;
  }

  double m[16];
  for (int i = 0; i < 16; ++i) {
    if ((i % 5) == 0) {
      m[i] = 1.0;
    }
    else {
      m[i] = 0.0;
    }
  }

  m[8] = -projectionDirection.fX / projectionDirection.fZ;
  m[9] = -projectionDirection.fY / projectionDirection.fZ;

  glMultMatrixd(m);
}

void TOpenGLApp::setCameraOrthogonalShape(const Vector3D& projectionDirection,
                                          double cameraLeftPlane,
                                          double cameraTopPlane,
                                          double cameraRightPlane,
                                          double cameraBottomPlane,
                                          double cameraDistanceOfNearPlane,
                                          double cameraDistanceOfFarPlane)
{
  setCameraObliqueShape(Vector3D(0.0,0.0,1.0),
                        cameraLeftPlane,cameraTopPlane,
                        cameraRightPlane,cameraBottomPlane,
                        cameraDistanceOfNearPlane,cameraDistanceOfFarPlane);
}

double TOpenGLApp::getCameraViewAngle(void) const
{
  return fCameraViewAngle;
}

double TOpenGLApp::getCameraAspectRatio(void) const
{
  return fCameraAspectRatio;
}

double TOpenGLApp::getCameraLeftPlane(void) const
{
  return fCameraLeftPlane;
}

double TOpenGLApp::getCameraTopPlane(void) const
{
  return fCameraTopPlane;
}

double TOpenGLApp::getCameraRightPlane(void) const
{
  return fCameraRightPlane;
}

double TOpenGLApp::getCameraBottomPlane(void) const
{
  return fCameraBottomPlane;
}

double TOpenGLApp::getCameraDistanceOfNearPlane(void) const
{
  return fCameraDistanceOfNearPlane;
}

double TOpenGLApp::getCameraDistanceOfFarPlane(void) const
{
  return fCameraDistanceOfFarPlane;
}

void TOpenGLApp::setCameraSlideStepSize(double cameraSlideStepSize)
{
  if (cameraSlideStepSize >= 0.0) {
    fCameraSlideStepSize = cameraSlideStepSize;
  }
  else {
    fCameraSlideStepSize = 0.0;
  }
}

void TOpenGLApp::setCameraRotateStepSize(double cameraRotateStepSizeInDegrees)
{
  if (cameraRotateStepSizeInDegrees >= 0.0) {
    fCameraRotateStepSize = cameraRotateStepSizeInDegrees;
  }
  else {
    fCameraRotateStepSize = 0.0;
  }
}

void TOpenGLApp::setCameraViewAngleStepSize(double cameraViewAngleStepSizeInDegrees)
{
  if (cameraViewAngleStepSizeInDegrees >= 0.0) {
    fCameraViewAngleStepSize = cameraViewAngleStepSizeInDegrees;
  }
  else {
    cameraViewAngleStepSizeInDegrees = 0.0;
  }
}

void TOpenGLApp::setCameraAspectRatioStepSize(double cameraAspectRatioStepSize)
{
  if (cameraAspectRatioStepSize >= 0.0) {
    fCameraAspectRatioStepSize = cameraAspectRatioStepSize;
  }
  else {
    fCameraAspectRatioStepSize = 0.0;
  }
}

void TOpenGLApp::setCameraAdvanceStepSize(double cameraAdvanceStepSize)
{
  if (cameraAdvanceStepSize >= 0.0) {
    fCameraAdvanceStepSize = cameraAdvanceStepSize;
  }
  else {
    fCameraAdvanceStepSize = 0.0;
  }
}

double TOpenGLApp::getCameraSlideStepSize(void) const
{
  return fCameraSlideStepSize;
}

double TOpenGLApp::getCameraRotateStepSize(void) const
{
  return fCameraRotateStepSize;
}

double TOpenGLApp::getCameraViewAngleStepSize(void) const
{
  return fCameraViewAngleStepSize;
}

double TOpenGLApp::getCameraAspectRatioStepSize(void) const
{
  return fCameraAspectRatioStepSize;
}

double TOpenGLApp::getCameraAdvanceStepSize(void) const
{
  return fCameraAdvanceStepSize;
}

void TOpenGLApp::loadCameraPath(string filename)
{
  fCameraNrOfControlPointsOnPath = 0;

  fstream inputFileStream;

  inputFileStream.open(filename.c_str(),ios::in);
  if (inputFileStream.fail()) {
    return;
  }

  inputFileStream >> fCameraNrOfControlPointsOnPath;

  // a path is defined by at least two points
  if (fCameraNrOfControlPointsOnPath >= 2) {

    fCameraControlPointsOnPath = new Point3D[fCameraNrOfControlPointsOnPath];

    for (int i = 0; i < fCameraNrOfControlPointsOnPath; ++i) {
      inputFileStream >> fCameraControlPointsOnPath[i].fX
                      >> fCameraControlPointsOnPath[i].fY
                      >> fCameraControlPointsOnPath[i].fZ; 
    }
  }
  else {
    fCameraNrOfControlPointsOnPath = 0;
  }

  inputFileStream.close();

  fCameraDistanceOnSegmentTraveled = 0.0;
  fCameraLastControlPointVisited = 0;
}

bool TOpenGLApp::advanceCameraAlongPath(void)
{
  if (fCameraNrOfControlPointsOnPath == 0) {
    return false;
  }

  if (fCameraLastControlPointVisited >= (fCameraNrOfControlPointsOnPath - 1)) {
    return false;
  }

  Point3D p = fCameraControlPointsOnPath[fCameraLastControlPointVisited];
  Point3D q = fCameraControlPointsOnPath[fCameraLastControlPointVisited + 1];
  double segmentDistance = sqrt(Math::sqr(p.fX - q.fX) +
                                Math::sqr(p.fY - q.fY) +
                                Math::sqr(p.fZ - q.fZ));

  // linearly interpolate the path between p and q
  double factor = fCameraDistanceOnSegmentTraveled / segmentDistance;
  Point3D cameraEyePoint(p.fX + ((q.fX - p.fX) * factor),
                         p.fY + ((q.fY - p.fY) * factor),
                         p.fZ + ((q.fZ - p.fZ) * factor));
  setCamera(cameraEyePoint,getCameraViewPoint(),getCameraUpDirection());

  // advance camera
  fCameraDistanceOnSegmentTraveled += fCameraAdvanceStepSize;

  // check if we pass the next control-point (i.e.: q)
  if (fCameraDistanceOnSegmentTraveled >= segmentDistance) {
    ++fCameraLastControlPointVisited;
    fCameraDistanceOnSegmentTraveled -= segmentDistance;
  }

  return true;
}

void TOpenGLApp::enableContinuousCameraMovements(void)
{
  fContinuousCameraMovements = true;
  fCurrentCameraMovement = kCameraStill;
}

void TOpenGLApp::disableContinuousCameraMovements(void)
{
  fContinuousCameraMovements = false;
  fCurrentCameraMovement = kCameraStill;
}

void TOpenGLApp::enableCameraMomentum(void)
{
  fCameraHasMomentum = true;
  disableContinuousCameraMovements();
}

void TOpenGLApp::disableCameraMomentum(void)
{
  fCameraHasMomentum = false;
}

void TOpenGLApp::decoupleCameraYawAndRoll(void)
{
  fCameraYawAndRollCoupled = false;
}

void TOpenGLApp::coupleCameraYawAndRoll(void)
{
  fCameraYawAndRollCoupled = true;
}

// --------------------------------------------------
// OpenGL color, fog, lighting and texture management
// --------------------------------------------------
void TOpenGLApp::clearScreen(void) const
{
  if (fEnableDepthBuffer) {
    glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);
  }
  else {
    glClear(GL_COLOR_BUFFER_BIT);
  }
}

void TOpenGLApp::setBackgroundColor(double red, double green, double blue, double alpha) const
{
  glClearColor((GLclampf) red,(GLclampf) green,(GLclampf) blue,(GLclampf) alpha);
}

void TOpenGLApp::setBackgroundColor(const Color& color) const
{
  setBackgroundColor(color.fRed,color.fGreen,color.fBlue,1.0);
}

void TOpenGLApp::setColor(const Color& color) const
{
  glColor3d((GLdouble) color.fRed,(GLdouble) color.fGreen,(GLdouble) color.fBlue);
}

void TOpenGLApp::enableSmoothShading(void) const
{
  glShadeModel(GL_SMOOTH);
}

void TOpenGLApp::enableFlatShading(void) const
{
  glShadeModel(GL_FLAT);
}

void TOpenGLApp::setMaterial(const Color& ambientColor,
                             const Color& diffuseColor,
                             const Color& specularColor,
                             double shininessCoefficient) const
{
  GLfloat ambientComponents[]   = {(GLfloat) ambientColor.fRed,
                                   (GLfloat) ambientColor.fGreen,
                                   (GLfloat) ambientColor.fBlue,
                                   (GLfloat) 1.0};
  GLfloat diffuseComponents[]   = {(GLfloat) diffuseColor.fRed,
                                   (GLfloat) diffuseColor.fGreen,
                                   (GLfloat) diffuseColor.fBlue,
                                   (GLfloat) 1.0};
  GLfloat specularComponents[]  = {(GLfloat) specularColor.fRed,
                                   (GLfloat) specularColor.fGreen,
                                   (GLfloat) specularColor.fBlue,
                                   (GLfloat) 1.0};
  if (shininessCoefficient > 128.0) {
    shininessCoefficient = 128.0;
  }
  GLfloat shininess[] = {(GLfloat) shininessCoefficient};

  glMaterialfv(GL_FRONT,GL_AMBIENT,ambientComponents);
  glMaterialfv(GL_FRONT,GL_DIFFUSE,diffuseComponents);
  glMaterialfv(GL_FRONT,GL_SPECULAR,specularComponents);
  glMaterialfv(GL_FRONT,GL_SHININESS,shininess);

  dontGlow();
}

void TOpenGLApp::setMaterialToBlackPlastic(void) const
{
  setMaterial(Color(0.0000,0.0000,0.0000),
              Color(0.0001,0.0001,0.0001),
              Color(0.5000,0.5000,0.5000),
              32.0);
}

void TOpenGLApp::setMaterialToBrass(void) const
{
  setMaterial(Color(0.3294,0.2235,0.0275),
              Color(0.7804,0.5686,0.1137),
              Color(0.9922,0.9412,0.8078),
              27.9);
}

void TOpenGLApp::setMaterialToBronze(void) const
{
  setMaterial(Color(0.2125,0.1275,0.0540),
              Color(0.7140,0.4284,0.1814),
              Color(0.3935,0.2179,0.1667),
              25.6);
}

void TOpenGLApp::setMaterialToChrome(void) const
{
  setMaterial(Color(0.2500,0.2500,0.2500),
              Color(0.4000,0.4000,0.4000),
              Color(0.7746,0.7746,0.7746),
              76.8);
}

void TOpenGLApp::setMaterialToCopper(void) const
{
  setMaterial(Color(0.1913,0.0735,0.0225),
              Color(0.7038,0.2705,0.0828),
              Color(0.2568,0.1376,0.0860),
              12.8);
}

void TOpenGLApp::setMaterialToGold(void) const
{
  setMaterial(Color(0.2473,0.1995,0.0745),
              Color(0.7516,0.6065,0.2265),
              Color(0.6283,0.5556,0.3661),
              51.2);
}

void TOpenGLApp::setMaterialToPewter(void) const
{
  setMaterial(Color(0.1059,0.0588,0.1137),
              Color(0.4275,0.4706,0.5412),
              Color(0.3333,0.3333,0.5216),
              9.8);
}

void TOpenGLApp::setMaterialToSilver(void) const
{
  setMaterial(Color(0.1923,0.1923,0.1923),
              Color(0.5075,0.5075,0.5075),
              Color(0.5083,0.5083,0.5083),
              51.2);
}

void TOpenGLApp::setMaterialToPolishedSilver(void) const
{
  setMaterial(Color(0.2313,0.2313,0.2313),
              Color(0.2775,0.2775,0.2775),
              Color(0.7740,0.7740,0.7740),
              89.6);
}

void TOpenGLApp::setGlow(const Color& glowColor) const
{
  GLfloat glow[] = {(GLfloat) glowColor.fRed,
                    (GLfloat) glowColor.fGreen,
                    (GLfloat) glowColor.fBlue,
                    (GLfloat) 1.0};
  glMaterialfv(GL_FRONT,GL_EMISSION,glow);
}

void TOpenGLApp::dontGlow(void) const
{
  setGlow(kBlack);
}

void TOpenGLApp::setHardColor(const Color& color) const
{
  setColor(color);
  setMaterial(kBlack,kBlack,kBlack,0.0);
  setGlow(color);
}

void TOpenGLApp::setGlobalAmbientLight(const Color& ambientColor) const
{
  GLfloat ambient[] = {(GLfloat) ambientColor.fRed,
                       (GLfloat) ambientColor.fGreen,
                       (GLfloat) ambientColor.fBlue,
                       (GLfloat) 1.0};
  glLightModelfv(GL_LIGHT_MODEL_AMBIENT,ambient);
}

void TOpenGLApp::enableExplicitViewpointCalculation(void) const
{
  glLightModeli(GL_LIGHT_MODEL_LOCAL_VIEWER,GL_TRUE);
}

void TOpenGLApp::disableExplicitViewpointCalculation(void) const
{
  glLightModeli(GL_LIGHT_MODEL_LOCAL_VIEWER,GL_FALSE);
}

void TOpenGLApp::enableLinearFog(const Color& fogColor, double start, double end) const
{
  GLfloat fogColor4[] = {(GLfloat) fogColor.fRed,
                         (GLfloat) fogColor.fGreen,
                         (GLfloat) fogColor.fBlue,
                         (GLfloat) 1.0};
  glFogfv(GL_FOG_COLOR,fogColor4);
  glHint(GL_FOG_HINT,GL_DONT_CARE);

  glFogi(GL_FOG_MODE,GL_LINEAR);
  glFogf(GL_FOG_START,(GLfloat) start);
  glFogf(GL_FOG_END,(GLfloat) end);

  glEnable(GL_FOG);
}

void TOpenGLApp::enableExponentialFog(const Color& fogColor, double density) const
{
  GLfloat fogColor4[] = {(GLfloat) fogColor.fRed,
                         (GLfloat) fogColor.fGreen,
                         (GLfloat) fogColor.fBlue,
                         (GLfloat) 1.0};
  glFogfv(GL_FOG_COLOR,fogColor4);
  glHint(GL_FOG_HINT,GL_DONT_CARE);

  glFogi(GL_FOG_MODE,GL_EXP2);
  glFogf(GL_FOG_DENSITY,(GLfloat) density);

  glEnable(GL_FOG);
}

void TOpenGLApp::disableFog(void) const
{
  glDisable(GL_FOG);
}

void TOpenGLApp::switchLightingOn(void) const
{
  glEnable(GL_LIGHTING);
}

void TOpenGLApp::switchLightingOff(void) const
{
  glDisable(GL_LIGHTING);
}

void TOpenGLApp::switchLightOn(int light) const
{
  glEnable((GLenum) light);
}

void TOpenGLApp::switchLightOff(int light) const
{
  glDisable((GLenum) light);
}

void TOpenGLApp::setLightPosition(int light, const Point3D& position) const
{
  GLfloat lightPosition[] = {(GLfloat) position.fX,
                             (GLfloat) position.fY,
                             (GLfloat) position.fZ,
                             (GLfloat) 1.0};
  glLightfv(GL_LIGHT0,GL_POSITION,lightPosition);
}

void TOpenGLApp::setLightDirection(int light, const Vector3D& direction) const
{
  GLfloat lightDirection[] = {(GLfloat) direction.fX,
                              (GLfloat) direction.fY,
                              (GLfloat) direction.fZ,
                              (GLfloat) 0.0};
  glLightfv(GL_LIGHT0,GL_POSITION,lightDirection);
}

void TOpenGLApp::setLightProperties(int light,
                                    const Color& ambient,
                                    const Color& diffuse,
                                    const Color& specular) const
{
  GLfloat lightAmbient[]  = {(GLfloat) ambient.fRed,
                             (GLfloat) ambient.fGreen,
                             (GLfloat) ambient.fBlue,
                             (GLfloat) 1.0};
  GLfloat lightDiffuse[]  = {(GLfloat) diffuse.fRed,
                             (GLfloat) diffuse.fGreen,
                             (GLfloat) diffuse.fBlue,
                             (GLfloat) 1.0};
  GLfloat lightSpecular[] = {(GLfloat) specular.fRed,
                             (GLfloat) specular.fGreen,
                             (GLfloat) specular.fBlue,
                             (GLfloat) 1.0};

  glLightfv((GLenum) light,GL_AMBIENT,lightAmbient);
  glLightfv((GLenum) light,GL_DIFFUSE,lightDiffuse);
  glLightfv((GLenum) light,GL_SPECULAR,lightSpecular);
}

void TOpenGLApp::makeSpotLight(int light,
                               const Vector3D& directionOfLightCone,
                               double cutOffAngle,
                               double exponent) const
{
  GLfloat spotLightDirection[] = {(GLfloat) directionOfLightCone.fX,
                                  (GLfloat) directionOfLightCone.fY,
                                  (GLfloat) directionOfLightCone.fZ};
  glLightfv((GLenum) light,GL_SPOT_DIRECTION,spotLightDirection);
  glLightf((GLenum) light,GL_SPOT_CUTOFF,(GLfloat) cutOffAngle);
  glLightf((GLenum) light,GL_SPOT_EXPONENT,(GLfloat) exponent);
}

void TOpenGLApp::enableLightAttenuation(int light,
                                        double constantAttenuation,
                                        double linearAttenuation,
                                        double quadraticAttenuation) const
{
  glLightf((GLenum) light,GL_CONSTANT_ATTENUATION,(GLfloat) constantAttenuation);
  glLightf((GLenum) light,GL_LINEAR_ATTENUATION,(GLfloat) linearAttenuation);
  glLightf((GLenum) light,GL_QUADRATIC_ATTENUATION,(GLfloat) quadraticAttenuation);
}

void TOpenGLApp::disableLightAttenuation(int light) const
{
  enableLightAttenuation(light,1.0,0.0,0.0);
}

int TOpenGLApp::getUniqueTextureID(void) const
{
  GLuint ID[1];
  glGenTextures(1,ID);
  return ID[0];
}

void TOpenGLApp::enableGoodTextureInterpolation(void)
{
  fGoodTextureInterpolation = true;
}

void TOpenGLApp::disableGoodTextureInterpolation(void)
{
  fGoodTextureInterpolation = false;
}

void TOpenGLApp::enableGlowingTextures(void)
{
  fGlowingTextures = true;
}

void TOpenGLApp::disableGlowingTextures(void)
{
  fGlowingTextures = false;
}

// --------------------
// 2D-graphics routines
// --------------------
void TOpenGLApp::point(double x, double y) const
{
  glBegin(GL_POINTS);
    glVertex2d((GLdouble) x,(GLdouble) y);
  glEnd();
}

void TOpenGLApp::point(const Point2D& p) const
{
  point(p.fX,p.fY);
}

void TOpenGLApp::setPointSize(double pointSize) const
{
  glPointSize((GLfloat) pointSize);
}

void TOpenGLApp::line(double x1, double y1, double x2, double y2) const
{
  glBegin(GL_LINES);
    glVertex2d((GLdouble) x1,(GLdouble) y1);
    glVertex2d((GLdouble) x2,(GLdouble) y2);
  glEnd();
}

void TOpenGLApp::line(const Point2D& p, const Point2D& q) const
{
  line(p.fX,p.fY,q.fX,q.fY);
}

void TOpenGLApp::lineTo(double x, double y)
{
  glBegin(GL_LINES);
    glVertex2d((GLdouble) fCurrentPosition.fX,(GLdouble) fCurrentPosition.fY);
    glVertex2d((GLdouble) x,(GLdouble) y);
  glEnd();
  moveTo(x,y);
}

void TOpenGLApp::lineTo(const Point2D& p)
{
  lineTo(p.fX,p.fY);
}

void TOpenGLApp::moveTo(double x, double y)
{
  fCurrentPosition.fX = x;
  fCurrentPosition.fY = y;
}

void TOpenGLApp::moveTo(const Point2D& p)
{
  moveTo(p.fX,p.fY);
}

void TOpenGLApp::lineRel(double dx, double dy)
{
  lineTo(fCurrentPosition.fX + dx,
         fCurrentPosition.fY + dy);
}

void TOpenGLApp::moveRel(double dx, double dy)
{
  fCurrentPosition.fX += dx;
  fCurrentPosition.fY += dy;
}

void TOpenGLApp::forward(double distance, bool isVisible)
{
  double x = fCurrentPosition.fX + (distance * cos(Math::degToRad(fCurrentDirection)));
  double y = fCurrentPosition.fY + (distance * sin(Math::degToRad(fCurrentDirection)));

  if (isVisible) {
    lineTo(x,y);
  }
  else {
    moveTo(x,y);
  }
}

void TOpenGLApp::turnTo(double degrees)
{
  fCurrentDirection = degrees;
}

void TOpenGLApp::turn(double degrees)
{
  fCurrentDirection += degrees;
}

void TOpenGLApp::ngon(int n, double x, double y, double radius, double startAngleInDegrees)
{
  if (n < 3) {
    return;
  }

  double angle = Math::degToRad(startAngleInDegrees);
  double angleInc = (2.0 * Math::kPi) / n;

  moveTo(x + (radius * cos(angle)),
         y + (radius * sin(angle)));

  for (int k = 0; k < n; k++) {
    angle += angleInc;
    lineTo(x + (radius * cos(angle)),
           y + (radius * sin(angle)));
  }
}

void TOpenGLApp::ngon(int n, const Point2D& center, double radius, double startAngleInDegrees)
{
  ngon(n,center.fX,center.fY,radius,startAngleInDegrees);
}

void TOpenGLApp::circle(double x, double y, double radius)
{
  ngon(kPolygonCircleApproximation,x,y,radius,0.0);
}

void TOpenGLApp::circle(const Point2D& center, double radius)
{
  circle(center.fX,center.fY,radius);
}

void TOpenGLApp::polyLineFromFile(string filename) const
{
  fstream inputFileStream;

  inputFileStream.open(filename.c_str(),ios::in);
  if (inputFileStream.fail()) {
    return;
  }

  GLint nrOfPolygons;
  inputFileStream >> nrOfPolygons;

  for (int i = 0; i < nrOfPolygons; ++i) {

    GLint nrOfLines;
    inputFileStream >> nrOfLines;

    glBegin(GL_LINE_STRIP);
      for (int j = 0; j < nrOfLines; ++j) {
        double x, y;
        inputFileStream >> x >> y;
        glVertex2d((GLdouble) x,(GLdouble) y);
      }
    glEnd();
  }

  inputFileStream.close();
}

void TOpenGLApp::drawString(int x, int y, string str, const Color& color)
{
  if (fEnableDepthBuffer) {
    glDisable(GL_DEPTH_TEST);
  }

  glMatrixMode(GL_PROJECTION);
  glPushMatrix();  
  setWorldWindow(0.0,(double) fClientWindowYSize,(double) fClientWindowXSize,0.0);

  glMatrixMode(GL_MODELVIEW);
  glPushMatrix();  

  // invert the y-axis and translate the origin to the upper-left corner of the client-window
  glLoadIdentity();
  glScaled((GLdouble) 1.0,(GLdouble) -1.0,(GLdouble) 1.0);
  glTranslated((GLdouble) 0.0,(GLdouble) -fClientWindowYSize,(GLdouble) 0.0);

  setHardColor(color);

  glRasterPos2i(x,y);
  for (int i = 0; i < str.length(); ++i) {
    glutBitmapCharacter(GLUT_BITMAP_9_BY_15,str[i]);
  }

  glMatrixMode(GL_PROJECTION);
  glPopMatrix();
  glMatrixMode(GL_MODELVIEW);
  glPopMatrix();

  if (fEnableDepthBuffer) {
    glEnable(GL_DEPTH_TEST);
  }
}

// --------------------
// 3D-graphics routines
// --------------------
void TOpenGLApp::point(double x, double y, double z) const
{
  glBegin(GL_POINTS);
    glVertex3d((GLdouble) x,(GLdouble) y,(GLdouble) z);
  glEnd();
}

void TOpenGLApp::point(const Point3D& p) const
{
  point(p.fX,p.fY,p.fZ);
}

void TOpenGLApp::line(double x1, double y1, double z1,
                      double x2, double y2, double z2) const
{
  glBegin(GL_LINES);
    glVertex3d((GLdouble) x1,(GLdouble) y1,(GLdouble) z1);
    glVertex3d((GLdouble) x2,(GLdouble) y2,(GLdouble) z2);
  glEnd();
}

void TOpenGLApp::line(const Point3D& p, const Point3D& q) const
{
  line(p.fX,p.fY,p.fZ,
       q.fX,q.fY,q.fZ);
}

void TOpenGLApp::lineTo(double x, double y, double z)
{
  glBegin(GL_LINES);
    glVertex3d((GLdouble) fCurrentPosition.fX,
               (GLdouble) fCurrentPosition.fY,
               (GLdouble) fCurrentPosition.fZ);
    glVertex3d((GLdouble) x,(GLdouble) y,(GLdouble) z);
  glEnd();
  moveTo(x,y,z);
}

void TOpenGLApp::lineTo(const Point3D& p)
{
  lineTo(p.fX,p.fY,p.fZ);
}

void TOpenGLApp::moveTo(double x, double y, double z)
{
  fCurrentPosition.set(x,y,z);
}

void TOpenGLApp::moveTo(const Point3D& p)
{
  moveTo(p.fX,p.fY,p.fZ);
}

void TOpenGLApp::lineRel(double dx, double dy, double dz)
{
  lineTo(fCurrentPosition.fX + dx,
         fCurrentPosition.fY + dy,
         fCurrentPosition.fZ + dz);
}

void TOpenGLApp::moveRel(double dx, double dy, double dz)
{
  fCurrentPosition.fX += dx;
  fCurrentPosition.fY += dy;
  fCurrentPosition.fZ += dz;
}

void TOpenGLApp::cone() const
{
  glutSolidCone((GLdouble) 1.0,
                (GLdouble) 1.0,
                (GLint) kTaperedCylinderSlices,
                (GLint) kTaperedCylinderStacks);
}

void TOpenGLApp::cube() const
{
  glutSolidCube((GLdouble) 2.0);
}

void TOpenGLApp::cylinder() const
{
  taperedCylinder(1.0);
}

void TOpenGLApp::dodecahedron() const
{
  glutSolidDodecahedron();
}

void TOpenGLApp::icosahedron() const
{
  glutSolidIcosahedron();
}

void TOpenGLApp::octahedron() const
{
  glutSolidOctahedron();
}

void TOpenGLApp::sphere() const
{
  glutSolidSphere((GLdouble) 1.0,
                  (GLint) kSphereSlices,
                  (GLint) kSphereStacks);
}

void TOpenGLApp::taperedCylinder(double topRadius) const
{
  GLUquadricObj* quadricObject = gluNewQuadric();
  gluQuadricDrawStyle(quadricObject,(GLenum) GLU_FILL);
  gluCylinder(quadricObject,
              (GLdouble) 1.0,
              (GLdouble) topRadius,
              (GLdouble) 1.0,
              (GLint) kTaperedCylinderSlices,
              (GLint) kTaperedCylinderStacks);
}

void TOpenGLApp::teapot() const
{
  glutSolidTeapot((GLdouble) 1.0);
}

void TOpenGLApp::tetrahedron() const
{
  glutSolidTetrahedron();
}

void TOpenGLApp::torus(double tubeRadius, double torusRadius) const
{
  glutSolidTorus((GLdouble) tubeRadius,
                 (GLdouble) torusRadius,
                 (GLint) kTorusSlices,
                 (GLint) kTorusStacks);
}

// ----------------------------------
// miscellaneous 3D-graphics routines
// ----------------------------------
void TOpenGLApp::textureQuad(const RGBApixmap* const pixmap,
                             const Point2D& tUpperLeft,
                             const Point2D& tLowerLeft,
                             const Point2D& tLowerRight,
                             const Point2D& tUpperRight,
                             const Point3D& pUpperLeft,
                             const Point3D& pLowerLeft,
                             const Point3D& pLowerRight,
                             const Point3D& pUpperRight) const
{
  glEnable(GL_TEXTURE_2D);

  glBindTexture(GL_TEXTURE_2D,(GLuint) pixmap->getTextureID());

  // make the texture repeat itself in the S- and T-direction
  glTexParameteri(GL_TEXTURE_2D,GL_TEXTURE_WRAP_S,GL_REPEAT);
  glTexParameteri(GL_TEXTURE_2D,GL_TEXTURE_WRAP_T,GL_REPEAT);

  if (fGlowingTextures) {
    glTexEnvf(GL_TEXTURE_ENV,GL_TEXTURE_ENV_MODE,GL_DECAL);
  }
  else {
    glTexEnvf(GL_TEXTURE_ENV,GL_TEXTURE_ENV_MODE,GL_MODULATE);
  }

  if (fGoodTextureInterpolation) {  
    glTexParameteri(GL_TEXTURE_2D,GL_TEXTURE_MAG_FILTER,GL_LINEAR);
    glTexParameteri(GL_TEXTURE_2D,GL_TEXTURE_MIN_FILTER,GL_LINEAR);
  }
  else {
    glTexParameteri(GL_TEXTURE_2D,GL_TEXTURE_MAG_FILTER,GL_NEAREST);
    glTexParameteri(GL_TEXTURE_2D,GL_TEXTURE_MIN_FILTER,GL_NEAREST);
  }

  glTexImage2D(GL_TEXTURE_2D,(GLint) 0,GL_RGBA,
               (GLsizei) pixmap->getNrOfColumns(),(GLsizei) pixmap->getNrOfRows(),
               (GLint) 0,GL_RGBA,GL_UNSIGNED_BYTE,pixmap->getPixmap());

  Vector3D normalVector = Vector3D::calcNewellNormal4(pUpperLeft,pLowerLeft,
                                                      pLowerRight,pUpperRight);

  glBegin(GL_QUADS);
    // set default normal for every-vertex in the plane
    glNormal3d((GLdouble) normalVector.fX,(GLdouble) normalVector.fY,(GLdouble) normalVector.fZ);

    glTexCoord2d((GLdouble) tUpperLeft.fX,(GLdouble) tUpperLeft.fY);
    glVertex3d((GLdouble) pUpperLeft.fX,(GLdouble) pUpperLeft.fY,(GLdouble) pUpperLeft.fZ);

    glTexCoord2d((GLdouble) tLowerLeft.fX,(GLdouble) tLowerLeft.fY);
    glVertex3d((GLdouble) pLowerLeft.fX,(GLdouble) pLowerLeft.fY,(GLdouble) pLowerLeft.fZ);

    glTexCoord2d((GLdouble) tLowerRight.fX,(GLdouble) tLowerRight.fY);
    glVertex3d((GLdouble) pLowerRight.fX,(GLdouble) pLowerRight.fY,(GLdouble) pLowerRight.fZ);

    glTexCoord2d((GLdouble) tUpperRight.fX,(GLdouble) tUpperRight.fY);
    glVertex3d((GLdouble) pUpperRight.fX,(GLdouble) pUpperRight.fY,(GLdouble) pUpperRight.fZ);
  glEnd();

  glDisable(GL_TEXTURE_2D);
}

void TOpenGLApp::drawLandscape(const TerrainMap& terrainMap,
                               bool useLandscapeMaterials,
                               bool smoothenVertexNormals,
                               bool applyTextureMapping) const
{
  int uSize = terrainMap.getUSize();
  int vSize = terrainMap.getVSize();

  Rectangle2D bounds = terrainMap.getBounds();
  Point2D p1(bounds.fLeft,bounds.fTop);
  Point2D p2(bounds.fRight,bounds.fBottom);

  for (int v = 0; v < (vSize - 1); ++v) {
    for (int u = 0; u < (uSize - 1); ++u) {

      double u1 = p1.fX + ((p2.fX - p1.fX) * (((double) u) / (uSize - 1)));
      double v1 = p1.fY + ((p2.fY - p1.fY) * (((double) v) / (vSize - 1)));

      double u2 = p1.fX + ((p2.fX - p1.fX) * (((double) (u + 1)) / (uSize - 1)));
      double v2 = p1.fY + ((p2.fY - p1.fY) * (((double) (v + 1)) / (vSize - 1)));

      if (!applyTextureMapping) {
        if (smoothenVertexNormals) {
          glBegin(GL_TRIANGLES);

            // draw the upper-left triangle
            glNormal3d((GLdouble) terrainMap.getVertexNormal(u,v).fX,
                       (GLdouble) terrainMap.getVertexNormal(u,v).fY,
                       (GLdouble) terrainMap.getVertexNormal(u,v).fZ);
            setColor(terrainMap.getColor(u,v));
            if (useLandscapeMaterials) {
              setMaterial(terrainMap.getColor(u,v),
                          terrainMap.getColor(u,v),
                          kWhite,128.0);
            }
            glVertex3d((GLdouble) u1,
                       (GLdouble) terrainMap.getHeight(u,v),
                       (GLdouble) v1);

            glNormal3d((GLdouble) terrainMap.getVertexNormal(u,v + 1).fX,
                       (GLdouble) terrainMap.getVertexNormal(u,v + 1).fY,
                       (GLdouble) terrainMap.getVertexNormal(u,v + 1).fZ);
            setColor(terrainMap.getColor(u,v + 1));
            if (useLandscapeMaterials) {
              setMaterial(terrainMap.getColor(u,v + 1),
                          terrainMap.getColor(u,v + 1),
                          kWhite,128.0);
            }
            glVertex3d((GLdouble) u1,
                       (GLdouble) terrainMap.getHeight(u,v + 1),
                       (GLdouble) v2);

            glNormal3d((GLdouble) terrainMap.getVertexNormal(u + 1,v).fX,
                       (GLdouble) terrainMap.getVertexNormal(u + 1,v).fY,
                       (GLdouble) terrainMap.getVertexNormal(u + 1,v).fZ);
            setColor(terrainMap.getColor(u + 1,v));
            if (useLandscapeMaterials) {
              setMaterial(terrainMap.getColor(u + 1,v),
                          terrainMap.getColor(u + 1,v),
                          kWhite,128.0);
            }
            glVertex3d((GLdouble) u2,
                       (GLdouble) terrainMap.getHeight(u + 1,v),
                       (GLdouble) v1);

            // draw the lower-right triangle
            glNormal3d((GLdouble) terrainMap.getVertexNormal(u + 1,v).fX,
                       (GLdouble) terrainMap.getVertexNormal(u + 1,v).fY,
                       (GLdouble) terrainMap.getVertexNormal(u + 1,v).fZ);
            setColor(terrainMap.getColor(u + 1,v));
            if (useLandscapeMaterials) {
              setMaterial(terrainMap.getColor(u + 1,v),
                          terrainMap.getColor(u + 1,v),
                          kWhite,128.0);
            }
            glVertex3d((GLdouble) u2,
                       (GLdouble) terrainMap.getHeight(u + 1,v),
                       (GLdouble) v1);

            glNormal3d((GLdouble) terrainMap.getVertexNormal(u,v + 1).fX,
                       (GLdouble) terrainMap.getVertexNormal(u,v + 1).fY,
                       (GLdouble) terrainMap.getVertexNormal(u,v + 1).fZ);
            setColor(terrainMap.getColor(u,v + 1));
            if (useLandscapeMaterials) {
              setMaterial(terrainMap.getColor(u,v + 1),
                          terrainMap.getColor(u,v + 1),
                          kWhite,128.0);
            }
            glVertex3d((GLdouble) u1,
                       (GLdouble) terrainMap.getHeight(u,v + 1),
                       (GLdouble) v2);

            glNormal3d((GLdouble) terrainMap.getVertexNormal(u + 1,v + 1).fX,
                       (GLdouble) terrainMap.getVertexNormal(u + 1,v + 1).fY,
                       (GLdouble) terrainMap.getVertexNormal(u + 1,v + 1).fZ);
            setColor(terrainMap.getColor(u + 1,v + 1));
            if (useLandscapeMaterials) {
              setMaterial(terrainMap.getColor(u + 1,v + 1),
                          terrainMap.getColor(u + 1,v + 1),
                          kWhite,128.0);
            }
            glVertex3d((GLdouble) u2,
                       (GLdouble) terrainMap.getHeight(u + 1,v + 1),
                       (GLdouble) v2);

          glEnd();
        }
        else {
          glBegin(GL_TRIANGLES);

            // draw the upper-left triangle
            Vector3D normal1 = Vector3D::calcNewellNormal3(Point3D(u1,terrainMap.getHeight(u,v),v1),
                                                           Point3D(u1,terrainMap.getHeight(u,v + 1),v2),
                                                           Point3D(u2,terrainMap.getHeight(u + 1,v),v1));
            glNormal3d((GLdouble) normal1.fX,
                       (GLdouble) normal1.fY,
                       (GLdouble) normal1.fZ);

            setColor(terrainMap.getColor(u,v));
            if (useLandscapeMaterials) {
              setMaterial(terrainMap.getColor(u,v),
                          terrainMap.getColor(u,v),
                          kWhite,128.0);
            }
            glVertex3d((GLdouble) u1,
                       (GLdouble) terrainMap.getHeight(u,v),
                       (GLdouble) v1);

            setColor(terrainMap.getColor(u,v + 1));
            if (useLandscapeMaterials) {
              setMaterial(terrainMap.getColor(u,v + 1),
                          terrainMap.getColor(u,v + 1),
                          kWhite,128.0);
            }
            glVertex3d((GLdouble) u1,
                       (GLdouble) terrainMap.getHeight(u,v + 1),
                       (GLdouble) v2);

            setColor(terrainMap.getColor(u + 1,v));
            if (useLandscapeMaterials) {
              setMaterial(terrainMap.getColor(u + 1,v),
                          terrainMap.getColor(u + 1,v),
                          kWhite,128.0);
            }
            glVertex3d((GLdouble) u2,
                       (GLdouble) terrainMap.getHeight(u + 1,v),
                       (GLdouble) v1);

            // draw the lower-right triangle
            Vector3D normal2 = Vector3D::calcNewellNormal3(Point3D(u2,terrainMap.getHeight(u + 1,v),v1),
                                                           Point3D(u1,terrainMap.getHeight(u,v + 1),v2),
                                                           Point3D(u2,terrainMap.getHeight(u + 1,v + 1),v2));
            glNormal3d((GLdouble) normal2.fX,
                       (GLdouble) normal2.fY,
                       (GLdouble) normal2.fZ);

            setColor(terrainMap.getColor(u + 1,v));
            if (useLandscapeMaterials) {
              setMaterial(terrainMap.getColor(u + 1,v),
                          terrainMap.getColor(u + 1,v),
                          kWhite,128.0);
            }
            glVertex3d((GLdouble) u2,
                       (GLdouble) terrainMap.getHeight(u + 1,v),
                       (GLdouble) v1);

            setColor(terrainMap.getColor(u,v + 1));
            if (useLandscapeMaterials) {
              setMaterial(terrainMap.getColor(u,v + 1),
                          terrainMap.getColor(u,v + 1),
                          kWhite,128.0);
            }
            glVertex3d((GLdouble) u1,
                       (GLdouble) terrainMap.getHeight(u,v + 1),
                       (GLdouble) v2);

            setColor(terrainMap.getColor(u + 1,v + 1));
            if (useLandscapeMaterials) {
              setMaterial(terrainMap.getColor(u + 1,v + 1),
                          terrainMap.getColor(u + 1,v + 1),
                          kWhite,128.0);
            }
            glVertex3d((GLdouble) u2,
                       (GLdouble) terrainMap.getHeight(u + 1,v + 1),
                       (GLdouble) v2);

          glEnd();
        }
      }
      else {
        textureQuad(terrainMap.getTexture(u,v),
                    Point2D(0.0,0.0),
                    Point2D(0.0,1.0),
                    Point2D(1.0,1.0),
                    Point2D(1.0,0.0),
                    Point3D(u1,terrainMap.getHeight(u,v),v1),
                    Point3D(u1,terrainMap.getHeight(u,v + 1),v2),
                    Point3D(u2,terrainMap.getHeight(u + 1,v + 1),v2),
                    Point3D(u2,terrainMap.getHeight(u + 1,v),v1));
      }
    }
  }
}

void TOpenGLApp::showAxes(double length, double thickness) const
{
  // show the (red) X-axis
  setMaterial(kRed,kRed,kBlack,0.0);

  // show the shaft
  pushCT();
    translate3D(-length / 6.0,0.0,0.0);
    rotate3D(90.0,0.0,1.0,0.0);
    scale3D(thickness,thickness,length);
    cylinder();
  popCT();

  // show the arrow
  pushCT();
    translate3D((5.0 * length) / 6.0,0.0,0.0);
    rotate3D(90.0,0.0,1.0,0.0);
    scale3D(2.0 * thickness,2.0 * thickness,length / 3.0);
    cone();
  popCT();

  // show the (green) Y-axis
  setMaterial(kGreen,kGreen,kBlack,0.0);

  // show the shaft
  pushCT();
    translate3D(0.0,-length / 6.0,0.0);
    rotate3D(-90.0,1.0,0.0,0.0);
    scale3D(thickness,thickness,length);
    cylinder();
  popCT();

  // show the arrow
  pushCT();
    translate3D(0.0,(5.0 * length) / 6.0,0.0);
    rotate3D(-90.0,1.0,0.0,0.0);
    scale3D(2.0 * thickness,2.0 * thickness,length / 3.0);
    cone();
  popCT();

  // show the (blue) Z-axis
  setMaterial(kBlue,kBlue,kBlack,0.0);

  // show the shaft
  pushCT();
    translate3D(0.0,0.0,-length / 6.0);
    scale3D(thickness,thickness,length);
    cylinder();
  popCT();

  // show the arrow
  pushCT();
    translate3D(0.0,0.0,(5.0 * length) / 6.0);
    scale3D(2.0 * thickness,2.0 * thickness,length / 3.0);
    cone();
  popCT();
}

// ----------------------------------------
// scene-rendering (OpenGL and ray tracing)
// ----------------------------------------
void TOpenGLApp::loadScene(string filename)
{
  fScene.readFromFile(filename);
}

void TOpenGLApp::clearScene(void)
{
  fScene.clear();
}

void TOpenGLApp::setSceneViewSettings(void)
{
  if (fScene.fEyePointSpecified) {
    setCamera(fScene.fEyePoint,
              getCameraViewPoint(),
              getCameraUpDirection());
  }

  if (fScene.fViewPointSpecified) {
    setCamera(getCameraEyePoint(),
              fScene.fViewPoint,
              getCameraUpDirection());
  }

  if (fScene.fViewAngleSpecified) {
    setCameraPerspectiveShape(fScene.fViewAngle,
                              getCameraAspectRatio(),
                              getCameraDistanceOfNearPlane(),
                              getCameraDistanceOfFarPlane());
  }

  if (fScene.fUpDirectionSpecified) {
    setCamera(getCameraEyePoint(),
              getCameraViewPoint(),
              fScene.fUpDirection);
  }
}

void TOpenGLApp::enableSceneLighting(void) const
{
  fScene.makeLightsOpenGL();
}

void TOpenGLApp::drawScene(void)
{
  fScene.drawOpenGL(fEnableDepthBuffer);
}

void TOpenGLApp::rayTraceScene(int blockSize,
                               int nrOfConeSamples,
                               EShadowComputationMode shadowComputationMode,
                               double shadowFeelerJitterStrength,
                               EAntiAliasMode antiAliasMode,
                               bool useStochasticSampling,
                               double jitterStrength,
                               bool normalizeColors,
                               double normalizeBrightness,
                               bool findNormals,
                               double findNormalsProbability,
                               bool followRays,
                               double followRaysProbability)
{
  // anti-aliasing must be disabled (we do it ourselves)
  if (fAntiAliasingEnabled) {
    cerr << "TOpenGLApp::rayTraceScene() -> OpenGL's own anti-aliasing must be disabled." << endl;
    return;
  }

  if (nrOfConeSamples < 1) {
    nrOfConeSamples = 1;
  }

  // backup the current view-settings
  glMatrixMode(GL_MODELVIEW);
  glPushMatrix();
  glLoadIdentity();

  glMatrixMode(GL_PROJECTION);
  glPushMatrix();
  setWorldWindow(0.0,(double) fClientWindowYSize,(double) fClientWindowXSize,0.0);

  // turn off all light-attenuation, fog and lighting because
  // we'll do it ourselves
  glDisable(GL_DEPTH_TEST);
  for (int lightNr = GL_LIGHT0; lightNr < 8; ++lightNr) {
    glLightf((GLenum) lightNr,GL_CONSTANT_ATTENUATION,(GLfloat) 1.0);
    glLightf((GLenum) lightNr,GL_LINEAR_ATTENUATION,(GLfloat) 0.0);
    glLightf((GLenum) lightNr,GL_QUADRATIC_ATTENUATION,(GLfloat) 0.0);
  }
  glDisable(GL_FOG);
  switchLightingOff();

  switch (shadowComputationMode) {

    case kNoShadows :
      fScene.dontComputeShadowsWhenRayTracing();
      break;

    case kHardShadows :
      fScene.computeShadowsWhenRayTracing();
      fScene.computeHardShadowsWhenRayTracing();
      fScene.setShadowFeelerJitterStrength(shadowFeelerJitterStrength);
      break;

    case kSoftShadows :
      fScene.computeShadowsWhenRayTracing();
      fScene.computeSoftShadowsWhenRayTracing();
      fScene.setShadowFeelerJitterStrength(shadowFeelerJitterStrength);
      break;
  }

  fScene.fLines.clear();

  if (findNormals) {
    fScene.findNormalsWhenRayTracing(findNormalsProbability);
  }
  else {
    fScene.dontFindNormalsWhenRayTracing();
  }

  if (followRays) {
    fScene.followRaysWhenRayTracing(followRaysProbability);
  }
  else {
    fScene.dontFollowRaysWhenRayTracing();
  }

  fScene.fEyePoint = fCameraEyePoint;

  fScene.resetStatistics();
  Chrono chrono;
  chrono.start();

  // create and initialize the render-buffer
  Color* renderBuffer = new Color[fClientWindowYSize * fClientWindowXSize];
  for (int i = 0; i < (fClientWindowXSize * fClientWindowYSize); ++i) {
    renderBuffer[i] = kBlack;
  }

  // determine the size of a pixel on the nearplane (in the scene's world)
  double pixelWidth = 2.0 * (fCameraRightPlane - fCameraLeftPlane) / (double) fClientWindowXSize;
  double pixelHeight = 2.0 * (fCameraTopPlane - fCameraBottomPlane) / (double) fClientWindowYSize;

  int row, column;

  // make a black border at the edge of the window
  setColor(kBlack);
  for (column = 0; column < fClientWindowXSize; ++column) {
    // draw the upper border
    glRecti((GLint) column,(GLint) (fClientWindowYSize - blockSize - 1),
            (GLint) (column + 1),(GLint) fClientWindowYSize);
    // draw the lower border
    glRecti((GLint) column,(GLint) 0,
            (GLint) (column + 1),(GLint) blockSize);
  }
  for (row = 0; row < fClientWindowYSize; ++row) {
    // draw the left border
    glRecti((GLint) 0,(GLint) row,
            (GLint) blockSize,(GLint) (row + 1));
    // draw the right border
    glRecti((GLint) (fClientWindowXSize - blockSize - 1),(GLint) row,
            (GLint) fClientWindowXSize,(GLint) (row + 1));
  }

  double timeIndexOffset = 0;
  double timeIndexSpan = 0;
  if (fScene.isMotionBlurEnabled()) {
    timeIndexOffset = -(fScene.getMotionBlurFrameLength() / 2.0);
    timeIndexSpan = fScene.getMotionBlurFrameLength() /
                    (fScene.getNrOfMotionBlurTimeSamples() - 1);
  }

  // start from top to bottom and from left to right (try to keep an offset of 1 block from the sides)
  for (row = (fClientWindowYSize - (2 * blockSize) - 1); row > 0; row -= blockSize) {
    for (column = blockSize; column < (fClientWindowXSize - blockSize); column += blockSize) {

      Color blockColor = kBlack;

      if (fScene.isMotionBlurEnabled()) {

        // sample all time-indices to compute motion blur
        for (int timeSample = 0; timeSample < fScene.getNrOfMotionBlurTimeSamples(); ++timeSample) {

          // find the timeIndex associated with this timeSample
          double timeIndex = timeIndexOffset + (timeSample * timeIndexSpan);

          // jitter this timeIndex to introduce stochastic temporal anti-aliasing
          timeIndex += (-1.0 + (2.0 * RNG::uniform())) * timeIndexOffset * jitterStrength;

          // transform objects with respect to the timeIndex to introduce motion blur
          fScene.applyShapesMotionBlurAffineTransformation(timeIndex);

          Color color = samplePixel(antiAliasMode,useStochasticSampling,
                                    row,column,
                                    nrOfConeSamples,blockSize,
                                    pixelWidth,pixelHeight,
                                    jitterStrength);

          // adjust this color's weight
          color /= fScene.getNrOfMotionBlurTimeSamples();

          blockColor += color;

          // restore all objects back to their original position, orientation and scale
          fScene.restoreShapes();

        } // timeSample
      }
      else {

        // motion blur is disabled
        blockColor = samplePixel(antiAliasMode,useStochasticSampling,
                                 row,column,
                                 nrOfConeSamples,blockSize,
                                 pixelWidth,pixelHeight,
                                 jitterStrength);
      }

      // show the colored block on the screen
      setColor(blockColor);
      glRecti((GLint) column,(GLint) row,
              (GLint) (column + blockSize),(GLint) (row + blockSize));

      // update the render-buffer with the current colored block
      for (int j = 0; j < blockSize; ++j) {
        for (int i = 0; i < blockSize; ++i) {
          int position = ((row + j) * fClientWindowXSize) + (column + i);
          if (position < (fClientWindowXSize * fClientWindowYSize)) {
            renderBuffer[position] = blockColor;
          }
        }
      }

    } // column
    // show the finished scanline
    finishFrame();
  } // row

  if (normalizeColors) {

    bool renderBufferOverflow = false;

    // check for overflow in the render-buffer
    Color highestColor = renderBuffer[0];
    for (int i = 0; i < (fClientWindowXSize * fClientWindowYSize); ++i) {
      if ((renderBuffer[i].fRed > 1.0) ||
          (renderBuffer[i].fGreen > 1.0) ||
          (renderBuffer[i].fBlue > 1.0)) {
        renderBufferOverflow = true;

        // find the highest color
        if ((renderBuffer[i].fRed > highestColor.fRed) ||
            (renderBuffer[i].fGreen > highestColor.fGreen) ||
            (renderBuffer[i].fBlue > highestColor.fBlue)) {
          highestColor = renderBuffer[i];
        }
      }
    }

    if (renderBufferOverflow) {

      // find the color-component that is out of range
      double scaleFactor = highestColor.fRed;
      if (highestColor.fGreen > scaleFactor) {
        scaleFactor = highestColor.fGreen;
      }
      if (highestColor.fBlue > scaleFactor) {
        scaleFactor = highestColor.fBlue;
      }

      // introduce the brightness factor (use wisely since this can again result in overflow !)
      scaleFactor /= normalizeBrightness;

      // normalize the render-buffer
      for (row = (fClientWindowYSize - 1); row >= 0; --row) {
        for (column = 0; column < fClientWindowXSize; ++column) {
          renderBuffer[(row * fClientWindowXSize) + column] /= scaleFactor;
        }
      }

      for (row = (fClientWindowYSize - 1); row >= 0; --row) {
        for (column = 0; column < fClientWindowXSize; ++column) {
          setColor(renderBuffer[(row * fClientWindowXSize) + column]);
          glRecti((GLint) column,(GLint) row,
                  (GLint) (column + 1),(GLint) (row + 1));
        }

        // show the finished scanline
        finishFrame();
      }
    }
  }

  delete[] renderBuffer;

  // restore the previous view-settings
  if (fEnableDepthBuffer) {
    glEnable(GL_DEPTH_TEST);
  }

  chrono.stop();
  long elapsedTime = chrono.getElapsedTime();
  cerr << "Render-time (" << Chrono::convertToHMS(elapsedTime)
       << "), #rays shot : " << fScene.getNrOfRaysShot() << endl;

  glMatrixMode(GL_PROJECTION);
  glPopMatrix();

  glMatrixMode(GL_MODELVIEW);
  glPopMatrix();

  // show the normals
  if (findNormals || followRays) {
    Scene::Lines::iterator p;
    for (p = fScene.fLines.begin(); p != fScene.fLines.end(); ++p) {
      setHardColor(p->fColor);
      line(p->fp1,p->fp2);
    }
  }
}

long TOpenGLApp::getNrOfRaysShot(void)
{
  return fScene.getNrOfRaysShot();
}

// --------------------------
// 2D- and 3D-transformations
// --------------------------
void TOpenGLApp::initCT(void) const
{
  glMatrixMode(GL_MODELVIEW);
  glLoadIdentity();
}

void TOpenGLApp::rotate2D(double degrees) const
{
  glMatrixMode(GL_MODELVIEW);
  glRotated((GLdouble) degrees,(GLdouble) 0.0,(GLdouble) 0.0,(GLdouble) 1.0);
}

void TOpenGLApp::rotate3D(double degrees, double x, double y, double z) const
{
  glMatrixMode(GL_MODELVIEW);
  glRotated((GLdouble) degrees,(GLdouble) x,(GLdouble) y,(GLdouble) z);
}

void TOpenGLApp::translate2D(double dx, double dy) const
{
  glMatrixMode(GL_MODELVIEW);
  glTranslated((GLdouble) dx,(GLdouble) dy,(GLdouble) 0.0);
}

void TOpenGLApp::translate2D(const Point2D& p) const
{
  translate2D(p.fX,p.fY);
}

void TOpenGLApp::translate3D(double dx, double dy, double dz) const
{
  glMatrixMode(GL_MODELVIEW);
  glTranslated((GLdouble) dx,(GLdouble) dy,(GLdouble) dz);
}

void TOpenGLApp::translate3D(const Point3D& p) const
{
  translate3D(p.fX,p.fY,p.fZ);
}

void TOpenGLApp::scale2D(double sx, double sy) const
{
  glMatrixMode(GL_MODELVIEW);
  glScaled((GLdouble) sx,(GLdouble) sy,(GLdouble) 1.0);
}

void TOpenGLApp::scale3D(double sx, double sy, double sz) const
{
  glMatrixMode(GL_MODELVIEW);
  glScaled((GLdouble) sx,(GLdouble) sy,(GLdouble) sz);
}

void TOpenGLApp::pushCT(void) const
{
  glMatrixMode(GL_MODELVIEW);
  glPushMatrix();
}

void TOpenGLApp::popCT(void) const
{
  glMatrixMode(GL_MODELVIEW);
  glPopMatrix();
}

// ---------------
// utility methods
// ---------------
string TOpenGLApp::convertDoubleToString(double x)
{
  ostringstream ostr;
  ostr << x;

  // terminate stream
  #ifndef _WIN32
    ostr << ends; // Microsoft Visual C++ 6.0 puts a strange symbol here
  #endif

  return ostr.str();
}

void TOpenGLApp::finishFrame(void) const
{
  if (!fEnableDoubleBuffering) {
    glFlush();
  }
  else {
    // glutSwapBuffers() automatically calls glFlush()
    glutSwapBuffers();
  }
}

void TOpenGLApp::setModelViewMatrix(void) const
{
  double m[16];

  Vector3D eyeVector(fCameraEyePoint.fX,fCameraEyePoint.fY,fCameraEyePoint.fZ);

  m[0] = fCameraU.fX; m[4] = fCameraU.fY; m[8]  = fCameraU.fZ; m[12] = -eyeVector.inner(fCameraU);
  m[1] = fCameraV.fX; m[5] = fCameraV.fY; m[9]  = fCameraV.fZ; m[13] = -eyeVector.inner(fCameraV);
  m[2] = fCameraN.fX; m[6] = fCameraN.fY; m[10] = fCameraN.fZ; m[14] = -eyeVector.inner(fCameraN);
  m[3] = 0.0;         m[7] = 0.0;         m[11] = 0.0;         m[15] = 1.0;

  glMatrixMode(GL_MODELVIEW);
  glLoadMatrixd(m);
}

void TOpenGLApp::rotateCameraAroundAxes(Vector3D& a, Vector3D& b, double degrees) const
{
  // rotate orthogonal vectors a (like x-axis) and b (like y-axis) through angle (expressed in degrees)
  double angle = Math::degToRad(degrees);
  double cosine = cos(angle);
  double sine = sin(angle);

  Vector3D t((cosine * a.fX) + (sine * b.fX),
             (cosine * a.fY) + (sine * b.fY),
             (cosine * a.fZ) + (sine * b.fZ));

  b.fX = (-sine * a.fX) + (cosine * b.fX);
  b.fY = (-sine * a.fY) + (cosine * b.fY);
  b.fZ = (-sine * a.fZ) + (cosine * b.fZ);

  a = t;
}

Color TOpenGLApp::samplePixel(EAntiAliasMode antiAliasMode,
                              bool useStochasticSampling,
                              int row,
                              int column,
                              int nrOfConeSamples,
                              int blockSize,
                              double pixelWidth,
                              double pixelHeight,
                              double jitterStrength)
{
  Color color = kBlack;

  if ((antiAliasMode == kUse5Samples) ||
      (antiAliasMode == kUse9Samples)) {

    // shoot nine rays and average their resulting colors (according to the alias-mode)
    for (int i = 0; i < 9; ++i) {

      fRays[i].clear();

      // all rays originate from our eye
      fRays[i].fStart = fCameraEyePoint;

      // build the ray
      fRays[i].fRow = row;
      fRays[i].fColumn = column;
      fRays[i].fNrOfConeSamples = nrOfConeSamples;
      fRays[i].fRecurseLevel = 1;

      // use the value in the middle of the block
      double x = -(fCameraRightPlane - fCameraLeftPlane) +
                  (column * pixelWidth) +
                 ((blockSize / 2.0) * pixelWidth);
      double y = -(fCameraTopPlane - fCameraBottomPlane) +
                  (row * pixelHeight) +
                  ((blockSize / 2.0) * pixelHeight);

      // reposition the ray to supersample nearby locations
      if ((i % 3) == 0) {
        x -= (blockSize / 2.0) * pixelWidth;
      }
      else if (((i + 1) % 3) == 0) {
        x += (blockSize / 2.0) * pixelWidth;
      }

      if (i <= 2) {
        y -= (blockSize / 2.0) * pixelHeight;
      }
      else if (i >= 6) {
        y += (blockSize / 2.0) * pixelHeight;
      }

      if (useStochasticSampling) {
        // jitter the supersample-locations just computed
        // to introduce a jittered regular distribution that
        // approximates a minimum distance Poisson distribution
        // (i.e. introduce some noise in the image)

        double xExcess = blockSize * pixelWidth * jitterStrength;
        double yExcess = blockSize * pixelHeight * jitterStrength;

        x = x - (xExcess / 2.0) + (RNG::uniform() * xExcess);
        y = y - (yExcess / 2.0) + (RNG::uniform() * yExcess);
      }

      fRays[i].fDirection = Vector3D(-(fCameraDistanceOfNearPlane * fCameraN.fX) +
                                      (x * fCameraU.fX) +
                                      (y * fCameraV.fX),
                                     -(fCameraDistanceOfNearPlane * fCameraN.fY) +
                                      (x * fCameraU.fY) +
                                      (y * fCameraV.fY),
                                     -(fCameraDistanceOfNearPlane * fCameraN.fZ) +
                                      (x * fCameraU.fZ) +
                                      (y * fCameraV.fZ));
      fRays[i].fDirection.normalize();

      Color rayColor(fScene.shootRay(fRays[i]));
      if (antiAliasMode == kUse5Samples) {
        rayColor /= 5.0;
      }
      else if (antiAliasMode == kUse9Samples) {
        rayColor /= 9.0;
      }
      color += rayColor;

      // skip a sample when not in full anti-alias mode (i.e. when less than 9 samples are used)
      if (antiAliasMode == kUse5Samples) {
        ++i;
      }
    }
  }
  else { // only use 1 sample (i.e. don't perform anti-aliasing)

    // the ray originates from our eye
    Ray ray(fCameraEyePoint);

    // build the ray
    ray.fRow = row;
    ray.fColumn = column;
    ray.fNrOfConeSamples = nrOfConeSamples;
    ray.fRecurseLevel = 1;

    // use the value in the middle of the block
    double x = -(fCameraRightPlane - fCameraLeftPlane) +
                (column * pixelWidth) +
                ((blockSize / 2.0) * pixelWidth);
    double y = -(fCameraTopPlane - fCameraBottomPlane) +
                (row * pixelHeight) +
                ((blockSize / 2.0) * pixelHeight);

    ray.fDirection = Vector3D(-(fCameraDistanceOfNearPlane * fCameraN.fX) +
                               (x * fCameraU.fX) +
                               (y * fCameraV.fX),
                              -(fCameraDistanceOfNearPlane * fCameraN.fY) +
                               (x * fCameraU.fY) +
                               (y * fCameraV.fY),
                              -(fCameraDistanceOfNearPlane * fCameraN.fZ) +
                               (x * fCameraU.fZ) +
                               (y * fCameraV.fZ));
    ray.fDirection.normalize();

    // shoot the ray into the scene
    color = fScene.shootRay(ray);
  }

  return color;
}
