// ------------------------------
// Filename      : primitives.h
// Author        : Sven Maerivoet
// Last modified : 20/04/2001
// Target        : C++ (OpenGL)
//
// All rights reserved.
// ------------------------------

#ifndef SM_PRIMITIVES_H // include-guard
#define SM_PRIMITIVES_H

#include <string>
#include <vector>

using namespace std;

class Point2D {
  public:
    // constructors
    Point2D(void);
    Point2D(double x, double y);
    Point2D(const Point2D& point2D);

    // destructor
    ~Point2D(void);

    // methods
    void set(double x, double y);

    // data-members
    double fX, fY;
};

class Point3D : public Point2D {
  public:
    // constructors
    Point3D(void);
    Point3D(double x, double y, double z);
    Point3D(const Point3D& point3D);

    // destructor
    ~Point3D(void);

    // methods
    void set(double x, double y, double z);
    void build4tuple(double v[]) const;
    void build4tuple(float v[]) const;

    // data-members
    double fZ;
};

class Rectangle2D {
  public:
    // constructors
    Rectangle2D(void);
    Rectangle2D(double left, double top, double right, double bottom);
    Rectangle2D(const Rectangle2D& rectangle2D);

    // destructor
    ~Rectangle2D(void);

    // methods
    void set(double left, double top, double right, double bottom);
    double getWidth(void) const;
    double getHeight(void) const;
    double getAspectRatio(void) const;

    // data-members
    double fLeft, fTop, fRight, fBottom;
};

class Vector2D {
  public:
    // constructors
    Vector2D(void);
    Vector2D(double x, double y);
    Vector2D(const Point2D& point2D);
    Vector2D(const Vector2D& vector2D);

    // destructor
    ~Vector2D(void);

    // vector arithmetic
    Vector2D& operator=(const Vector2D& vector2D);
    double length(void) const;
    void normalize(void);
    void flip(void);

    Vector2D operator+(const Vector2D& vector2D) const;
    Vector2D& operator+=(const Vector2D& vector2D);

    Vector2D operator-(const Vector2D& vector2D) const;
    Vector2D& operator-=(const Vector2D& vector2D);

    Vector2D operator*(double scalar) const;
    Vector2D& operator*=(double scalar);

    Vector2D operator/(double scalar) const;
    Vector2D& operator/=(double scalar);

    double inner(const Vector2D& vector2D) const;
    double angle(const Vector2D& vector2D) const;

    // methods
    void set(double x, double y);

    // data-members
    double fX, fY;
};

class Vector3D {
  public:
    // constructors
    Vector3D(void);
    Vector3D(double x, double y, double z);
    Vector3D(const Point3D& point3D);
    Vector3D(const Vector3D& vector3D);

    // destructor
    ~Vector3D(void);

    // vector arithmetic
    Vector3D& operator=(const Vector3D& vector3D);
    double length(void) const;
    void normalize(void);
    void flip(void);

    Vector3D operator+(const Vector3D& vector3D) const;
    Vector3D& operator+=(const Vector3D& vector3D);

    Vector3D operator-(const Vector3D& vector3D) const;
    Vector3D& operator-=(const Vector3D& vector3D);

    Vector3D operator*(double scalar) const;
    Vector3D& operator*=(double scalar);

    Vector3D operator/(double scalar) const;
    Vector3D& operator/=(double scalar);

    double inner(const Vector3D& vector3D) const;
    Vector3D outer(const Vector3D& vector3D) const;
    double angle(const Vector3D& vector3D) const;

    // methods
    void set(double x, double y, double z);
    static Vector3D calcNewellNormal3(const Point3D& p1,
                                      const Point3D& p2,
                                      const Point3D& p3);
    static Vector3D calcNewellNormal4(const Point3D& p1,
                                      const Point3D& p2,
                                      const Point3D& p3,
                                      const Point3D& p4);
    void build4tuple(double v[]) const;
    void build4tuple(float v[]) const;

    // data-members
    double fX, fY, fZ;
};

class Color {
  public:
    // constructors
    Color(void);
    Color(double red, double green, double blue);
    Color(const Color& color);

    // destructor
    ~Color(void);

    // color arithmetic
    Color& operator=(const Color& color);

    Color operator+(const Color& color) const;
    Color operator+(double value) const;
    Color& operator+=(const Color& color);
    Color& operator+=(double value);

    Color operator-(const Color& color) const;
    Color operator-(double value) const;
    Color& operator-=(const Color& color);
    Color& operator-=(double value);

    Color operator*(const Color& color) const;
    Color operator*(double value) const;
    Color& operator*=(const Color& color);
    Color& operator*=(double value);

    Color operator/(const Color& color) const;
    Color operator/(double value) const;
    Color& operator/=(const Color& color);
    Color& operator/=(double value);

    double calcGrayLevel(const Color& color) const;
    Color calcGrayLevelColor(const Color& color) const;

    void clamp(void);

    // methods
    void set(double red, double green, double blue);
    void build4tuple(double v[]) const;
    void build4tuple(float v[]) const;

    // data-members
    double fRed, fGreen, fBlue;
};

#endif
