// ------------------------------
// Filename      : scene.h
// Author        : Sven Maerivoet
// Last modified : 16/06/2001
// Target        : C++ (OpenGL)
//
// All rights reserved.
// ------------------------------

#ifndef SM_SCENE_H // include-guard
#define SM_SCENE_H

#ifdef _WIN32
  #include <windows.h>
#endif

#include "affine.h"
#include "constants.h"
#include "primitives.h"
#include "proceduraltexturer.h"
#include "ray.h"
#include "shapes.h"
#include "solidtexturer.h"
#include <GL/gl.h>
#include <GL/glu.h>
#include <GL/glut.h>
#include <string>
#include <vector>

// string-streams are not always compatible
// => Microsoft Visual C++ 6.0 and UNIX/Linux are treated separately
#ifdef _WIN32
  #include <sstream>
#else
  #include <strstream>
  #define  stringstream strstream
#endif

using namespace std;

class Light {
  public:
    // constructor
    Light(void);

    // destructor
    ~Light(void);

    // data-members
    Point3D  fPosition;
    Color    fColor;
    bool     fIsSpotLight;
    Vector3D fSpotLightDirection;
    double   fCutOffAngle;
    double   fExponent;
};

class DefUnit {
  public:
    // constructor
    DefUnit(string name, string stuff);

    // destructor
    ~DefUnit(void);

    // data-members
    string fName;
    string fStuff;
};

class DefUnitStack {
  public:
    // constructor
    DefUnitStack(void);

    // destructor
    ~DefUnitStack(void);

    // methods
    void push(string name, string stuff);
    bool search(string searchName) const;
    string contents(string searchName) const;
    void release(void);
    void add(DefUnitStack* defUnitStack);

    // data-members
    struct D4S {
      DefUnit*    fCurrent;
      struct D4S* fNext;
    } d4s;

    D4S* fStack;
};

class Line {
  public:
    // constructor
    Line(void);

    // destructor
    ~Line(void);

    // data-members
    Point3D fp1, fp2;
    Color   fColor;
};

class Scene {
  public:
    // constructors
    Scene(void);
    Scene(string filename);

    // destructor
    ~Scene(void);

    // methods
    void clear(void);
    void makeLightsOpenGL(void) const;
    void drawOpenGL(bool depthBufferEnabled) const;
    void readFromFile(string filename);
    bool sceneLoaded(void) const;
    void addLight(Light* light);
    void addGeometricObject(GeometricObject* geometricObject);
    Color shootRay(const Ray& ray);

    void computeShadowsWhenRayTracing(void);
    void dontComputeShadowsWhenRayTracing(void);
    void computeSoftShadowsWhenRayTracing(void);
    void computeHardShadowsWhenRayTracing(void);
    void setShadowFeelerJitterStrength(double shadowFeelerJitterStrength);

    void findNormalsWhenRayTracing(double findNormalsProbability);
    void dontFindNormalsWhenRayTracing(void);

    void followRaysWhenRayTracing(double followRaysProbability);
    void dontFollowRaysWhenRayTracing(void);

    bool isMotionBlurEnabled(void) const;
    double getMotionBlurFrameLength(void) const;
    int getNrOfMotionBlurTimeSamples(void) const;
    void applyShapesMotionBlurAffineTransformation(double timeIndex);
    void restoreShapes(void);

    long getNrOfRaysShot(void) const;
    void resetStatistics(void);
 
    // data-members
    typedef vector<Light*>           Lights;
    typedef vector<GeometricObject*> GeometricObjects;
    typedef vector<Line>             Lines;
    DefUnitStack*                    fDefStack;
    Lights*                          fLights;
    GeometricObjects*                fGeometricObjects;
    bool                             fEyePointSpecified;
    bool                             fViewPointSpecified;
    bool                             fViewAngleSpecified;
    bool                             fUpDirectionSpecified;
    Point3D                          fEyePoint;
    Point3D                          fViewPoint;
    double                           fViewAngle;
    Vector3D                         fUpDirection;
    Lines                            fLines;

  private:
    enum ETokenType {kIDENT,
                     kINCLUDE,
                     kLIGHT,kSPOTLIGHT,
                     kIDENTITYAFFINE,kROTATE,kTRANSLATE,kSCALE,kTRANSFORM,
                     kPUSH,kPOP,
                     kMBCLEAR,kMBROTATE,kMBTRANSLATE,kMBSCALE,
                     kMBENABLE,kMOTIONBLUR,
                     kBUCKYBALL,kCONE,kCUBE,kCYLINDER,kDIAMOND,kDODECAHEDRON,kICOSAHEDRON,kMESH,
                     kOCTAHEDRON,kPYRAMID,kSPHERE,kTAPEREDCYLINDER,kTEAPOT,kTETRAHEDRON,kTORUS,
                     kONESHEETHYPERBOLOID,kTWOSHEETHYPERBOLOID,
                     kELLIPTICCYLINDER,kELLIPTICCONE,
                     kELLIPTICPARABOLOID,kHYPERBOLICPARABOLOID,
                     kPINCHEDSURFACEWITHHOLES,kPINCHEDSURFACEWITHLOBE,
                     kFOURLOBES,kSTEINERSROMANSURFACE,kTOOTHCUBE,
                     kDIFFERENCEBOOL,kINTERSECTIONBOOL,kUNIONBOOL,
                     kDEFAULTMATERIALS,
                     kSURFACEROUGHNESS,
                     kEMISSIVE,kAMBIENT,kDIFFUSE,kSPECULAR,kSPECULAREXPONENT,kLIGHTBACKFACES,
                     kSPEEDOFLIGHT,kTRANSPARENCY,kREFLECTIVITY,kGLOSSSTRENGTH,kTRANSLUCENCYSTRENGTH,kPRIORITY,
                     kRETAINAMBIENTCOLOR,kRETAINDIFFUSECOLOR,kRETAINSPECULARREFLECTION,kDISABLEREFRACTION,
                     kEYEPOINT,kVIEWPOINT,kVIEWANGLE,kUPDIRECTION,
                     kBACKGROUND,kGLOBALAMBIENT,kLIGHTATTENUATION,kATMOSPHERICATTENUATION,kFOG,
                     kSHADOWFEELEREPSILON,kREFLECTIVITYEPSILON,kTRANSPARENCYEPSILON,kMESHEPSILON,
                     kADAPTEDSHADOWCOMPUTATION,kMINREFLECTIVITY,kMINTRANSPARENCY,kMAXRECURSIONDEPTH,
                     kSHOWEXTENTS,
                     kSOLIDTEXTURE,kIMAGETEXTURE,kPROCEDURALTEXTURE,kFLATTEXTUREOFFSET,
                     kLEFTBRACE,kRIGHTBRACE,kDEF,kUSE,
                     kT_NULL,kF_EOF,kUNKNOWN};

    // utility methods
    void initialize(void);
    GeometricObject* getGeometricObject(void);
    string getNextToken(void);
    double getDouble(void);
    bool isIdentifier(string keyword) const;
    ETokenType getToken(string keyword) const;
    double calcShadowFactor(const Ray& ray);
    void resetMotionBlurTransformations(void);

    // data-members
    bool                      fSceneLoaded;
    SolidTexturer             fSolidTexturer;
    ProceduralTexturer        fProceduralTexturer;
    int                       fCurrentLine;
    int                       fNextLine;
    stringstream*             fInputStream;
    AffineTransformationStack fAffineStack;
    AffineTransformationStack fMotionBlurAffineStack;
    Material                  fCurrentMaterial;
    bool                      fComputeShadows;
    bool                      fComputeSoftShadows;
    double                    fShadowFeelerJitterStrength;
    bool                      fShowExtents;
    Color                     fBackgroundColor;
    Color                     fGlobalAmbientColor;
    double                    fLightAttenuationConstantFactor;
    double                    fLightAttenuationLinearFactor;
    double                    fLightAttenuationQuadraticFactor;
    double                    fAtmosphericAttenuationFrontPlane;
    double                    fAtmosphericAttenuationFrontScale;
    double                    fAtmosphericAttenuationBackPlane;
    double                    fAtmosphericAttenuationBackScale;
    Color                     fAtmosphericAttenuationColor;
    double                    fFogStart;
    double                    fFogEnd;
    Color                     fFogColor;
    double                    fShadowFeelerEpsilon;
    double                    fReflectivityEpsilon;
    double                    fTransparencyEpsilon;
    bool                      fAdaptedShadowComputation;
    double                    fMinReflectivity;
    double                    fMinTransparency;
    int                       fMaxRecursionDepth;
    bool                      fFindNormals;
    double                    fFindNormalsProbability;
    bool                      fFollowRays;
    double                    fFollowRaysProbability;
    bool                      fMotionBlurEnabled;
    double                    fMotionBlurFrameLength;
    int                       fNrOfMotionBlurTimeSamples;
    double                    fMotionBlurRotationAngle;
    Vector3D                  fMotionBlurRotationAxis;
    Vector3D                  fMotionBlurTranslation;
    Vector3D                  fMotionBlurScale;
    long                      fNrOfRaysShot;
};

#endif
