// ------------------------------
// Filename      : shapes.cpp
// Author        : Sven Maerivoet
// Last modified : 16/06/2001
// Target        : C++ (OpenGL)
//
// All rights reserved.
// ------------------------------

/*
  If you want to use extents, derive a specific extent or use the box-
  or sphere-extent and override the following method :

    void UserDefinedShape::computeExtent(void)
    {
      // create the point-cluster for the generic user-defined-shape
      PointCluster pointCluster(2);
      pointCluster.fPoints[0] = Point3D(-1.0,0.0,0.0); // e.g.
      pointCluster.fPoints[1] = Point3D(1.0,0.0,0.0); // e.g.

      // apply the user-defined-shape's specific affine transformation
      // to express the point-cluster in world coordinates
      // instead of generic shape coordinates
      pointCluster.applyAffineTransformation(fAffineTransformation);

      // note that instead you can only perform the translation-part of the
      // specified affine transformation, which retains the shape of the
      // generic extent
      // to do this, use :
      //   pointCluster.applyTranslation(fAffineTransformation);

      fWorldExtent = new SphereExtent(); // e.g.: we use the sphere-extent
      // use a small excess for the extent (e.g. 0.1)
      fWorldExtent->create(pointCluster,0.1);
    }

  You must also include an extent-intersection-test :

    bool UserDefinedShape::computeIntersectionsWithRay(const Ray& ray,
                                                       Intersection& intersection)
    {
      // perform extent-intersection-test (in world-coordinates !)
      if (!fWorldExtent->rayIntersectsExtent(ray)) {
        return false;
      }

      // rest of code...
    }
*/

#include "glm.h"
#include "math.h"
#include "shapes.h"
#include <iostream>
#include <fstream>
#include <math.h>

using namespace std;

// define for easier use
#define MT(x) model->triangles[(x)]

// --------------
// class Material
// --------------

// -----------
// constructor
// -----------
Material::Material(void)
{
  setDefault();
}

// ----------
// destructor
// ----------
Material::~Material(void)
{
}

// -------
// methods
// -------
void Material::setDefault(void)
{
  fSurfaceRoughness = 0.0;
  fEmissive.set(0.0,0.0,0.0);
  fAmbient.set(0.1,0.1,0.1);
  fDiffuse.set(0.8,0.8,0.8);
  fSpecular.set(0.0,0.0,0.0);
  fSpecularExponent = 1.0;
  fLightBackFaces = false;
  fReflectivity = 0.0;
  fGlossStrength = 0.0;
  fTransparency = 0.0;
  fPriority = 0;
  fTranslucencyStrength = 0.0;
  fSpeedOfLight = 1.0;
  fRetainAmbientColor = false;
  fRetainDiffuseColor = false;
  fRetainSpecularReflection = false;
  fDisableRefraction = false;
  fSolidTextureType = SolidTexturer::kNone;
  int i;
  for (i = 0; i < kMaxSolidTextureParameters; ++i) {
    fSolidTextureParameters[i] = 0.0;
  }
  fHasImageTexture = false;
  fImageTextureBitmap = 0;
  fImageTextureTileS = 1.0;
  fImageTextureTileT = 1.0;
  fProceduralTextureType = ProceduralTexturer::kNone;
  for (i = 0; i < kMaxProceduralTextureParameters; ++i) {
    fProceduralTextureParameters[i] = 0.0;
  }
  fFlatTextureIsModulating = false;
  fFlatTextureSOffset = 0.0;
  fFlatTextureTOffset = 0.0;
  fMotionBlurEnabled = false;
}

void Material::set(const Material& material)
{
  fSurfaceRoughness = material.fSurfaceRoughness;
  fEmissive = material.fEmissive;
  fAmbient = material.fAmbient;
  fDiffuse = material.fDiffuse;
  fSpecular = material.fSpecular;
  fSpecularExponent = material.fSpecularExponent;
  fLightBackFaces = material.fLightBackFaces;
  fReflectivity = material.fReflectivity;
  fGlossStrength = material.fGlossStrength;
  fTransparency = material.fTransparency;
  fPriority = material.fPriority;
  fTranslucencyStrength = material.fTranslucencyStrength;
  fSpeedOfLight = material.fSpeedOfLight;
  fRetainAmbientColor = material.fRetainAmbientColor;
  fRetainDiffuseColor = material.fRetainDiffuseColor;
  fRetainSpecularReflection = material.fRetainSpecularReflection;
  fDisableRefraction = material.fDisableRefraction;
  fSolidTextureType = material.fSolidTextureType;
  int i;
  for (i = 0; i < kMaxSolidTextureParameters; ++i) {
    fSolidTextureParameters[i] = material.fSolidTextureParameters[i];
  }
  fHasImageTexture = material.fHasImageTexture;
  fImageTextureBitmap = material.fImageTextureBitmap;
  fImageTextureTileS = material.fImageTextureTileS;
  fImageTextureTileT = material.fImageTextureTileT;
  fProceduralTextureType = material.fProceduralTextureType;
  for (i = 0; i < kMaxProceduralTextureParameters; ++i) {
    fProceduralTextureParameters[i] = material.fProceduralTextureParameters[i];
  }
  fFlatTextureIsModulating = material.fFlatTextureIsModulating;
  fFlatTextureSOffset = material.fFlatTextureSOffset;
  fFlatTextureTOffset = material.fFlatTextureTOffset;
  fMotionBlurEnabled = material.fMotionBlurEnabled;
}

// ---------------------
// class GeometricObject
// ---------------------

// -----------
// constructor
// -----------
GeometricObject::GeometricObject(void) : fWorldExtent(0),
                                         fWarningShown(false)
{
}

// ----------
// destructor
// ----------
GeometricObject::~GeometricObject(void)
{
  delete fWorldExtent;
}

// -------
// methods
// -------
void GeometricObject::computeExtent(void)
{
}

void GeometricObject::drawExtent(void) const
{
  if (fWorldExtent != 0) {
    fWorldExtent->draw();
  }
}

void GeometricObject::setMaterial(const Material& material)
{
  fMaterial.set(material);
}

void GeometricObject::drawOpenGL(void) const
{
}

void GeometricObject::tellMaterialsGL(void) const
{
}

bool GeometricObject::computeIntersectionsWithRay(const Ray& ray, Intersection& intersection)
{
  return false;
}

void GeometricObject::textureLookup(const HitInfo& hitInfo, double& s, double& t) const
{
  s = 0.0;
  t = 0.0;
}

void GeometricObject::computeTextureCoordinates(const HitInfo& hitInfo, double& s, double& t) const
{
  textureLookup(hitInfo,s,t);
  
  // apply the texture-offset
  s += fMaterial.fFlatTextureSOffset;
  if (s > 1.0) {
    s = Math::frac(s);
  }

  t += fMaterial.fFlatTextureTOffset;
  if (t > 1.0) {
    t = Math::frac(t);
  }

  // if necessary, clip s and t in the [0,1] range
  if (s < 0.0) {
    s = 0.0;
  }
  else if (s > 1.0) {
    s = 1.0;
  }

  if (t < 0.0) {
    t = 0.0;
  }
  else if (t > 1.0) {
    t = 1.0;
  }

  // tile the texture
  s = Math::frac(s * fMaterial.fImageTextureTileS);
  t = Math::frac(t * fMaterial.fImageTextureTileT);
}

void GeometricObject::applyMotionBlurTransformation(double timeIndex, double frameLength)
{
}

void GeometricObject::restoreOriginalTransformation(void)
{
}

// -------------
// class Boolean
// -------------

// -----------
// constructor
// -----------
Boolean::Boolean(void) : fLeft(0),
                         fRight(0)
{
}

// ----------
// destructor
// ----------
Boolean::~Boolean(void)
{
}

// -------
// methods
// -------
void Boolean::drawOpenGL(void) const
{
  // draw the children
  if(fLeft != 0) {
    fLeft->drawOpenGL();
  }

  if(fRight != 0) {
    fRight->drawOpenGL();
  }
}

bool Boolean::computeIntersectionsWithRay(const Ray &ray, Intersection& intersection)
{
  return false;
}

void Boolean::applyMotionBlurTransformation(double timeIndex, double frameLength)
{
  // traverse the CSG-tree
  if (fLeft != 0) {
    fLeft->applyMotionBlurTransformation(timeIndex,frameLength);
  }

  if (fRight != 0) {
    fRight->applyMotionBlurTransformation(timeIndex,frameLength);
  }
}

void Boolean::restoreOriginalTransformation(void)
{
  // traverse the CSG-tree
  if (fLeft != 0) {
    fLeft->restoreOriginalTransformation();
  }

  if (fRight != 0) {
    fRight->restoreOriginalTransformation();
  }
}

// ---------------
// class UnionBool
// ---------------

// -----------
// constructor
// -----------
UnionBool::UnionBool(void)
{
}

// ----------
// destructor
// ----------
UnionBool::~UnionBool(void)
{
}

// -------
// methods
// -------
bool UnionBool::computeIntersectionsWithRay(const Ray &ray, Intersection& intersection)
{
  Intersection leftIntersection;
  Intersection rightIntersection;
  leftIntersection.fNrOfHits = 0;
  rightIntersection.fNrOfHits = 0;

  // test for early-out
  bool leftHit = fLeft->computeIntersectionsWithRay(ray,leftIntersection);
  bool rightHit = fRight->computeIntersectionsWithRay(ray,rightIntersection);
  if ((!leftHit) && (!rightHit)) {
    return false;
  }

  // make combined lists
  bool leftInside = false;
  if (leftIntersection.fNrOfHits > 0) {
    leftInside = !(leftIntersection.fHits[0].fIsEntering);
  }
  bool rightInside = false;
  if (rightIntersection.fNrOfHits > 0) {
    rightInside = !(rightIntersection.fHits[0].fIsEntering);
  }
  bool combinedInside = leftInside || rightInside;

  // move through the lists until one is consumed
  int indexLeft = 0;
  int indexRight = 0;
  int indexCombined = -1;
  HitInfo nextHit;

  // consider all hit-times until at least one list has been consumed
  while ((indexLeft < leftIntersection.fNrOfHits) && 
         (indexRight < rightIntersection.fNrOfHits)) {

    bool newInside;

    if (leftIntersection.fHits[indexLeft].fHitTime <= rightIntersection.fHits[indexRight].fHitTime) {
      nextHit = leftIntersection.fHits[indexLeft];
      leftInside = !leftInside;
      newInside = leftInside || rightInside;
      ++indexLeft;
    }
    else {
      nextHit = rightIntersection.fHits[indexRight];
      rightInside = !rightInside;
      newInside = leftInside || rightInside;
      ++indexRight;
    } // else

    if (newInside != combinedInside) {
      ++indexCombined;
      intersection.fHits[indexCombined] = nextHit;
      combinedInside = newInside;
      intersection.fHits[indexCombined].fIsEntering = combinedInside;
    }
  } // while

  // gobble up the rest of the unconsumed left-list
  while (indexLeft < leftIntersection.fNrOfHits) {
    nextHit = leftIntersection.fHits[indexLeft];
    leftInside = !leftInside;
    ++indexCombined;
    intersection.fHits[indexCombined] = nextHit;
    intersection.fHits[indexCombined].fIsEntering = leftInside;
    ++indexLeft;
  }

  // gobble up the rest of the unconsumed right-list
  while (indexRight < rightIntersection.fNrOfHits) {
    nextHit = rightIntersection.fHits[indexRight];
    rightInside = !rightInside;
    ++indexCombined;
    intersection.fHits[indexCombined] = nextHit;
    intersection.fHits[indexCombined].fIsEntering = rightInside;
    ++indexRight;
  }

  intersection.fNrOfHits = indexCombined + 1;
  return (intersection.fNrOfHits > 0);
}

// ----------------------
// class IntersectionBool
// ----------------------

// -----------
// constructor
// -----------
IntersectionBool::IntersectionBool(void)
{
}

// ----------
// destructor
// ----------
IntersectionBool::~IntersectionBool(void)
{
}

// -------
// methods
// -------
bool IntersectionBool::computeIntersectionsWithRay(const Ray &ray, Intersection& intersection)
{
  Intersection leftIntersection;
  Intersection rightIntersection;
  leftIntersection.fNrOfHits = 0;
  rightIntersection.fNrOfHits = 0;

  // test for early-out
  bool leftHit = fLeft->computeIntersectionsWithRay(ray,leftIntersection);
  bool rightHit = fRight->computeIntersectionsWithRay(ray,rightIntersection);
  if ((!leftHit) || (!rightHit)) {
    return false;
  }

  // make combined lists
  bool leftInside = !(leftIntersection.fHits[0].fIsEntering);
  bool rightInside = !(rightIntersection.fHits[0].fIsEntering);
  bool combinedInside = leftInside && rightInside;

  // move through the lists until one is consumed
  int indexLeft = 0;
  int indexRight = 0;
  int indexCombined = -1;
  HitInfo nextHit;

  // consider all hit-times until at least one list has been consumed
  while ((indexLeft < leftIntersection.fNrOfHits) && 
         (indexRight < rightIntersection.fNrOfHits)) {

    bool newInside;

    if (leftIntersection.fHits[indexLeft].fHitTime <= rightIntersection.fHits[indexRight].fHitTime) {
      nextHit = leftIntersection.fHits[indexLeft];
      leftInside = !leftInside;
      newInside = leftInside && rightInside;
      ++indexLeft;
    }
    else {
      nextHit = rightIntersection.fHits[indexRight];
      rightInside = !rightInside;
      newInside = leftInside && rightInside;
      ++indexRight;
    } // else

    if (newInside != combinedInside) {
      ++indexCombined;
      intersection.fHits[indexCombined] = nextHit;
      combinedInside = newInside;
      intersection.fHits[indexCombined].fIsEntering = combinedInside;
    }
  } // while

  // ignore the rest of the unconsumed lists

  intersection.fNrOfHits = indexCombined + 1;
  return (intersection.fNrOfHits > 0);
}

// --------------------
// class DifferenceBool
// --------------------

// -----------
// constructor
// -----------
DifferenceBool::DifferenceBool(void)
{
}

// ----------
// destructor
// ----------
DifferenceBool::~DifferenceBool(void)
{
}

// -------
// methods
// -------
bool DifferenceBool::computeIntersectionsWithRay(const Ray &ray, Intersection& intersection)
{
  Intersection leftIntersection;
  Intersection rightIntersection;
  leftIntersection.fNrOfHits = 0;
  rightIntersection.fNrOfHits = 0;

  // test for early-out
  if (!fLeft->computeIntersectionsWithRay(ray,leftIntersection)) {
    return false;
  }

  if (!fRight->computeIntersectionsWithRay(ray,rightIntersection)) {
    intersection = leftIntersection;
    return true;
  }

  // make combined lists
  bool leftInside = !(leftIntersection.fHits[0].fIsEntering);
  bool rightInside = !(rightIntersection.fHits[0].fIsEntering);
  bool combinedInside = leftInside && (!rightInside);

  // move through the lists until one is consumed
  int indexLeft = 0;
  int indexRight = 0;
  int indexCombined = -1;
  HitInfo nextHit;

  GeometricObject* leftObject = leftIntersection.fHits[0].fObjectHit;

  // consider all hit-times until at least one list has been consumed
  while ((indexLeft < leftIntersection.fNrOfHits) && 
         (indexRight < rightIntersection.fNrOfHits)) {

    bool newInside;

    if (leftIntersection.fHits[indexLeft].fHitTime <= rightIntersection.fHits[indexRight].fHitTime) {
      nextHit = leftIntersection.fHits[indexLeft];
      leftInside = !leftInside;
      newInside = leftInside && (!rightInside);
      leftObject = nextHit.fObjectHit;
      ++indexLeft;

      if (newInside != combinedInside) {
        ++indexCombined;
        intersection.fHits[indexCombined] = nextHit;
        combinedInside = newInside;
        intersection.fHits[indexCombined].fIsEntering = combinedInside;
      }
    }
    else {
      nextHit = rightIntersection.fHits[indexRight];
      rightInside = !rightInside;
      newInside = leftInside && (!rightInside);
      ++indexRight;

      if (newInside != combinedInside) {
        ++indexCombined;
        intersection.fHits[indexCombined] = nextHit;
        combinedInside = newInside;
        intersection.fHits[indexCombined].fIsEntering = combinedInside;

        // correctly model the "hole" in the object
        if (combinedInside) {
          nextHit.fHitNormal.flip();
          intersection.fHits[indexCombined].fObjectHit = leftObject;
        }
      }
    } // else
  } // while

  // gobble up the rest of the unconsumed left-list
  while (indexLeft < leftIntersection.fNrOfHits) {
    nextHit = leftIntersection.fHits[indexLeft];
    leftInside = !leftInside;
    ++indexCombined;
    intersection.fHits[indexCombined] = nextHit;
    intersection.fHits[indexCombined].fIsEntering = leftInside;
    ++indexLeft;
  }

  intersection.fNrOfHits = indexCombined + 1;
  return (intersection.fNrOfHits > 0);
}

// -----------
// class Shape
// -----------

// -----------
// constructor
// -----------
Shape::Shape(void) : fMotionBlurRotationAngle(0.0),
                     fMotionBlurRotationAxis(Vector3D(0.0,1.0,0.0)),
                     fMotionBlurTranslation(Vector3D(0.0,0.0,0.0)),
                     fMotionBlurScale(Vector3D(1.0,1.0,1.0))
{
}

// ----------
// destructor
// ----------
Shape::~Shape(void)
{
}

// -------
// methods
// -------
void Shape::computeExtent(void)
{
}

void Shape::drawOpenGL(void) const
{
}

void Shape::tellMaterialsGL(void) const
{
  GLfloat ambient[4]  = {(GLfloat) 0.0,
                         (GLfloat) 0.0,
                         (GLfloat) 0.0,
                         (GLfloat) 1.0};
  GLfloat diffuse[4]  = {(GLfloat) 0.0,
                         (GLfloat) 0.0,
                         (GLfloat) 0.0,
                         (GLfloat) 1.0};
  GLfloat specular[4] = {(GLfloat) 0.0,
                         (GLfloat) 0.0,
                         (GLfloat) 0.0,
                         (GLfloat) 1.0};
  GLfloat emissive[4] = {(GLfloat) 0.0,
                         (GLfloat) 0.0,
                         (GLfloat) 0.0,
                         (GLfloat) 1.0};

  fMaterial.fAmbient.build4tuple(ambient);
  fMaterial.fDiffuse.build4tuple(diffuse);
  fMaterial.fSpecular.build4tuple(specular);
  fMaterial.fEmissive.build4tuple(emissive);

  glMaterialfv(GL_FRONT,GL_AMBIENT,ambient);
  glMaterialfv(GL_FRONT,GL_DIFFUSE,diffuse);
  glMaterialfv(GL_FRONT,GL_SPECULAR,specular);
  glMaterialfv(GL_FRONT,GL_EMISSION,emissive);
  glMaterialf(GL_FRONT,GL_SHININESS,(GLfloat) fMaterial.fSpecularExponent);
}

bool Shape::computeIntersectionsWithRay(const Ray& ray, Intersection& intersection)
{
  return false;
}

void Shape::textureLookup(const HitInfo& hitInfo, double& s, double& t) const
{
  s = 0.0;
  t = 0.0;
}

void Shape::applyMotionBlurTransformation(double timeIndex, double frameLength)
{
  if (fMaterial.fMotionBlurEnabled) {

    // adjust the motion blur to the current time-index 

    // interpolate between the scale at the beginning (1.0) and the end
    double fraction = (timeIndex  / frameLength) + 0.5;
    if (fraction < 0.0) {
      fraction = 0.0;
    }
    else if (fraction > 1.0) {
      fraction = 1.0;
    }
    double sx = 1.0 + (fraction * (fMotionBlurScale.fX - 1.0));
    double sy = 1.0 + (fraction * (fMotionBlurScale.fY - 1.0));
    double sz = 1.0 + (fraction * (fMotionBlurScale.fZ - 1.0));

    double rotationAngle = fMotionBlurRotationAngle * timeIndex;

    Vector3D translation(fMotionBlurTranslation);
    translation *= timeIndex;

    AffineTransformation temp;

    // construct homogenous motion blur affine transformation
    AffineTransformation motionBlurAffineTransformation(
      AffineTransformation::convertScalingToTransformation(sx,sy,sz));
    temp = AffineTransformation::convertRotationToTransformation(rotationAngle,
                                                                 fMotionBlurRotationAxis);
    motionBlurAffineTransformation.postMult(temp);
    temp = AffineTransformation::convertTranslationToTransformation(translation);
    motionBlurAffineTransformation.postMult(temp);

    // construct homogenous motion blur inverse affine transformation
    AffineTransformation motionBlurInverseAffineTransformation(
      AffineTransformation::convertScalingToTransformation(1.0 / sx,1.0 / sy,1.0 / sz));
    temp = AffineTransformation::convertRotationToTransformation(-rotationAngle,
                                                                 fMotionBlurRotationAxis);
    motionBlurInverseAffineTransformation.preMult(temp);
    temp = AffineTransformation::convertTranslationToTransformation(translation * (-1.0));
    motionBlurInverseAffineTransformation.preMult(temp);

    // apply the motion blur transformation
    fAffineTransformation.postMult(motionBlurAffineTransformation);
    fInverseAffineTransformation.preMult(motionBlurInverseAffineTransformation);

    // recompute the extent because the object might have changed scale/orientation/position
    computeExtent();
  }
}

void Shape::restoreOriginalTransformation(void)
{
  if (fMaterial.fMotionBlurEnabled) {

    fAffineTransformation.set(fOriginalAffineTransformation);
    fInverseAffineTransformation.set(fOriginalInverseAffineTransformation);

    // recompute the extent because the object might have changed position/orientation/scale
    computeExtent();
  }
}

// ----------
// class Cube
// ----------

// -----------
// constructor
// -----------
Cube::Cube(void)
{
}

// ----------
// destructor
// ----------
Cube::~Cube(void)
{
}

// -------
// methods
// -------
void Cube::drawOpenGL(void) const
{
  tellMaterialsGL();
  glPushMatrix();
    glMultMatrixd(fAffineTransformation.m);
    glutSolidCube((GLdouble) 2.0);
  glPopMatrix();
}

bool Cube::computeIntersectionsWithRay(const Ray& ray, Intersection& intersection)
{
  // there's no need to perform extent-testing
  Ray genericRay;
  Transformations::transformRay(genericRay,fInverseAffineTransformation,ray);

  double hitTime, numerator, denominator;
  double hitTimeIn = -Math::kInfinity;
  double hitTimeOut = Math::kInfinity;
  int innerSurface, outerSurface;

  for (int i = 0; i < 6; ++i) {

    // find the cube's plane to test
    switch (i) {

      case 0 :
        numerator = 1.0 - genericRay.fStart.fY;
        denominator = genericRay.fDirection.fY;
        break;

      case 1 :
        numerator = 1.0 + genericRay.fStart.fY;
        denominator = -genericRay.fDirection.fY;
        break;

      case 2 :
        numerator = 1.0 - genericRay.fStart.fX;
        denominator = genericRay.fDirection.fX;
        break;

      case 3 :
        numerator = 1.0 + genericRay.fStart.fX;
        denominator = -genericRay.fDirection.fX;
        break;

      case 4 :
        numerator = 1.0 - genericRay.fStart.fZ;
        denominator = genericRay.fDirection.fZ;
        break;

      case 5 :
        numerator = 1.0 + genericRay.fStart.fZ;
        denominator = -genericRay.fDirection.fZ;
        break;
    }

    if (Math::abs(denominator) < Math::kEpsilon) {
      // ray is parallel to plane
      if (numerator < 0.0) {
        // ray is outside
        return false;
      }
    }
    else {
      hitTime = numerator / denominator;
      if (denominator > 0.0) {
        // ray is exiting
        if (hitTime < hitTimeOut) {
          // a new earlier exit
          hitTimeOut = hitTime;
          outerSurface = i;
        }
      }
      else {
        // ray is entering
        if (hitTime > hitTimeIn) {
          // a new later entrance
          hitTimeIn = hitTime;
          innerSurface = i;
        }
      }
    }

    if (hitTimeIn >= hitTimeOut) {
      // a miss - early out
      return false;
    }
  }

  intersection.fNrOfHits = 0;

  if (hitTimeIn > 0.0) { // a positive hit-time

    // find the hit-point
    Point3D p(Transformations::calcRayPosition(genericRay,hitTimeIn));

    intersection.fHits[0].fHitTime = hitTimeIn;
    intersection.fHits[0].fObjectHit = this;
    intersection.fHits[0].fSurface = innerSurface;
    intersection.fHits[0].fIsEntering = true;
    intersection.fHits[0].fHitPoint = p;
    intersection.fHits[0].fHitNormal = calcNormal(innerSurface);
    ++intersection.fNrOfHits;
  }

  if (hitTimeOut > 0.0) { // a positive hit-time

    // find the hit-point
    Point3D p(Transformations::calcRayPosition(genericRay,hitTimeOut));

    intersection.fHits[intersection.fNrOfHits].fHitTime = hitTimeOut;
    intersection.fHits[intersection.fNrOfHits].fObjectHit = this;
    intersection.fHits[intersection.fNrOfHits].fSurface = outerSurface;
    intersection.fHits[intersection.fNrOfHits].fIsEntering = false;
    intersection.fHits[intersection.fNrOfHits].fHitPoint = p;
    intersection.fHits[intersection.fNrOfHits].fHitNormal = calcNormal(innerSurface);

    ++intersection.fNrOfHits;
  }

  return (intersection.fNrOfHits > 0);
}

void Cube::textureLookup(const HitInfo& hitInfo, double& s, double& t) const
{
  Point3D p(hitInfo.fHitPoint);

  switch (hitInfo.fSurface) {

    case 0 :
      s = (1.0 + p.fX) / 2.0;
      t = (1.0 - p.fZ) / 2.0;
      break;

    case 1 :
      s = (1.0 + p.fX) / 2.0;
      t = (1.0 + p.fZ) / 2.0;
      break;

    case 2 :
      s = (1.0 - p.fZ) / 2.0;
      t = (1.0 + p.fY) / 2.0;
      break;

    case 3 :
      s = (1.0 + p.fZ) / 2.0;
      t = (1.0 + p.fY) / 2.0;
      break;

    case 4 :
      s = (1.0 + p.fX) / 2.0;
      t = (1.0 + p.fY) / 2.0;
      break;

    case 5 :
      s = (1.0 - p.fX) / 2.0;
      t = (1.0 + p.fY) / 2.0;
      break;
  }
}

// ---------------
// utility methods
// ---------------
Vector3D Cube::calcNormal(int surface) const
{
  Vector3D normal;

  switch (surface) {

    case 0 :
      normal.set(0.0,1.0,0.0);
      break;

    case 1 :
      normal.set(0.0,-1.0,0.0);
      break;

    case 2 :
      normal.set(1.0,0.0,0.0);
      break;

    case 3 :
      normal.set(-1.0,0.0,0.0);
      break;

    case 4 :
      normal.set(0.0,0.0,1.0);
      break;

    case 5 :
      normal.set(0.0,0.0,-1.0);
      break;
  }

  return normal;
}

// ----------
// class Face
// ----------

// -----------
// constructor
// -----------
Face::Face(void) : fNrOfVertices(0),
                   fVertices(0)
{
}

// ----------
// destructor
// ----------
Face::~Face(void)
{
  if (fVertices != 0) {
    delete[] fVertices;
  }
  fNrOfVertices = 0;
}

// ----------
// class Mesh
// ----------

// -----------
// constructor
// -----------
Mesh::Mesh(string filename, Mesh::EMeshType meshType) : fInterpolateNormals(false),
                                                        fNrOfVertices(0),
                                                        fNrOfNormals(0),
                                                        fNrOfFaces(0),
                                                        fNrOfTriangulatedFaces(0),
                                                        fVertices(0),
                                                        fNormals(0),
                                                        fFaces(0),
                                                        fTriangulatedFaces(0),
                                                        fLastVertexUsed(-1),
                                                        fLastNormalUsed(-1),
                                                        fLastFaceUsed(-1),
                                                        fFilename(),
                                                        fIsTriangulated(false)
{
  readFromFile(filename,meshType);
}

// ----------
// destructor
// ----------
Mesh::~Mesh(void)
{
}

// -------
// methods
// -------
void Mesh::readFromFile(string filename, Mesh::EMeshType meshType)
{
  fFilename = filename;
  fstream inStream;
  inStream.open(fFilename.c_str(),ios::in);

  if (inStream.fail() ||
      inStream.eof()) {
    cerr << "Mesh::readFromFile() -> can't open inputfile " << fFilename
         << " for reading or EOF detected." << endl;
    clear();
    return;
  }

  cerr << "Reading mesh from [" << fFilename << "]..." << endl;
 
  if (meshType == k3VN) {
    inStream >> fNrOfVertices
             >> fNrOfNormals
             >> fNrOfFaces;

    fVertices = new Point3D[fNrOfVertices];
    fNormals = new Vector3D[fNrOfNormals];
    fFaces = new Face[fNrOfFaces];

    int i;
    for (i = 0; i < fNrOfVertices; ++i) {
      inStream >> fVertices[i].fX
               >> fVertices[i].fY
               >> fVertices[i].fZ;
    }

    for (i = 0; i < fNrOfNormals; ++i) {
      inStream >> fNormals[i].fX
               >> fNormals[i].fY
               >> fNormals[i].fZ;
    }
	
    for (i = 0; i < fNrOfFaces; ++i) {
      inStream >> fFaces[i].fNrOfVertices;

      int n = fFaces[i].fNrOfVertices;

      fFaces[i].fVertices = new VertexID[n];

      int k = 0;
      for (k = 0; k < n; ++k) {
        inStream >> fFaces[i].fVertices[k].fVertexIndex;
      }

      for (k = 0; k < n; ++k) {
        inStream >> fFaces[i].fVertices[k].fNormalIndex;
      }
    }
  }
  else if (meshType == k3DV) {
    inStream >> fNrOfVertices >> fNrOfFaces;
    fNrOfNormals = fNrOfVertices;

    fVertices = new Point3D[fNrOfVertices];
    fNormals = new Vector3D[fNrOfNormals];
    fFaces = new Face[fNrOfFaces];

    int i;
    for (i = 0; i < fNrOfVertices; ++i) {
      inStream >> fVertices[i].fX
               >> fVertices[i].fY
               >> fVertices[i].fZ
               >> fNormals[i].fX
               >> fNormals[i].fY
               >> fNormals[i].fZ;
    }

    for (i = 0; i < fNrOfFaces; ++i) {
      inStream >> fFaces[i].fNrOfVertices;

      int n = fFaces[i].fNrOfVertices;

      fFaces[i].fVertices = new VertexID[n];

      int k = 0;
      for (k = 0; k < n; ++k) {
        inStream >> fFaces[i].fVertices[k].fVertexIndex;
        fFaces[i].fVertices[k].fNormalIndex = fFaces[i].fVertices[k].fVertexIndex;
      }
    }
  }
  else if (meshType == k3DM) {
    inStream >> fNrOfVertices >> fNrOfFaces;
    fNrOfNormals = fNrOfFaces;

    fVertices = new Point3D[fNrOfVertices];
    fNormals = new Vector3D[fNrOfNormals];
    fFaces = new Face[fNrOfFaces];

    int i;
    for (i = 0; i < fNrOfVertices; ++i) {
      inStream >> fVertices[i].fX
               >> fVertices[i].fY
               >> fVertices[i].fZ;
    }

    for (i = 0; i < fNrOfFaces; ++i) {
      inStream >> fFaces[i].fNrOfVertices;

      int n = fFaces[i].fNrOfVertices;

      fFaces[i].fVertices = new VertexID[n];

      int k;
      for (k = 0; k < n; ++k) {
        inStream >> fFaces[i].fVertices[k].fVertexIndex;
      }

      // manually calculate the normal on this surface using Martin Newell's method
      fNormals[i].set(0.0,0.0,0.0);
      for (int j = 0; j < n; ++j) {
        int successor = (j + 1) % n;
        Point3D p1 = fVertices[fFaces[i].fVertices[j].fVertexIndex];
        Point3D p2 = fVertices[fFaces[i].fVertices[successor].fVertexIndex];

        fNormals[i].fX += (p1.fY - p2.fY) * (p1.fZ + p2.fZ);
        fNormals[i].fY += (p1.fZ - p2.fZ) * (p1.fX + p2.fX);
        fNormals[i].fZ += (p1.fX - p2.fX) * (p1.fY + p2.fY);
      }

      // assign this face's normal's index to all the vertices in the face
      for (k = 0; k < n; ++k) {
        fFaces[i].fVertices[k].fNormalIndex = i;
      }
    }
  }
  else if (meshType == kOBJ) {

    GLMmodel *model = glmReadOBJ(filename.c_str());
    if (!model) {
      cerr << "Mesh::readFromFile() -> error while parsing Wavefront Object file ["
           << fFilename << "]." << endl;
      clear();
      return;
    }

    if (model->numnormals == 0) {
      glmFacetNormals(model);
      glmVertexNormals(model,35.0f);
    }

    // convert the model to a mesh
    glmUnitize(model);

    // calculate the total number of vertices in the model
    fNrOfVertices = model->numvertices;
    fVertices = new Point3D[fNrOfVertices];
    for (int vertex = 1; vertex <= fNrOfVertices; ++vertex) {
      fVertices[vertex - 1] = Point3D(model->vertices[3 * vertex],
                                      model->vertices[(3 * vertex) + 1],
                                      model->vertices[(3 * vertex) + 2]);
    }
    
    // calculate the total number of normals in the model
    fNrOfNormals = model->numnormals;
    fNormals = new Vector3D[fNrOfNormals];
    for (int normal = 1; normal <= fNrOfNormals; ++normal) {
      fNormals[normal - 1] = Vector3D(model->normals[3 * normal],
                                      model->normals[(3 * normal) + 1],
                                      model->normals[(3 * normal) + 2]);
    }

    // calculate the total number of faces in the model
    fNrOfFaces = 0;
    GLMgroup* group = model->groups;
    while (group) {
      fNrOfFaces += group->numtriangles;
      group = group->next;
    }
    fFaces = new Face[fNrOfFaces];

    // construct the faces
    int face = 0;
    group = model->groups;
    while (group) {

      for (int triangle = 0; triangle < group->numtriangles; ++triangle) {

        fFaces[face].fNrOfVertices = 3;

        fFaces[face].fVertices = new VertexID[fFaces[face].fNrOfVertices];
        for (int k = 0; k < fFaces[face].fNrOfVertices; ++k) {
          fFaces[face].fVertices[k].fVertexIndex = MT(group->triangles[triangle]).vindices[k] - 1;
          fFaces[face].fVertices[k].fNormalIndex = MT(group->triangles[triangle]).nindices[k] - 1;
        }
        ++face;
      }
      group = group->next;
    }

    glmDelete(model);
  }

  inStream.close();

  cerr << "  -> #vertices(" << fNrOfVertices
       << "), #normals(" << fNrOfNormals
       << "), #faces(" << fNrOfFaces << ")" << endl;
}

void Mesh::writeToFile(string filename, EMeshType meshType) const
{
  if (isEmpty()) {
    return;
  }

  fstream outStream;
  outStream.open(filename.c_str(),ios::out);

  if (outStream.fail()) {
    cerr << "Mesh::writeToFile() -> can't open outputfile [" << filename
         << "] for writing." << endl;
    return;
  }

  if (meshType == k3VN) {
    outStream << fNrOfVertices << endl;
    outStream << fNrOfNormals << endl;
    outStream << fNrOfFaces << endl;

    int i;
    for (i = 0; i < fNrOfVertices; ++i) {
      outStream << fVertices[i].fX << " "
                << fVertices[i].fY << " "
                << fVertices[i].fZ << endl;
    }

    for (i = 0; i < fNrOfNormals; ++i) {
      outStream << fNormals[i].fX << " "
                << fNormals[i].fY << " "
                << fNormals[i].fZ << endl;
    }

    for (i = 0; i < fNrOfFaces; ++i) {
      outStream << fFaces[i].fNrOfVertices << endl;

      int n = fFaces[i].fNrOfVertices;

      int k = 0;
      for (k = 0; k < n; ++k) {
        outStream << fFaces[i].fVertices[k].fVertexIndex << " ";
      }
      outStream << endl;

      for (k = 0; k < n; ++k) {
        outStream << fFaces[i].fVertices[k].fNormalIndex << " ";
      }
      outStream << endl;
    }
  }
  else if (meshType == k3DV) {
    outStream << fNrOfVertices << endl;
    outStream << fNrOfFaces << endl;

    int i;
    for (i = 0; i < fNrOfVertices; ++i) {
      outStream << fVertices[i].fX << " "
                << fVertices[i].fY << " "
                << fVertices[i].fZ << " "
                << fNormals[i].fX << " "
                << fNormals[i].fY << " "
                << fNormals[i].fZ << endl;
    }

    for (i = 0; i < fNrOfFaces; ++i) {
      outStream << fFaces[i].fNrOfVertices << endl;

      int n = fFaces[i].fNrOfVertices;

      int k = 0;
      for (k = 0; k < n; ++k) {
        outStream << fFaces[i].fVertices[k].fVertexIndex << " ";
      }
      outStream << endl;
    }
  }
  else if (meshType == k3DM) {
    outStream << fNrOfVertices << endl;
    outStream << fNrOfFaces << endl;

    int i;
    for (i = 0; i < fNrOfVertices; ++i) {
      outStream << fVertices[i].fX << " "
                << fVertices[i].fY << " "
                << fVertices[i].fZ << endl;
    }

    for (i = 0; i < fNrOfFaces; ++i) {
      outStream << fFaces[i].fNrOfVertices << endl;

      int n = fFaces[i].fNrOfVertices;

      int k;
      for (k = 0; k < n; ++k) {
        outStream << fFaces[i].fVertices[k].fVertexIndex << " ";
      }
      outStream << endl;
    }
  }
  else if (meshType == kOBJ) {
    outStream << "Mesh::writeToFile() -> saving to Wavefront Object File (OBJ) "
              << "not yet supported." << endl;
  }

  outStream.close();
}

bool Mesh::isEmpty(void) const
{
  return ((fNrOfVertices == 0) ||
          (fNrOfNormals == 0) ||
          (fNrOfNormals == 0));
}

void Mesh::clear(void)
{
  delete[] fVertices;
  delete[] fNormals;

  int i;
  for (i = 0; i < fNrOfFaces; ++i) {
    delete[] fFaces[i].fVertices;
  }

  delete[] fFaces;

  for (i = 0; i < fNrOfTriangulatedFaces; ++i) {
    delete[] fTriangulatedFaces[i].fVertices;
  }

  delete[] fTriangulatedFaces;

  fVertices = 0;
  fNormals = 0;
  fFaces = 0;
  fTriangulatedFaces = 0;

  fNrOfVertices = 0;
  fNrOfNormals = 0;
  fNrOfFaces = 0;
  fNrOfTriangulatedFaces = 0;
}

void Mesh::triangulateFaces(void)
{
  cerr << "Triangulating mesh, ";
  
  fNrOfTriangulatedFaces = 0;
  int i;
  for (i = 0; i < fNrOfFaces; ++i) {
    // every face's n-polygon can be decomposed in n-2 triangles
    fNrOfTriangulatedFaces += (fFaces[i].fNrOfVertices - 2);
  }
  fTriangulatedFaces = new Face[fNrOfTriangulatedFaces];
  int lastTriangleProcessed = 0;

  // triangulate every face of the mesh
  for (i = 0; i < fNrOfFaces; ++i) {

    // find the normal-vector of the face's polygon using Martin Newell's method
    Vector3D normal(0.0,0.0,0.0);
    for (int j = 0; j < fFaces[i].fNrOfVertices; ++j) {
      int successor = (j + 1) % fFaces[i].fNrOfVertices;
      Point3D p1 = fVertices[fFaces[i].fVertices[j].fVertexIndex];
      Point3D p2 = fVertices[fFaces[i].fVertices[successor].fVertexIndex];

      normal.fX += (p1.fY - p2.fY) * (p1.fZ + p2.fZ);
      normal.fY += (p1.fZ - p2.fZ) * (p1.fX + p2.fX);
      normal.fZ += (p1.fX - p2.fX) * (p1.fY + p2.fY);
    }
    
    // create a 2D-polygon from the 3D-face-polygon
    // by projecting it's vertices onto the plane perpendicular to the
    // dominant axis of the normal-vector of the original polygon
    Point2D* projectedVertices = new Point3D[fFaces[i].fNrOfVertices];

    normal.fX = Math::abs(normal.fX);
    normal.fY = Math::abs(normal.fY);
    normal.fZ = Math::abs(normal.fZ);

    int k;

    if (((normal.fY <= normal.fZ) &&
         (normal.fZ <= normal.fX)) ||
        ((normal.fZ <= normal.fY) &&
         (normal.fY <= normal.fX))) {
      // project all the polygon's vertices onto the YZ-plane
      for (k = 0; k < fFaces[i].fNrOfVertices; ++k) {
        projectedVertices[k].fX = fVertices[fFaces[i].fVertices[k].fVertexIndex].fY;
        projectedVertices[k].fY = fVertices[fFaces[i].fVertices[k].fVertexIndex].fZ;
      }
    }
    else if (((normal.fX <= normal.fZ) &&
              (normal.fZ <= normal.fY)) ||
             ((normal.fZ <= normal.fX) &&
              (normal.fX <= normal.fY))) {
      // project all the polygon's vertices onto the XZ-plane
      for (k = 0; k < fFaces[i].fNrOfVertices; ++k) {
        projectedVertices[k].fX = fVertices[fFaces[i].fVertices[k].fVertexIndex].fZ;
        projectedVertices[k].fY = fVertices[fFaces[i].fVertices[k].fVertexIndex].fX;
      }
    }
    else {
      // project all the polygon's vertices onto the XY-plane
      for (k = 0; k < fFaces[i].fNrOfVertices; ++k) {
        projectedVertices[k].fX = fVertices[fFaces[i].fVertices[k].fVertexIndex].fX;
        projectedVertices[k].fY = fVertices[fFaces[i].fVertices[k].fVertexIndex].fY;
      }
    }

    // triangulate the resulting 2D-polygon
    int nrOfVerticesToProcess = fFaces[i].fNrOfVertices;

    int* currentPolygon = new int[nrOfVerticesToProcess];
    for (k = 0; k < nrOfVerticesToProcess; ++k) {
      currentPolygon[k] = k;
    }

    // the following triangulation-algorithm works great for polygons that
    // are counterclockwise specified (and not completely convex when clockwise
    // specified)
    while (nrOfVerticesToProcess >= 3) {

      int minimumIndex = 0;
      int previousIndex = 0;
      int nextIndex = 0;

      double smallestDiagonal = Math::kInfinity;

      for (k = 0; k < nrOfVerticesToProcess; ++k) {
        previousIndex = (k + nrOfVerticesToProcess - 1) % nrOfVerticesToProcess;
        nextIndex = (k + 1) % nrOfVerticesToProcess;
        if (isCounterClockWise(projectedVertices[currentPolygon[previousIndex]],
                               projectedVertices[currentPolygon[k]],
                               projectedVertices[currentPolygon[nextIndex]])) {
          double diagonal = lengthSquared(projectedVertices[currentPolygon[previousIndex]],
                                          projectedVertices[currentPolygon[nextIndex]]);
          if (diagonal < smallestDiagonal) {
            smallestDiagonal = diagonal;
            minimumIndex = k;
          }
        }
      }

      previousIndex = (minimumIndex + nrOfVerticesToProcess - 1) % nrOfVerticesToProcess;
      nextIndex = (minimumIndex + 1) % nrOfVerticesToProcess;

      // create the triangle
      fTriangulatedFaces[lastTriangleProcessed].fNrOfVertices = 3;
      fTriangulatedFaces[lastTriangleProcessed].fVertices = new VertexID[3];

      fTriangulatedFaces[lastTriangleProcessed].fVertices[0].fVertexIndex =
        fFaces[i].fVertices[currentPolygon[previousIndex]].fVertexIndex;
      fTriangulatedFaces[lastTriangleProcessed].fVertices[0].fNormalIndex =
        fFaces[i].fVertices[currentPolygon[previousIndex]].fNormalIndex;

      fTriangulatedFaces[lastTriangleProcessed].fVertices[1].fVertexIndex =
        fFaces[i].fVertices[currentPolygon[minimumIndex]].fVertexIndex;
      fTriangulatedFaces[lastTriangleProcessed].fVertices[1].fNormalIndex =
        fFaces[i].fVertices[currentPolygon[minimumIndex]].fNormalIndex;

      fTriangulatedFaces[lastTriangleProcessed].fVertices[2].fVertexIndex =
        fFaces[i].fVertices[currentPolygon[nextIndex]].fVertexIndex;
      fTriangulatedFaces[lastTriangleProcessed].fVertices[2].fNormalIndex =
        fFaces[i].fVertices[currentPolygon[nextIndex]].fNormalIndex;

      ++lastTriangleProcessed;

      --nrOfVerticesToProcess;
      for (k = minimumIndex; k < nrOfVerticesToProcess; ++k) {
         currentPolygon[k] =  currentPolygon[k + 1];
      }
    }
  }

  fIsTriangulated = true;

  cerr << "done (" << fNrOfTriangulatedFaces << " triangles generated)." << endl;
}

bool Mesh::isTriangulated(void) const
{
  return fIsTriangulated;
}

void Mesh::computeExtent(void)
{
  // create the point-cluster
  PointCluster pointCluster(fNrOfVertices);
  for (int i = 0; i < fNrOfVertices; ++i) {
    pointCluster.fPoints[i] = fVertices[i];
  }

  // apply the mesh's specific affine transformation
  // to express the point-cluster in world coordinates
  // instead of generic shape coordinates
  pointCluster.applyAffineTransformation(fAffineTransformation);

  fWorldExtent = new BoxExtent();
  // use a small excess for the extent (e.g. 0.1)
  fWorldExtent->create(pointCluster,0.1);
}

void Mesh::drawOpenGL(void) const
{
  if (isEmpty()) {
    return;
  }

  tellMaterialsGL(); 
  glPushMatrix();
    glMultMatrixd(fAffineTransformation.m); 

    for (int i = 0; i < fNrOfFaces; ++i) {
      int n = fFaces[i].fNrOfVertices;

      glBegin(GL_POLYGON);
        for (int v = 0; v < n; ++v) {

          int normalIndex = fFaces[i].fVertices[v].fNormalIndex;
          glNormal3d((GLdouble) fNormals[normalIndex].fX,
                     (GLdouble) fNormals[normalIndex].fY,
                     (GLdouble) fNormals[normalIndex].fZ);

          int vertexIndex = fFaces[i].fVertices[v].fVertexIndex;
          glVertex3d((GLdouble) fVertices[vertexIndex].fX,
                     (GLdouble) fVertices[vertexIndex].fY,
                     (GLdouble) fVertices[vertexIndex].fZ);
        }
      glEnd();
    }
  glPopMatrix();
}

bool Mesh::computeIntersectionsWithRay(const Ray& ray, Intersection& intersection)
{
  // perform extent-intersection-test (in world-coordinates !)
  if (!fWorldExtent->rayIntersectsExtent(ray)) {
    return false;
  }

  Ray genericRay;
  Transformations::transformRay(genericRay,fInverseAffineTransformation,ray);

  if (fIsTriangulated) {

    // find all intersections with all the triangles
    // using barycentric coordinates

    intersection.fNrOfHits = 0;
    for (int i = 0; i < fNrOfTriangulatedFaces; ++i) {

      Vector3D v1 = Vector3D(fVertices[fTriangulatedFaces[i].fVertices[1].fVertexIndex]) -
                    Vector3D(fVertices[fTriangulatedFaces[i].fVertices[0].fVertexIndex]);
      Vector3D v2 = Vector3D(fVertices[fTriangulatedFaces[i].fVertices[2].fVertexIndex]) -
                    Vector3D(fVertices[fTriangulatedFaces[i].fVertices[0].fVertexIndex]);
      Vector3D vp(genericRay.fDirection.outer(v2));

      double denominator = v1.inner(vp);
      // the generic ray is parallel to the triangle's plane
      if (Math::abs(denominator) < Math::kEpsilon) {
        continue;
      }

      denominator = 1.0 / denominator;

      Vector3D vt(Vector3D(genericRay.fStart) -
                  Vector3D(fVertices[fTriangulatedFaces[i].fVertices[0].fVertexIndex]));
      double u = (vt.inner(vp)) * denominator;
      if ((u < 0.0) ||
          (u > 1.0)) {
        // the generic ray is on or behind the triangle
        continue;
      }

      Vector3D vq(vt.outer(v1));
      double v = (genericRay.fDirection.inner(vq)) * denominator;
      if ((v < 0.0) ||
          ((u + v) > 1.0)) {
        // the intersection point doesn't lie within the triangle
        continue;
      }

      double hitTime = (v2.inner(vq)) * denominator;

      if (hitTime <= fMeshEpsilon) {
        continue;
      }

      Point3D p(Transformations::calcRayPosition(genericRay,hitTime));

      intersection.fHits[intersection.fNrOfHits].fHitTime = hitTime;
      intersection.fHits[intersection.fNrOfHits].fObjectHit = this;
      intersection.fHits[intersection.fNrOfHits].fSurface = i;
      intersection.fHits[intersection.fNrOfHits].fHitPoint = p;
      intersection.fHits[intersection.fNrOfHits].fBarycentricU = u;
      intersection.fHits[intersection.fNrOfHits].fBarycentricV = v;

      if (fInterpolateNormals) {
        // use Phong-shading by interpolating the normal with the use of the barycentric coordinates
        intersection.fHits[intersection.fNrOfHits].fHitNormal =
          (fNormals[fTriangulatedFaces[i].fVertices[0].fNormalIndex] * (1.0 - u - v)) +
          (fNormals[fTriangulatedFaces[i].fVertices[1].fNormalIndex] * u) +
          (fNormals[fTriangulatedFaces[i].fVertices[2].fNormalIndex] * v);
      }
      else {
        // use flat-shading with the use of the plane's polygon
        intersection.fHits[intersection.fNrOfHits].fHitNormal = v1.outer(v2);
      }

      ++intersection.fNrOfHits;
    }

    if (intersection.fNrOfHits > 0) {
      int j, k;

      // bubblesort the intersections
      // (the number of intersections is rather small, so bubblesorting isn't that time-consuming)      
      for (k = 0; k < (intersection.fNrOfHits - 1); ++k) {
        for (j = (intersection.fNrOfHits - 1); j > k; --j) {
          if (intersection.fHits[j - 1].fHitTime >
              intersection.fHits[j].fHitTime) {
            HitInfo temporary;
            temporary.set(intersection.fHits[j - 1]);
            intersection.fHits[j - 1].set(intersection.fHits[j]);
            intersection.fHits[j].set(temporary);
          }
        }
      }

      // adjust the inside-outside fields and find the first hit with a negative hit-time
      int negativeHitIndex = -1;
      for (j = 0; j < intersection.fNrOfHits; ++j) {
        intersection.fHits[j].fIsEntering = ((j % 2) == 0);
        if ((intersection.fHits[j].fHitTime < 0.0) &&
            (negativeHitIndex == -1)) {
          negativeHitIndex = j;
        }
      }

      // eliminate all hits with a negative hit-time
      if (negativeHitIndex != -1) {
        for (j = negativeHitIndex; j < (intersection.fNrOfHits - 1); ++j) {
          intersection.fHits[j - negativeHitIndex].set(intersection.fHits[j + 1]);
        }
        intersection.fNrOfHits -= (negativeHitIndex + 1);
      }

      return (intersection.fNrOfHits > 0);
    }
    else {
      return false;
    }
  }
  else {
    // this routine assumes a convex mesh

    double hitTime, numerator, denominator;
    double hitTimeIn = -Math::kInfinity;
    double hitTimeOut = Math::kInfinity;
    int innerSurface, outerSurface;

    for (int i = 0; i < fNrOfFaces; ++i) {

      Vector3D normal(fNormals[fFaces[i].fVertices[0].fNormalIndex]);
      Point3D vertex(fVertices[fFaces[i].fVertices[0].fVertexIndex]);

      Vector3D difference = Vector3D(vertex) - Vector3D(genericRay.fStart);
      numerator = Transformations::dot3D(normal,difference);
      denominator = Transformations::dot3D(normal,genericRay.fDirection);

      if (Math::abs(denominator) < Math::kEpsilon) {
        // ray is parallel to plane
        if (numerator < 0.0) {
          // ray is outside
          return false;
        }
      }
      else {
        hitTime = numerator / denominator;
        if (denominator > 0.0) {
          // ray is exiting
          if (hitTime < hitTimeOut) {
            // a new earlier exit
            hitTimeOut = hitTime;
            outerSurface = i;
          }
        }
        else {
          // ray is entering
          if (hitTime > hitTimeIn) {
            // a new later entrance
            hitTimeIn = hitTime;
            innerSurface = i;
          }
        }
      }

      if (hitTimeIn >= hitTimeOut) {
        // a miss - early out
        return false;
      }
    }

    intersection.fNrOfHits = 0;

    if (hitTimeIn > 0.0) { // a positive hit-time

      // find the hit-point
      Point3D p(Transformations::calcRayPosition(genericRay,hitTimeIn));

      intersection.fHits[0].fHitTime = hitTimeIn;
      intersection.fHits[0].fObjectHit = this;
      intersection.fHits[0].fSurface = innerSurface;
      intersection.fHits[0].fIsEntering = true;
      intersection.fHits[0].fHitPoint = p;
      intersection.fHits[0].fHitNormal = fNormals[fFaces[innerSurface].fVertices[0].fNormalIndex];
      ++intersection.fNrOfHits;
    }

    if (hitTimeOut > 0.0) { // a positive hit-time

      // find the hit-point
      Point3D p(Transformations::calcRayPosition(genericRay,hitTimeOut));

      intersection.fHits[intersection.fNrOfHits].fHitTime = hitTimeOut;
      intersection.fHits[intersection.fNrOfHits].fObjectHit = this;
      intersection.fHits[intersection.fNrOfHits].fSurface = outerSurface;
      intersection.fHits[intersection.fNrOfHits].fIsEntering = false;
      intersection.fHits[intersection.fNrOfHits].fHitPoint = p;
      intersection.fHits[intersection.fNrOfHits].fHitNormal = fNormals[fFaces[outerSurface].fVertices[0].fNormalIndex];
      ++intersection.fNrOfHits;
    }

    return (intersection.fNrOfHits > 0);
  }
}

void Mesh::textureLookup(const HitInfo& hitInfo, double& s, double& t) const
{
  Point3D p(hitInfo.fHitPoint);

  // shrink-wrap the texture
  double radius, phi, theta;
  Math::convertCartesianToSpherical(p.fX,p.fY,p.fZ,
                                    radius,phi,theta);

  s = phi / (Math::kPi * 2.0);
  t = theta / Math::kPi;
}

// ---------------
// utility methods
// ---------------
bool Mesh::isCounterClockWise(const Point2D& p1, const Point2D& p2, const Point2D& p3) const
{
  double x2Minusx1 = p2.fX - p1.fX;
  double y2Minusy1 = p2.fY - p1.fY;
  double x3Minusx1 = p3.fX - p1.fX;
  double y3Minusy1 = p3.fY - p1.fY;

  double determinant = (x2Minusx1 * y3Minusy1) - (y2Minusy1 * x3Minusx1);

  return (determinant >= 0.0);
}

double Mesh::lengthSquared(const Point2D& p1, const Point2D& p2) const
{
  double x2Minusx1 = p2.fX - p1.fX;
  double y2Minusy1 = p2.fY - p1.fY;
  return ((x2Minusx1 * x2Minusx1) + (y2Minusy1 * y2Minusy1));
}

// ------------
// class Sphere
// ------------

// -----------
// constructor
// -----------
Sphere::Sphere(void)
{
}

// ----------
// destructor
// ----------
Sphere::~Sphere(void)
{
}

// -------
// methods
// -------
void Sphere::drawOpenGL(void) const
{
  tellMaterialsGL();
  glPushMatrix();
    glMultMatrixd(fAffineTransformation.m);
    glutSolidSphere((GLdouble) 1.0,
                    (GLint) kSphereSlices,
                    (GLint) kSphereStacks);
    glPopMatrix();
}

bool Sphere::computeIntersectionsWithRay(const Ray& ray, Intersection& intersection)
{
  // there's no need to perform extent-testing
  Ray genericRay;
  Transformations::transformRay(genericRay,fInverseAffineTransformation,ray);

  double A = Transformations::dot3D(genericRay.fDirection,genericRay.fDirection);
  if (A <= 0.0) {
    return false;
  }
  double B = Transformations::dot3D(genericRay.fStart,genericRay.fDirection);
  double C = Transformations::dot3D(genericRay.fStart,genericRay.fStart) - 1.0;

  double discriminant = (B * B) - (A * C);
  if (discriminant < 0.0) {
    return false;
  }

  double sqrtDiscrimant = sqrt(discriminant);
  intersection.fNrOfHits = 0;

  // earlier hit
  double t1 = (-B - sqrtDiscrimant) / A;
  if (t1 > 0.0) { // a positive hit-time

    // find the hit-point
    Point3D p(Transformations::calcRayPosition(genericRay,t1));

    intersection.fHits[0].fHitTime = t1;
    intersection.fHits[0].fObjectHit = this;
    intersection.fHits[0].fSurface = 0;
    intersection.fHits[0].fIsEntering = true;
    intersection.fHits[0].fHitPoint = p;
    intersection.fHits[0].fHitNormal.set(p.fX,p.fY,p.fZ);
    ++intersection.fNrOfHits;
  }

  // second hit
  double t2 = (-B + sqrtDiscrimant) / A;
  if (t2 > 0.0) { // a positive hit-time

    // find the hit-point
    Point3D p(Transformations::calcRayPosition(genericRay,t2));

    intersection.fHits[intersection.fNrOfHits].fHitTime = t2;
    intersection.fHits[intersection.fNrOfHits].fObjectHit = this;
    intersection.fHits[intersection.fNrOfHits].fSurface = 0;
    intersection.fHits[intersection.fNrOfHits].fIsEntering = false;
    intersection.fHits[intersection.fNrOfHits].fHitPoint = p;
    intersection.fHits[intersection.fNrOfHits].fHitNormal.set(p.fX,p.fY,p.fZ);

    ++intersection.fNrOfHits;
  }

  return (intersection.fNrOfHits > 0);
}

void Sphere::textureLookup(const HitInfo& hitInfo, double& s, double& t) const
{
  Point3D p(hitInfo.fHitPoint);

  double radius, phi, theta;
  Math::convertCartesianToSpherical(p.fX,p.fY,p.fZ,
                                    radius,phi,theta);

  s = phi / (Math::kPi * 2.0);
  t = 1.0 - (theta / Math::kPi);
}

// ---------------------
// class TaperedCylinder
// ---------------------

// -----------
// constructor
// -----------
TaperedCylinder::TaperedCylinder(void) : fSmallRadius(0.0),
                                         fDistortTexture(true)
{
}

// ----------
// destructor
// ----------
TaperedCylinder::~TaperedCylinder(void)
{
}

// -------
// methods
// -------
void TaperedCylinder::drawOpenGL(void) const
{
  tellMaterialsGL();
  glPushMatrix();
    glMultMatrixd(fAffineTransformation.m);
    GLUquadricObj* quadricObject = gluNewQuadric();
    gluQuadricDrawStyle(quadricObject,(GLenum) GLU_FILL);
    gluCylinder(quadricObject,
                (GLdouble) 1.0,
                (GLdouble) fSmallRadius,
                (GLdouble) 1.0,
                (GLint) kTaperedCylinderSlices,
                (GLint) kTaperedCylinderStacks);
  glPopMatrix();
}

bool TaperedCylinder::computeIntersectionsWithRay(const Ray& ray, Intersection& intersection)
{
  // there's no need to perform extent-testing
  Ray genericRay;
  Transformations::transformRay(genericRay,fInverseAffineTransformation,ray);

  double smallRadius = fSmallRadius - 1.0;
  double direction = smallRadius * genericRay.fDirection.fZ;
  double start = smallRadius * genericRay.fStart.fZ + 1.0;

  double A = (genericRay.fDirection.fX * genericRay.fDirection.fX) +
             (genericRay.fDirection.fY * genericRay.fDirection.fY) -
             (direction * direction);
  double B = (genericRay.fStart.fX * genericRay.fDirection.fX) +
             (genericRay.fStart.fY * genericRay.fDirection.fY) -
             (start * direction);
  double C = (genericRay.fStart.fX * genericRay.fStart.fX) +
             (genericRay.fStart.fY * genericRay.fStart.fY) -
             (start * start);

  double discriminant = (B * B) - (A * C);

  intersection.fNrOfHits = 0;

  if (discriminant > 0.0) {
    double sqrtDiscriminant = sqrt(discriminant);
    double t1 = (-B - sqrtDiscriminant) / A; // earlier hit
    double zHit = genericRay.fStart.fZ + (genericRay.fDirection.fZ * t1);
    if ((t1 > Math::kEpsilon) &&
        (zHit <= 1.0) &&
        (zHit >= 0.0)) {

      intersection.fHits[intersection.fNrOfHits].fHitTime = t1;
      intersection.fHits[intersection.fNrOfHits].fSurface = 0; // the wall was hit
      ++intersection.fNrOfHits;
    }
    
    double t2 = (-B + sqrtDiscriminant) / A; // earlier hit
    zHit = genericRay.fStart.fZ + (genericRay.fDirection.fZ * t2);
    if ((t2 > Math::kEpsilon) &&
        (zHit <= 1.0) &&
        (zHit >= 0.0)) {

      intersection.fHits[intersection.fNrOfHits].fHitTime = t2;
      intersection.fHits[intersection.fNrOfHits].fSurface = 0; // the wall was hit
      ++intersection.fNrOfHits;
    }
  }

  // test the base at z = 0
  double tB = -genericRay.fStart.fZ / genericRay.fDirection.fZ; // the hit-time at the z = 0 plane
  if ((tB > Math::kEpsilon) &&
       ((Math::sqr(genericRay.fStart.fX + (genericRay.fDirection.fX * tB)) +
         Math::sqr(genericRay.fStart.fY + (genericRay.fDirection.fY * tB))) < 1.0)) {

    intersection.fHits[intersection.fNrOfHits].fHitTime = tB;
    intersection.fHits[intersection.fNrOfHits].fSurface = 1; // the base was hit
    ++intersection.fNrOfHits;
  }

  // test the cap at z = 1
  double tC = (1.0 - genericRay.fStart.fZ) / genericRay.fDirection.fZ; // the hit-time at the z = 1 plane
  if ((tC > Math::kEpsilon) &&
       ((Math::sqr(genericRay.fStart.fX + (genericRay.fDirection.fX * tC)) +
         Math::sqr(genericRay.fStart.fY + (genericRay.fDirection.fY * tC))) < Math::sqr(fSmallRadius))) {

    intersection.fHits[intersection.fNrOfHits].fHitTime = tC;
    intersection.fHits[intersection.fNrOfHits].fSurface = 2; // the cap was hit
    ++intersection.fNrOfHits;
  }

  if (intersection.fNrOfHits == 0) {
    // the tapered cylinder was missed (or lies behind the eye)
    return false;
  }

  if (intersection.fNrOfHits == 1) {
    intersection.fHits[0].fIsEntering = false;
    intersection.fHits[0].fObjectHit = this;
  }
  else {
    // sort the two hits
    if (intersection.fHits[0].fHitTime > intersection.fHits[1].fHitTime) {

      // only swap the hit-time- and surface-fields
      double tempt = intersection.fHits[0].fHitTime;
      intersection.fHits[0].fHitTime = intersection.fHits[1].fHitTime;
      intersection.fHits[1].fHitTime = tempt;

      int temps = intersection.fHits[0].fSurface;
      intersection.fHits[0].fSurface = intersection.fHits[1].fSurface;
      intersection.fHits[1].fSurface = temps;
    }

    intersection.fHits[0].fIsEntering = true;
    intersection.fHits[0].fObjectHit = this;
    intersection.fHits[1].fIsEntering = false;
    intersection.fHits[1].fObjectHit = this;
  }
    
  // set the hit-point and normal at the hit-point
  for (int i = 0; i < intersection.fNrOfHits; ++i) {

    // find the hit-point
    Point3D p(Transformations::calcRayPosition(genericRay,intersection.fHits[i].fHitTime));
    intersection.fHits[i].fHitPoint = p;

    // set the normal
    if (intersection.fHits[i].fSurface == 0) {
      // at the wall
      intersection.fHits[i].fHitNormal.set(p.fX,p.fY,-smallRadius * (1 + (smallRadius * p.fZ)));
    }
    else if (intersection.fHits[i].fSurface == 1) {
      // at the base
      intersection.fHits[i].fHitNormal.set(0.0,0.0,-1.0);
    }
    else {
      // at the cap
      intersection.fHits[i].fHitNormal.set(0.0,0.0,1.0);
    }
  }

  return true;
}

void TaperedCylinder::textureLookup(const HitInfo& hitInfo, double& s, double& t) const
{
  Point3D p(hitInfo.fHitPoint);

  if (fDistortTexture) {
    if (Math::abs(p.fZ - 1.0) < Math::kEpsilon) {
      // handle the front of the tapered cylinder
      // map the circle to a square
      double alpha = Math::arctan(p.fY,p.fX);
      double radius = sqrt(Math::sqr(p.fX) + Math::sqr(p.fY));

      double beta = ((2.0 * floor(alpha / (Math::kPi / 2.0))) + 1.0) * (Math::kPi / 4.0);
      double alphaPrime = -Math::abs(alpha - beta) + beta;

      if (alpha < (Math::kPi / 2.0)) {
        s = cos(alpha) / cos(alphaPrime);
        t = sin(alpha) / cos(alphaPrime);
      }
      else if (alpha < Math::kPi) {
        s = cos(alpha) / sin(alphaPrime);
        t = sin(alpha) / sin(alphaPrime);
      }
      else if (alpha < (3.0 * (Math::kPi / 2.0))) {
        s = -cos(alpha) / cos(alphaPrime);
        t = -sin(alpha) / cos(alphaPrime);
      }
      else {
        s = -cos(alpha) / sin(alphaPrime);
        t = -sin(alpha) / sin(alphaPrime);
      }

      s = (1.0 + ((radius / fSmallRadius) * s)) / 2.0;
      t = (1.0 + ((radius / fSmallRadius) * t)) / 2.0;
    }
    else if (Math::abs(p.fZ) < Math::kEpsilon) {
      // handle the back of the tapered cylinder (radius is 1)
      // map the circle to a square
      double alpha = Math::arctan(p.fY,p.fX);
      double radius = sqrt(Math::sqr(p.fX) + Math::sqr(p.fY));

      double beta = ((2.0 * floor(alpha / (Math::kPi / 2.0))) + 1.0) * (Math::kPi / 4.0);
      double alphaPrime = -Math::abs(alpha - beta) + beta;

      if (alpha < (Math::kPi / 2.0)) {
        s = cos(alpha) / cos(alphaPrime);
        t = sin(alpha) / cos(alphaPrime);
      }
      else if (alpha < Math::kPi) {
        s = cos(alpha) / sin(alphaPrime);
        t = sin(alpha) / sin(alphaPrime);
      }
      else if (alpha < (3.0 * (Math::kPi / 2.0))) {
        s = -cos(alpha) / cos(alphaPrime);
        t = -sin(alpha) / cos(alphaPrime);
      }
      else {
        s = -cos(alpha) / sin(alphaPrime);
        t = -sin(alpha) / sin(alphaPrime);
      }

      s = (1.0 + (radius * (-s))) / 2.0;
      t = (1.0 + (radius * t)) / 2.0;
    }
    else {
      // handle the wall of the tapered cylinder
      s = Math::arctan(p.fY,p.fX) / (2.0 * Math::kPi);
      t = p.fZ;
    }
  }
  else {
    // clip excess texture-parts
    if (Math::abs(p.fZ - 1.0) < Math::kEpsilon) {
      // handle the front of the tapered cylinder
      s = (p.fX + fSmallRadius) / (2.0 * fSmallRadius);
      t = (p.fY + fSmallRadius) / (2.0 * fSmallRadius);
    }
    else if (Math::abs(p.fZ) < Math::kEpsilon) {
      // handle the back of the tapered cylinder (radius is 1)
      s = (1.0 - p.fX) / 2.0;
      t = (p.fY + 1.0) / 2.0;
    }
    else {
      // handle the wall of the tapered cylinder
      s = Math::arctan(p.fY,p.fX) / (2.0 * Math::kPi);
      t = p.fZ;
    }
  }
}

// -----------
// class Torus
// -----------

// -----------
// constructor
// -----------
Torus::Torus(void) : fTubeRadius(0.5),
                     fTorusRadius(1.0)
{
}

// ----------
// destructor
// ----------
Torus::~Torus(void)
{
}

// -------
// methods
// -------
void Torus::computeExtent(void)
{
  // create the point-cluster
  PointCluster pointCluster(8);
  pointCluster.fPoints[0] = Point3D(fTorusRadius + fTubeRadius,
                                    -(fTorusRadius + fTubeRadius),
                                    -fTubeRadius);
  pointCluster.fPoints[1] = Point3D(fTorusRadius + fTubeRadius,
                                    fTorusRadius + fTubeRadius,
                                    -fTubeRadius);
  pointCluster.fPoints[2] = Point3D(-(fTorusRadius + fTubeRadius),
                                    fTorusRadius + fTubeRadius,
                                    -fTubeRadius);
  pointCluster.fPoints[3] = Point3D(-(fTorusRadius + fTubeRadius),
                                    -(fTorusRadius + fTubeRadius),
                                    -fTubeRadius);
  
  pointCluster.fPoints[4] = Point3D(fTorusRadius + fTubeRadius,
                                    -(fTorusRadius + fTubeRadius),
                                    fTubeRadius);
  pointCluster.fPoints[5] = Point3D(fTorusRadius + fTubeRadius,
                                    fTorusRadius + fTubeRadius,
                                    fTubeRadius);
  pointCluster.fPoints[6] = Point3D(-(fTorusRadius + fTubeRadius),
                                    fTorusRadius + fTubeRadius,
                                    fTubeRadius);
  pointCluster.fPoints[7] = Point3D(-(fTorusRadius + fTubeRadius),
                                    -(fTorusRadius + fTubeRadius),
                                    fTubeRadius);
  
  // apply the torus's specific affine transformation
  // to express the point-cluster in world coordinates
  // instead of generic shape coordinates
  pointCluster.applyAffineTransformation(fAffineTransformation);

  fWorldExtent = new BoxExtent();
  // use a small excess for the extent
  fWorldExtent->create(pointCluster,0.1);
}

void Torus::drawOpenGL(void) const
{
  tellMaterialsGL();
  glPushMatrix();
    glMultMatrixd(fAffineTransformation.m);
    glutSolidTorus((GLdouble) fTubeRadius,
                   (GLdouble) fTorusRadius,
                   (GLint) kTorusSlices,
                   (GLint) kTorusStacks);
  glPopMatrix();
}

bool Torus::computeIntersectionsWithRay(const Ray& ray, Intersection& intersection)
{
  // perform extent-intersection-test (in world-coordinates !)
  if (!fWorldExtent->rayIntersectsExtent(ray)) {
    return false;
  }

  Ray genericRay;
  Transformations::transformRay(genericRay,fInverseAffineTransformation,ray);

  double alpha = Math::sqr(genericRay.fDirection.fX) +
                 Math::sqr(genericRay.fDirection.fY) +
                 Math::sqr(genericRay.fDirection.fZ);
  double beta = 2.0 * ((genericRay.fStart.fX * genericRay.fDirection.fX) +
                       (genericRay.fStart.fY * genericRay.fDirection.fY) +
                       (genericRay.fStart.fZ * genericRay.fDirection.fZ));
  double gamma = Math::sqr(genericRay.fStart.fX) +
                 Math::sqr(genericRay.fStart.fY) +
                 Math::sqr(genericRay.fStart.fZ) +
                 Math::sqr(fTorusRadius) -
                 Math::sqr(fTubeRadius);
  double delta = 4.0 * Math::sqr(fTorusRadius) *
                 (Math::sqr(genericRay.fDirection.fX) + Math::sqr(genericRay.fDirection.fY));
  double epsilon = 8.0 * Math::sqr(fTorusRadius) *
                   ((genericRay.fStart.fX * genericRay.fDirection.fX) +
                    (genericRay.fStart.fY * genericRay.fDirection.fY));
  double mu = 4.0 * Math::sqr(fTorusRadius) *
              (Math::sqr(genericRay.fStart.fX) + Math::sqr(genericRay.fStart.fY));

  double coefficients[5] = {Math::sqr(gamma) - mu,
                            (2.0 * beta * gamma) - epsilon,
                            (2.0 * alpha * gamma) + Math::sqr(beta) - delta,
                            2.0 * alpha * beta,
                            Math::sqr(alpha)};

  double solutions[4];
  int nrOfSolutions = Math::solveQuarticEquation(coefficients,solutions);

  int j, k;

  // bubblesort the hit-times
  // (the number of intersections is rather small, so bubblesorting isn't that time-consuming)      
  for (k = 0; k < (nrOfSolutions - 1); ++k) {
    for (j = (nrOfSolutions - 1); j > k; --j) {
      if (solutions[j - 1] > solutions[j]) {
        double temporary = solutions[j - 1];
        solutions[j - 1] = solutions[j];
        solutions[j] = temporary;
      }
    }
  }

  // adjust the inside-outside fields and find the first hit with a negative hit-time
  int negativeHitIndex = -1;
  for (j = 0; j < nrOfSolutions; ++j) {
    intersection.fHits[j].fIsEntering = ((j % 2) == 0);
    if ((solutions[j] < 0.0) &&
        (negativeHitIndex == -1)) {
      negativeHitIndex = j;
    }
  }

  // eliminate all hits with a negative hit-time
  if (negativeHitIndex != -1) {
    for (j = negativeHitIndex; j < (nrOfSolutions - 1); ++j) {
      solutions[j - negativeHitIndex] = solutions[j + 1];
    }
    nrOfSolutions -= (negativeHitIndex + 1);
  }

  intersection.fNrOfHits = 0;

  for (int i = 0; i < nrOfSolutions; ++i) {

    if (solutions[i] > 0.0) { // a positive hit-time
      // find the hit-point
      Point3D p(Transformations::calcRayPosition(genericRay,solutions[i]));

      intersection.fHits[intersection.fNrOfHits].fHitTime = solutions[i];
      intersection.fHits[intersection.fNrOfHits].fObjectHit = this;
      intersection.fHits[intersection.fNrOfHits].fSurface = 0;
      intersection.fHits[intersection.fNrOfHits].fHitPoint = p;
      intersection.fHits[intersection.fNrOfHits].fHitNormal = calcNormal(p);
      intersection.fHits[intersection.fNrOfHits].fHitNormal.normalize();

      ++intersection.fNrOfHits;
    }
  }

  return (intersection.fNrOfHits > 0);
}

void Torus::textureLookup(const HitInfo& hitInfo, double& s, double& t) const
{
  Point3D p(hitInfo.fHitPoint);

  s = Math::arctan(p.fY,p.fX) / (2.0 * Math::kPi);

  double cosPhi = -(sqrt(Math::sqr(p.fX) + Math::sqr(p.fY)) - fTorusRadius) / fTubeRadius;

  if (cosPhi > 1.0) {
    t = 0.0;
  }
  else if (cosPhi < -1.0) {
    t = 0.5;
  }
  else {
    t = acos(cosPhi) / (2.0 * Math::kPi);
  }

  if (p.fZ > 0.0) {
    t = 1.0 - t;
  }
}

// ------------------
// utlitity functions
// ------------------
Vector3D Torus::calcNormal(const Point3D& p) const
{
  if ((Math::abs(p.fX) < Math::kEpsilon) &&
      (Math::abs(p.fY) < Math::kEpsilon)) {
    return Vector3D(0.0,0.0,1.0);
  }
  else {
    double squareRoot = sqrt(Math::sqr(p.fX) + Math::sqr(p.fY));
    double partialDerivativeToX = (2.0 * p.fX * (squareRoot - fTorusRadius)) / squareRoot;
    double partialDerivativeToY = (2.0 * p.fY * (squareRoot - fTorusRadius)) / squareRoot;
    double partialDerivativeToZ = 2.0 * p.fZ;

    return Vector3D(partialDerivativeToX,partialDerivativeToY,partialDerivativeToZ);
  }
}

// ----------------------
// class AlgebraicSurface
// ----------------------

// -----------
// constructor
// -----------
AlgebraicSurface::AlgebraicSurface(void) : fBoxExtentWidth(1.0),
                                           fBoxExtentHeight(1.0),
                                           fBoxExtentDepth(1.0)
{
  for (int i = 0; i < (kMaxAlgebraicSurfaceDegree + 1); ++i) {
    fCoefficients[i] = 0.0;
  }
}

// ----------
// destructor
// ----------
AlgebraicSurface::~AlgebraicSurface(void)
{
}

// -------
// methods
// -------
void AlgebraicSurface::computeExtent(void)
{
  // create the point-cluster
  PointCluster pointCluster(2);
  pointCluster.fPoints[0] = Point3D(-fBoxExtentWidth / 2.0,
                                    -fBoxExtentHeight / 2.0,
                                    -fBoxExtentDepth / 2.0);
  pointCluster.fPoints[1] = Point3D(fBoxExtentWidth / 2.0,
                                    fBoxExtentHeight / 2.0,
                                    fBoxExtentDepth / 2.0);

  // apply the translation-part of the algebraic surface's  affine transformation
  // to express the point-cluster in world coordinates
  // instead of generic shape coordinates
  pointCluster.applyTranslation(fAffineTransformation);

  fWorldExtent = new BoxExtent();
  // don't use an excess
  fWorldExtent->create(pointCluster,0.0);
}

void AlgebraicSurface::drawOpenGL(void) const
{
}

bool AlgebraicSurface::computeIntersectionsWithRay(const Ray& ray, Intersection& intersection)
{
  // perform extent-intersection-test in world-coordinates
  if (!fWorldExtent->rayIntersectsExtent(ray)) {
    return false;
  }

  Ray genericRay;
  Transformations::transformRay(genericRay,fInverseAffineTransformation,ray);

  setCoefficients(genericRay);

  double solutions[4];
  int nrOfSolutions = 0;

  // determine the degree of the polynomial
  if (fCoefficients[4] != 0.0) {
    // solve quartic equation
    nrOfSolutions = Math::solveQuarticEquation(fCoefficients,solutions);
  }
  else if (fCoefficients[3] != 0.0) {
    // solve cubic equation
    nrOfSolutions = Math::solveCubicEquation(fCoefficients,solutions);
  }
  else {
    // solve quadratic equation
    nrOfSolutions = Math::solveQuadraticEquation(fCoefficients,solutions);
  }

  int j, k;

  // bubblesort the hit-times
  // (the number of intersections is rather small, so bubblesorting isn't that time-consuming)      
  for (k = 0; k < (nrOfSolutions - 1); ++k) {
    for (j = (nrOfSolutions - 1); j > k; --j) {
      if (solutions[j - 1] > solutions[j]) {
        double temporary = solutions[j - 1];
        solutions[j - 1] = solutions[j];
        solutions[j] = temporary;
      }
    }
  }

  // adjust the inside-outside fields and find the first hit with a negative hit-time
  int negativeHitIndex = -1;
  for (j = 0; j < nrOfSolutions; ++j) {
    intersection.fHits[j].fIsEntering = ((j % 2) == 0);
    if ((solutions[j] < 0.0) &&
        (negativeHitIndex == -1)) {
      negativeHitIndex = j;
    }
  }

  // eliminate all hits with a negative hit-time
  if (negativeHitIndex != -1) {
    for (j = negativeHitIndex; j < (nrOfSolutions - 1); ++j) {
      solutions[j - negativeHitIndex] = solutions[j + 1];
    }
    nrOfSolutions -= (negativeHitIndex + 1);
  }

  intersection.fNrOfHits = 0;

  for (int i = 0; i < nrOfSolutions; ++i) {

    if (solutions[i] > 0.0) { // a positive hit-time
      // find the hit-point
      Point3D p(Transformations::calcRayPosition(genericRay,solutions[i]));

      intersection.fHits[intersection.fNrOfHits].fHitTime = solutions[i];
      intersection.fHits[intersection.fNrOfHits].fObjectHit = this;
      intersection.fHits[intersection.fNrOfHits].fSurface = 0;
      intersection.fHits[intersection.fNrOfHits].fHitPoint = p;
      intersection.fHits[intersection.fNrOfHits].fHitNormal = calcNormal(p);
      intersection.fHits[intersection.fNrOfHits].fHitNormal.normalize();

      ++intersection.fNrOfHits;
    }
  }

  return (intersection.fNrOfHits > 0);
}

void AlgebraicSurface::textureLookup(const HitInfo& hitInfo, double& s, double& t) const
{
  Point3D p(hitInfo.fHitPoint);

  // shrink-wrap the texture
  s = Math::arctan(p.fY,p.fX) / (2.0 * Math::kPi);
  t = 1.0 - ((p.fZ + (fBoxExtentDepth / 2.0)) / fBoxExtentDepth);
}

// ------------------------------------
// algebraic surface specific functions
// ------------------------------------
void AlgebraicSurface::setCoefficients(const Ray& genericRay)
{
  // note that the coefficients are used as follows :
  //   c[0] +
  //   c[1] * t +
  //   c[2] * t^2 +
  //   c[3] * t^3 +
  //   ...
  // to distinguish between the different degrees of the
  // algebraic surfaces, use 0.0 as the highest coefficients
}

Vector3D AlgebraicSurface::calcNormal(const Point3D& p) const
{
  return Vector3D(1.0,1.0,1.0);
}

// -------------------------
// class OneSheetHyperboloid
// -------------------------

// -----------
// constructor
// -----------
OneSheetHyperboloid::OneSheetHyperboloid(void)
{
}

// ----------
// destructor
// ----------
OneSheetHyperboloid::~OneSheetHyperboloid(void)
{
}

// ----------------------------------
// quadratic surface specific methods
// ----------------------------------
void OneSheetHyperboloid::setCoefficients(const Ray& genericRay)
{
  fCoefficients[0] = Math::sqr(genericRay.fStart.fX) +
                     Math::sqr(genericRay.fStart.fY) -
                     Math::sqr(genericRay.fStart.fZ) - 1.0;

  fCoefficients[1] = 2.0 * ((genericRay.fStart.fX * genericRay.fDirection.fX) +
                            (genericRay.fStart.fY * genericRay.fDirection.fY) -
                            (genericRay.fStart.fZ * genericRay.fDirection.fZ));

  fCoefficients[2] = Math::sqr(genericRay.fDirection.fX) +
                     Math::sqr(genericRay.fDirection.fY) -
                     Math::sqr(genericRay.fDirection.fZ);

  fCoefficients[3] = 0.0;
  fCoefficients[4] = 0.0;
}

Vector3D OneSheetHyperboloid::calcNormal(const Point3D& p) const
{
  return Vector3D( 2.0 * p.fX,
                   2.0 * p.fY,
                  -2.0 * p.fZ);
}

// -------------------------
// class TwoSheetHyperboloid
// -------------------------

// -----------
// constructor
// -----------
TwoSheetHyperboloid::TwoSheetHyperboloid(void)
{
}

// ----------
// destructor
// ----------
TwoSheetHyperboloid::~TwoSheetHyperboloid(void)
{
}

// ----------------------------------
// quadratic surface specific methods
// ----------------------------------
void TwoSheetHyperboloid::setCoefficients(const Ray& genericRay)
{
  fCoefficients[0] = Math::sqr(genericRay.fStart.fX) -
                     Math::sqr(genericRay.fStart.fY) -
                     Math::sqr(genericRay.fStart.fZ) - 1.0;

  fCoefficients[1] = 2.0 * ((genericRay.fStart.fX * genericRay.fDirection.fX) -
                            (genericRay.fStart.fY * genericRay.fDirection.fY) -
                            (genericRay.fStart.fZ * genericRay.fDirection.fZ));

  fCoefficients[2] = Math::sqr(genericRay.fDirection.fX) -
                     Math::sqr(genericRay.fDirection.fY) -
                     Math::sqr(genericRay.fDirection.fZ);

  fCoefficients[3] = 0.0;
  fCoefficients[4] = 0.0;
}

Vector3D TwoSheetHyperboloid::calcNormal(const Point3D& p) const
{
  return Vector3D( 2.0 * p.fX,
                   2.0 * p.fY,
                  -2.0 * p.fZ);
}

// ----------------------
// class EllipticCylinder
// ----------------------

// -----------
// constructor
// -----------
EllipticCylinder::EllipticCylinder(void)
{
}

// ----------
// destructor
// ----------
EllipticCylinder::~EllipticCylinder(void)
{
}

// ----------------------------------
// quadratic surface specific methods
// ----------------------------------
void EllipticCylinder::setCoefficients(const Ray& genericRay)
{
  fCoefficients[0] = Math::sqr(genericRay.fStart.fX) +
                     Math::sqr(genericRay.fStart.fY) - 1.0;

  fCoefficients[1] = 2.0 * ((genericRay.fStart.fX * genericRay.fDirection.fX) +
                            (genericRay.fStart.fY * genericRay.fDirection.fY));

  fCoefficients[2] = Math::sqr(genericRay.fDirection.fX) +
                     Math::sqr(genericRay.fDirection.fY);

  fCoefficients[3] = 0.0;
  fCoefficients[4] = 0.0;
}

Vector3D EllipticCylinder::calcNormal(const Point3D& p) const
{
  return Vector3D(2.0 * p.fX,
                  2.0 * p.fY,
                  0.0);
}

// ------------------
// class EllipticCone
// ------------------

// -----------
// constructor
// -----------
EllipticCone::EllipticCone(void)
{
}

// ----------
// destructor
// ----------
EllipticCone::~EllipticCone(void)
{
}

// ----------------------------------
// quadratic surface specific methods
// ----------------------------------
void EllipticCone::setCoefficients(const Ray& genericRay)
{
  fCoefficients[0] = Math::sqr(genericRay.fStart.fX) +
                     Math::sqr(genericRay.fStart.fY) -
                     Math::sqr(genericRay.fStart.fZ);

  fCoefficients[1] = 2.0 * ((genericRay.fStart.fX * genericRay.fDirection.fX) +
                            (genericRay.fStart.fY * genericRay.fDirection.fY) -
                            (genericRay.fStart.fZ * genericRay.fDirection.fZ));

  fCoefficients[2] = Math::sqr(genericRay.fDirection.fX) +
                     Math::sqr(genericRay.fDirection.fY) -
                     Math::sqr(genericRay.fDirection.fZ);

  fCoefficients[3] = 0.0;
  fCoefficients[4] = 0.0;
}

Vector3D EllipticCone::calcNormal(const Point3D& p) const
{
  return Vector3D( 2.0 * p.fX,
                   2.0 * p.fY,
                  -2.0 * p.fZ);
}

// ------------------------
// class EllipticParaboloid
// ------------------------

// -----------
// constructor
// -----------
EllipticParaboloid::EllipticParaboloid(void)
{
}

// ----------
// destructor
// ----------
EllipticParaboloid::~EllipticParaboloid(void)
{
}

// ----------------------------------
// quadratic surface specific methods
// ----------------------------------
void EllipticParaboloid::setCoefficients(const Ray& genericRay)
{
  fCoefficients[0] = Math::sqr(genericRay.fStart.fX) +
                     Math::sqr(genericRay.fStart.fY) -
                     genericRay.fStart.fZ;

  fCoefficients[1] = (2.0 * ((genericRay.fStart.fX * genericRay.fDirection.fX) +
                             (genericRay.fStart.fY * genericRay.fDirection.fY))) -
                     genericRay.fDirection.fZ;

  fCoefficients[2] = Math::sqr(genericRay.fDirection.fX) +
                     Math::sqr(genericRay.fDirection.fY);

  fCoefficients[3] = 0.0;
  fCoefficients[4] = 0.0;
}

Vector3D EllipticParaboloid::calcNormal(const Point3D& p) const
{
  return Vector3D( 2.0 * p.fX,
                   2.0 * p.fY,
                  -1.0);
}

// --------------------------
// class HyperbolicParaboloid
// --------------------------

// -----------
// constructor
// -----------
HyperbolicParaboloid::HyperbolicParaboloid(void)
{
}

// ----------
// destructor
// ----------
HyperbolicParaboloid::~HyperbolicParaboloid(void)
{
}

// ----------------------------------
// quadratic surface specific methods
// ----------------------------------
void HyperbolicParaboloid::setCoefficients(const Ray& genericRay)
{
  fCoefficients[0] = Math::sqr(genericRay.fStart.fX) -
                     Math::sqr(genericRay.fStart.fY) -
                     genericRay.fStart.fZ;

  fCoefficients[1] = (2.0 * ((genericRay.fStart.fX * genericRay.fDirection.fX) -
                             (genericRay.fStart.fY * genericRay.fDirection.fY))) -
                     genericRay.fDirection.fZ;

  fCoefficients[2] = Math::sqr(genericRay.fDirection.fX) -
                     Math::sqr(genericRay.fDirection.fY);

  fCoefficients[3] = 0.0;
  fCoefficients[4] = 0.0;
}

Vector3D HyperbolicParaboloid::calcNormal(const Point3D& p) const
{
  return Vector3D( 2.0 * p.fX,
                  -2.0 * p.fY,
                  -1.0);
}

// -----------------------------
// class PinchedSurfaceWithHoles
// -----------------------------

// -----------
// constructor
// -----------
PinchedSurfaceWithHoles::PinchedSurfaceWithHoles(void)
{
}

// ----------
// destructor
// ----------
PinchedSurfaceWithHoles::~PinchedSurfaceWithHoles(void)
{
}

// ------------------------------
// cubic surface specific methods
// ------------------------------
void PinchedSurfaceWithHoles::setCoefficients(const Ray& genericRay)
{
  fCoefficients[0] = Math::cube(genericRay.fStart.fX) +
                     Math::cube(genericRay.fStart.fY) +
                     Math::cube(genericRay.fStart.fZ) -
                     genericRay.fStart.fX -
                     genericRay.fStart.fY -
                     genericRay.fStart.fZ;

  fCoefficients[1] = (3.0 * ((Math::sqr(genericRay.fStart.fX) *
                                        genericRay.fDirection.fX) +
                             (Math::sqr(genericRay.fStart.fY) *
                                        genericRay.fDirection.fY) +
                             (Math::sqr(genericRay.fStart.fZ) *
                                        genericRay.fDirection.fZ))) -
                     genericRay.fDirection.fX -
                     genericRay.fDirection.fY -
                     genericRay.fDirection.fZ;

  fCoefficients[2] = 3.0 * ((genericRay.fStart.fX *
                             Math::sqr(genericRay.fDirection.fX)) +
                            (genericRay.fStart.fY *
                             Math::sqr(genericRay.fDirection.fY)) +
                            (genericRay.fStart.fZ *
                             Math::sqr(genericRay.fDirection.fZ)));

  fCoefficients[3] = Math::cube(genericRay.fDirection.fX) +
                     Math::cube(genericRay.fDirection.fY) +
                     Math::cube(genericRay.fDirection.fZ);

  fCoefficients[4] = 0.0;
}

Vector3D PinchedSurfaceWithHoles::calcNormal(const Point3D& p) const
{
  double partialDerivativeToX = (3.0 * Math::sqr(p.fX)) - 1.0;
  double partialDerivativeToY = (3.0 * Math::sqr(p.fY)) - 1.0;
  double partialDerivativeToZ = (3.0 * Math::sqr(p.fZ)) - 1.0;

  return Vector3D(partialDerivativeToX,partialDerivativeToY,partialDerivativeToZ);
}

// ----------------------------
// class PinchedSurfaceWithLobe
// ----------------------------

// -----------
// constructor
// -----------
PinchedSurfaceWithLobe::PinchedSurfaceWithLobe(void)
{
}

// ----------
// destructor
// ----------
PinchedSurfaceWithLobe::~PinchedSurfaceWithLobe(void)
{
}

// ------------------------------
// cubic surface specific methods
// ------------------------------
void PinchedSurfaceWithLobe::setCoefficients(const Ray& genericRay)
{
  fCoefficients[0] = Math::sqr(genericRay.fStart.fX) -
                     Math::sqr(genericRay.fStart.fY) +
                     Math::sqr(genericRay.fStart.fZ) +
                     (Math::sqr(genericRay.fStart.fY) *
                      genericRay.fStart.fY);

  fCoefficients[1] = (2.0 * ((genericRay.fStart.fX *
                              genericRay.fDirection.fX) -
                             (genericRay.fStart.fY *
                              genericRay.fDirection.fY) +
                             (genericRay.fStart.fZ *
                              genericRay.fDirection.fZ))) +
                     (3.0 * Math::sqr(genericRay.fStart.fY) *
                            genericRay.fDirection.fY);

  fCoefficients[2] = Math::sqr(genericRay.fDirection.fX) +
                     (3.0 * genericRay.fStart.fY *
                            Math::sqr(genericRay.fDirection.fY)) -
                     Math::sqr(genericRay.fDirection.fY) +
                     Math::sqr(genericRay.fDirection.fZ);

  fCoefficients[3] = Math::cube(genericRay.fDirection.fY);
}

Vector3D PinchedSurfaceWithLobe::calcNormal(const Point3D& p) const
{
  double partialDerivativeToX = 2.0 * p.fX;
  double partialDerivativeToY = (3.0 * Math::sqr(p.fY)) - (2.0 * p.fY);
  double partialDerivativeToZ = 2.0 * p.fZ;

  return Vector3D(partialDerivativeToX,partialDerivativeToY,partialDerivativeToZ);
}

// ---------------
// class FourLobes
// ---------------

// -----------
// constructor
// -----------
FourLobes::FourLobes(void)
{
}

// ----------
// destructor
// ----------
FourLobes::~FourLobes(void)
{
}

// --------------------------------
// quartic surface specific methods
// --------------------------------
void FourLobes::setCoefficients(const Ray& genericRay)
{
  fCoefficients[0] = (5.0 * (genericRay.fStart.fX *
                             genericRay.fStart.fY *
                             genericRay.fStart.fZ)) +
                     Math::quart(genericRay.fStart.fX) +
                     Math::quart(genericRay.fStart.fY) +
                     Math::quart(genericRay.fStart.fZ);

  fCoefficients[1] = (5.0 * ((genericRay.fStart.fX *
                              genericRay.fStart.fY *
                              genericRay.fDirection.fZ) +
                             (genericRay.fStart.fX *
                              genericRay.fDirection.fY *
                              genericRay.fStart.fZ) +
                             (genericRay.fDirection.fX *
                              genericRay.fStart.fY *
                              genericRay.fStart.fZ))) +
                     (4.0 * ((Math::cube(genericRay.fStart.fX) *
                              genericRay.fDirection.fX) +
                             (Math::cube(genericRay.fStart.fY) *
                              genericRay.fDirection.fY) +
                             (Math::cube(genericRay.fStart.fZ) *
                              genericRay.fDirection.fZ)));

  fCoefficients[2] = (5.0 * ((genericRay.fStart.fX *
                              genericRay.fDirection.fY *
                              genericRay.fDirection.fZ) +
                             (genericRay.fDirection.fX *
                              genericRay.fStart.fY *
                              genericRay.fDirection.fZ) +
                             (genericRay.fDirection.fX *
                              genericRay.fDirection.fY *
                              genericRay.fStart.fZ))) +
                     (6.0 * ((Math::sqr(genericRay.fStart.fX) *
                              Math::sqr(genericRay.fDirection.fX)) +
                             (Math::sqr(genericRay.fStart.fY) *
                              Math::sqr(genericRay.fDirection.fY)) +
                             (Math::sqr(genericRay.fStart.fZ) *
                              Math::sqr(genericRay.fDirection.fZ))));

  fCoefficients[3] = (5.0 * (genericRay.fDirection.fX *
                             genericRay.fDirection.fY *
                             genericRay.fDirection.fZ)) +
                     (4.0 * ((genericRay.fStart.fX *
                              Math::cube(genericRay.fDirection.fX)) +
                             (genericRay.fStart.fY *
                              Math::cube(genericRay.fDirection.fY)) +
                             (genericRay.fStart.fZ *
                              Math::cube(genericRay.fDirection.fZ))));

  fCoefficients[4] = Math::quart(genericRay.fDirection.fX) +
                     Math::quart(genericRay.fDirection.fY) +
                     Math::quart(genericRay.fDirection.fZ);
}

Vector3D FourLobes::calcNormal(const Point3D& p) const
{
  double partialDerivativeToX = (4.0 * Math::cube(p.fX)) + (5.0 * p.fY * p.fZ);
  double partialDerivativeToY = (4.0 * Math::cube(p.fY)) + (5.0 * p.fX * p.fZ);
  double partialDerivativeToZ = (4.0 * Math::cube(p.fZ)) + (5.0 * p.fX * p.fY);

  return Vector3D(partialDerivativeToX,partialDerivativeToY,partialDerivativeToZ);
}

// --------------------------
// class SteinersRomanSurface
// --------------------------

// -----------
// constructor
// -----------
SteinersRomanSurface::SteinersRomanSurface(void)
{
}

// ----------
// destructor
// ----------
SteinersRomanSurface::~SteinersRomanSurface(void)
{
}

// --------------------------------
// quartic surface specific methods
// --------------------------------
void SteinersRomanSurface::setCoefficients(const Ray& genericRay)
{
  fCoefficients[0] = (Math::sqr(genericRay.fStart.fX) *
                      Math::sqr(genericRay.fStart.fY)) +
                     (Math::sqr(genericRay.fStart.fX) *
                      Math::sqr(genericRay.fStart.fZ)) +
                     (Math::sqr(genericRay.fStart.fY) *
                      Math::sqr(genericRay.fStart.fZ)) -
                     (genericRay.fStart.fX *
                      genericRay.fStart.fY *
                      genericRay.fStart.fZ);

  fCoefficients[1] = (2.0 * ((Math::sqr(genericRay.fStart.fX) *
                              genericRay.fStart.fY *
                              genericRay.fDirection.fY) +
                             (genericRay.fStart.fX *
                              genericRay.fDirection.fX *
                              Math::sqr(genericRay.fStart.fY)) +
                             (Math::sqr(genericRay.fStart.fX) *
                              genericRay.fStart.fZ *
                              genericRay.fDirection.fZ) +
                             (genericRay.fStart.fX *
                              genericRay.fDirection.fX *
                              Math::sqr(genericRay.fStart.fZ)) +
                             (Math::sqr(genericRay.fStart.fY) *
                              genericRay.fStart.fZ *
                              genericRay.fDirection.fZ) +
                             (genericRay.fStart.fY *
                              genericRay.fDirection.fY *
                              Math::sqr(genericRay.fStart.fZ)))) -
                     (genericRay.fStart.fX *
                      genericRay.fStart.fY *
                      genericRay.fDirection.fZ) -
                     (genericRay.fStart.fX *
                      genericRay.fDirection.fY *
                      genericRay.fStart.fZ) -
                     (genericRay.fDirection.fX *
                      genericRay.fStart.fY *
                      genericRay.fStart.fZ);

  fCoefficients[2] = (Math::sqr(genericRay.fStart.fX) *
                      Math::sqr(genericRay.fDirection.fY)) +
                     (4.0 * (genericRay.fStart.fX *
                             genericRay.fDirection.fX *
                             genericRay.fStart.fY *
                             genericRay.fDirection.fY)) +
                     (Math::sqr(genericRay.fDirection.fX) *
                      Math::sqr(genericRay.fStart.fY)) +
                     (Math::sqr(genericRay.fStart.fX) *
                      Math::sqr(genericRay.fDirection.fZ)) +
                     (4.0 * (genericRay.fStart.fX *
                             genericRay.fDirection.fX *
                             genericRay.fStart.fZ *
                             genericRay.fDirection.fZ)) +
                     (Math::sqr(genericRay.fDirection.fX) *
                      Math::sqr(genericRay.fStart.fZ)) +
                     (Math::sqr(genericRay.fStart.fY) *
                      Math::sqr(genericRay.fDirection.fZ)) +
                     (4.0 * (genericRay.fStart.fY *
                             genericRay.fDirection.fY *
                             genericRay.fStart.fZ *
                             genericRay.fDirection.fZ)) +
                     (Math::sqr(genericRay.fDirection.fY) *
                      Math::sqr(genericRay.fStart.fZ)) -
                     (genericRay.fStart.fX *
                      genericRay.fDirection.fY *
                      genericRay.fDirection.fZ) -
                     (genericRay.fDirection.fX *
                      genericRay.fStart.fY *
                      genericRay.fDirection.fZ) -
                     (genericRay.fDirection.fX *
                      genericRay.fDirection.fY *
                      genericRay.fStart.fZ);

  fCoefficients[3] = (2.0 * ((genericRay.fStart.fX *
                              genericRay.fDirection.fX *
                              Math::sqr(genericRay.fDirection.fY)) +
                             (Math::sqr(genericRay.fDirection.fX) *
                              genericRay.fStart.fY *
                              genericRay.fDirection.fY) +
                             (genericRay.fStart.fX *
                              genericRay.fDirection.fX *
                              Math::sqr(genericRay.fDirection.fZ)) +
                             (Math::sqr(genericRay.fDirection.fX) *
                              genericRay.fStart.fZ *
                              genericRay.fDirection.fZ) +
                             (genericRay.fStart.fY *
                              genericRay.fDirection.fY *
                              Math::sqr(genericRay.fDirection.fZ)) +
                             (Math::sqr(genericRay.fDirection.fY) *
                              genericRay.fStart.fZ *
                              genericRay.fDirection.fZ))) -
                     (genericRay.fDirection.fX *
                      genericRay.fDirection.fY *
                      genericRay.fDirection.fZ);

  fCoefficients[4] = (Math::sqr(genericRay.fDirection.fX) *
                      Math::sqr(genericRay.fDirection.fY)) +
                     (Math::sqr(genericRay.fDirection.fX) *
                      Math::sqr(genericRay.fDirection.fZ)) +
                     (Math::sqr(genericRay.fDirection.fY) *
                      Math::sqr(genericRay.fDirection.fZ));
}

Vector3D SteinersRomanSurface::calcNormal(const Point3D& p) const
{
  double partialDerivativeToX = (2.0 * p.fX * (Math::sqr(p.fY) + Math::sqr(p.fZ))) - (p.fY * p.fZ);
  double partialDerivativeToY = (2.0 * p.fY * (Math::sqr(p.fX) + Math::sqr(p.fZ))) - (p.fX * p.fZ);
  double partialDerivativeToZ = (2.0 * p.fZ * (Math::sqr(p.fX) + Math::sqr(p.fY))) - (p.fX * p.fY);

  return Vector3D(partialDerivativeToX,partialDerivativeToY,partialDerivativeToZ);
}

// ---------------
// class ToothCube
// ---------------

// -----------
// constructor
// -----------
ToothCube::ToothCube(void)
{
}

// ----------
// destructor
// ----------
ToothCube::~ToothCube(void)
{
}

// --------------------------------
// quartic surface specific methods
// --------------------------------
void ToothCube::setCoefficients(const Ray& genericRay)
{
  fCoefficients[0] = Math::quart(genericRay.fStart.fX) +
                     Math::quart(genericRay.fStart.fY) +
                     Math::quart(genericRay.fStart.fZ) -
                     Math::sqr(genericRay.fStart.fX) -
                     Math::sqr(genericRay.fStart.fY) -
                     Math::sqr(genericRay.fStart.fZ);

  fCoefficients[1] = (4.0 * ((Math::cube(genericRay.fStart.fX) * 
                              genericRay.fDirection.fX) +
                             (Math::cube(genericRay.fStart.fY) * 
                              genericRay.fDirection.fY) +
                             (Math::cube(genericRay.fStart.fZ) * 
                              genericRay.fDirection.fZ))) -
                     (2.0 * ((genericRay.fStart.fX *
                              genericRay.fDirection.fX) +
                             (genericRay.fStart.fY *
                              genericRay.fDirection.fY) +
                             (genericRay.fStart.fZ *
                              genericRay.fDirection.fZ)));

  fCoefficients[2] = (6.0 * ((Math::sqr(genericRay.fStart.fX) *
                              Math::sqr(genericRay.fDirection.fX)) +
                             (Math::sqr(genericRay.fStart.fY) *
                              Math::sqr(genericRay.fDirection.fY)) +
                             (Math::sqr(genericRay.fStart.fZ) *
                              Math::sqr(genericRay.fDirection.fZ)))) -
                     Math::sqr(genericRay.fDirection.fX) -
                     Math::sqr(genericRay.fDirection.fY) -
                     Math::sqr(genericRay.fDirection.fZ);

  fCoefficients[3] = 4.0 * ((genericRay.fStart.fX *
                             Math::cube(genericRay.fDirection.fX)) +
                            (genericRay.fStart.fY *
                             Math::cube(genericRay.fDirection.fY)) +
                            (genericRay.fStart.fZ *
                             Math::cube(genericRay.fDirection.fZ)));

  fCoefficients[4] = Math::quart(genericRay.fDirection.fX) +
                     Math::quart(genericRay.fDirection.fY) +
                     Math::quart(genericRay.fDirection.fZ);
}

Vector3D ToothCube::calcNormal(const Point3D& p) const
{
  double partialDerivativeToX = 2.0 * p.fX * (2.0 * Math::sqr(p.fX) - 1.0);
  double partialDerivativeToY = 2.0 * p.fY * (2.0 * Math::sqr(p.fY) - 1.0);
  double partialDerivativeToZ = 2.0 * p.fZ * (2.0 * Math::sqr(p.fZ) - 1.0);

  return Vector3D(partialDerivativeToX,partialDerivativeToY,partialDerivativeToZ);
}
