// ---------------------------------
// Filename      : solidtexturer.cpp
// Author        : Sven Maerivoet
// Last modified : 31/05/2001
// Target        : C++ (OpenGL)
//
// All rights reserved.
// ---------------------------------

#ifdef _WIN32
  #include <windows.h>
#endif

#include "math.h"
#include "rng.h"
#include "solidtexturer.h"
#include <GL/gl.h>
#include <GL/glu.h>
#include <GL/glut.h>
#include <math.h>
#include <stdlib.h>

// -------------------
// class SolidTexturer
// -------------------

// -----------
// constructor
// -----------
SolidTexturer::SolidTexturer(void) : kOffset(100.0)
{
}

// ----------
// destructor
// ----------
SolidTexturer::~SolidTexturer(void)
{
}

// -------
// methods
// -------
void SolidTexturer::initialize(void)
{
  int i;

  fIndex = new unsigned char[256];

  for (i = 0; i < 256; ++i) {
    fIndex[i] = i;
  }

  for (i = 0; i < 256; ++i) {
    int which = RNG::uniformInt() % 256;
    unsigned char tmp = fIndex[which];
    fIndex[which] = fIndex[i];
    fIndex[i] = tmp;
  }

  fNoiseTable = new double[256];
  for (i = 0; i < 256; ++i) {
    fNoiseTable[i] = RNG::uniform();
  }
}

Color SolidTexturer::sampleCheckerboard(const Point3D& p,
                                        const Vector3D& scale) const
{
  double intensity = ((int) (kOffset + (p.fX / scale.fX)) +
                      (int) (kOffset + (p.fY / scale.fY)) +
                      (int) (kOffset + (p.fZ / scale.fZ))) % 2;

  return Color(intensity,intensity,intensity);
}

Color SolidTexturer::sampleStackedColorBars(const Point3D& p,
                                            const Vector3D& scale) const
{
  return Color(1.0 - Math::abs((2.0 * Math::frac(p.fX / scale.fX)) - 1.0),
               1.0 - Math::abs((2.0 * Math::frac(p.fY / scale.fX)) - 1.0),
               1.0 - Math::abs((2.0 * Math::frac(p.fZ / scale.fX)) - 1.0));
}

Color SolidTexturer::sampleRGBCubes(const Point3D& p,
                                    const Vector3D& scale) const
{
  return Color(Math::frac(kOffset + (p.fX / scale.fX)),
               Math::frac(kOffset + (p.fY / scale.fY)),
               Math::frac(kOffset + (p.fZ / scale.fZ)));
}

Color SolidTexturer::sampleOneCoordinateBasedContourLines(const Point3D& p,
                                                          int coordinateInvolved,
                                                          double scale) const
{
  double intensity;

  if (coordinateInvolved == 0) {
    intensity = (int) (kOffset + (p.fX / scale)) % 2;
  }
  else if (coordinateInvolved == 1) {
    intensity = (int) (kOffset + (p.fY / scale)) % 2;
  }
  else {
    intensity = (int) (kOffset + (p.fZ / scale)) % 2;
  }

  return Color(intensity,intensity,intensity);
}

Color SolidTexturer::sampleTwoCoordinateBasedContourLines(const Point3D& p,
                                                          int coordinatesInvolved,
                                                          int distanceMeasureToUse,
                                                          const Vector2D& scale) const
{
  double a, b;

  if (coordinatesInvolved == 0) {
    a = p.fX;
    b = p.fY;
  }
  else if (coordinatesInvolved == 1) {
    a = p.fY;
    b = p.fZ;
  }
  else {
    a = p.fX;
    b = p.fZ;
  }

  double distance;

  if (distanceMeasureToUse == 0) {
    a /= scale.fX;
    b /= scale.fY;
    distance = sqrt((a * a) + (b * b));
  }
  else if (distanceMeasureToUse == 1) {
    a /= scale.fX;
    b /= scale.fY;
    distance = Math::abs(a) + Math::abs(b);
  }
  else if (distanceMeasureToUse == 2) {
    distance = (sin(a) / scale.fX) * (sin(b) / scale.fY);
  }
  else {
    distance = (sin(a) / scale.fX) * (cos(b) / scale.fY);
  }

  double intensity = ((int) (kOffset + distance)) % 2;

  return Color(intensity,intensity,intensity);
}

Color SolidTexturer::sampleThreeCoordinateBasedContourLines(const Point3D& p,
                                                            int distanceMeasureToUse,
                                                            const Vector3D& scale) const
{
  double a, b, c, distance;

  if (distanceMeasureToUse == 0) {
    a = (p.fX / scale.fX);
    b = (p.fY / scale.fY);
    c = (p.fZ / scale.fZ);
    distance = sqrt((a * a) + (b * b) + (c * c));
  }
  else {
    a = (p.fX / scale.fX);
    b = (p.fY / scale.fY);
    c = (p.fZ / scale.fZ);
    distance = Math::abs(a) + Math::abs(b) + Math::abs(c);
  }

  double intensity = ((int) (kOffset + distance)) % 2;

  return Color(intensity,intensity,intensity);
}

Color SolidTexturer::sampleWoodGrain(const Point3D& p,
                                     double ringThickness,
                                     double wobbleStrength,
                                     int nrOfWobbles,
                                     double twistStrength,
                                     const Color& color1,
                                     const Color& color2,
                                     bool interpolateSmoothly) const
{
  double r = sqrt((p.fX * p.fX) + (p.fY * p.fY));

  // adjust rings' thickness
  r /= ringThickness;

  // wobble, skew, rotate and twist the rings
  double theta = Math::arctan(p.fY,p.fX);
  r += wobbleStrength * (sin((theta * nrOfWobbles) + (twistStrength * p.fZ)));

  Color ringColor;

  if (interpolateSmoothly) {
    double rings = Math::frac(r);
    ringColor = (color1 * rings) + (color2 * (1.0 - rings));
  }
  else {
    double rings = ((int) floor(r)) % 2;
    if (rings == 0) {
      ringColor = color1;
    }
    else {
      ringColor = color2;
    }
  }

  return ringColor;
}

Color SolidTexturer::sampleMarble(const Point3D& p,
                                  int coordinateInvolved,
                                  double turbulenceStrength,
                                  double redWeight,
                                  double greenWeight,
                                  double blueWeight) const
{
  double coordinate;

  if (coordinateInvolved == 0) {
    coordinate = p.fX;
  }
  else if (coordinateInvolved == 1) {
    coordinate = p.fY;
  }
  else if (coordinateInvolved == 2) {
    coordinate = p.fZ;
  }

  double red = spline(sin((redWeight * coordinate) + (turbulenceStrength * turbulence(10.0,p))));
  double green = spline(sin((greenWeight * coordinate) + (turbulenceStrength * turbulence(10.0,p))));
  double blue = spline(sin((blueWeight * coordinate) + (turbulenceStrength * turbulence(10.0,p))));

  return Color(red,green,blue);
}

Color SolidTexturer::sampleCowSpots(const Point3D& p,
                                    double treshold) const
{
  double distance = sqrt(Math::sqr(p.fX) + Math::sqr(p.fY) + Math::sqr(p.fZ));
  // only use 1 frequency for the noise
  double disturbance = noise(10.0,p) / 2.0;
  
  double value = spline(sin(distance) + disturbance);

  if (Math::sqr(value) < Math::sqr(treshold)) {
    return Color(0.0,0.0,0.0);
  }
  else {
    return Color(1.0,1.0,1.0);
  }
}

// ---------------
// utility methods
// ---------------
double SolidTexturer::spline(double x) const
{
  // a nice spline
  if (x < -0.4) {
    return 0.15 + (2.857 * Math::sqr(x + 0.75));
  }
  else if (x < 0.4) {
    return 0.95 - (2.8125 * Math::sqr(x));
  }
  else {
    return 0.26 + (2.666 * Math::sqr(x - 0.7));
  }
}

double SolidTexturer::turbulence(double scale, const Point3D& p) const
{
  return ((noise(scale,p) / 2.0) +
          (noise(2.0 * scale,p) / 4.0) +
          (noise(4.0 * scale,p) / 8.0) +
          (noise(8.0 * scale,p) / 16.0));
}

double SolidTexturer::noise(double scale, const Point3D& p) const
{
  #define lerp(f,A,B) A + f * (B - A)

  double d[2][2][2];
  Point3D q;

  q.fX = (p.fX * scale) + 10000.0;
  q.fY = (p.fY * scale) + 10000.0;
  q.fZ = (p.fZ * scale) + 10000.0;

  long ix = (long) floor(q.fX);
  long iy = (long) floor(q.fY);
  long iz = (long) floor(q.fZ);

  double tx = Math::frac(q.fX);
  double ty = Math::frac(q.fY);
  double tz = Math::frac(q.fZ);

  double mtx = 1.0 - tx;
  double mty = 1.0 - ty;
  double mtz = 1.0 - tz;

  for (int k = 0; k <= 1; ++k) {
    for (int j = 0; j <= 1; ++j) {
      for (int i = 0; i <= 1; ++i) {
        d[k][j][i] = latticeNoise(ix + i,iy + j,iz + k);
      }
    }
  }

  // perform linear interpolation in 3D
  double x0 = lerp(tx,d[0][0][0],d[0][0][1]);
  double x1 = lerp(tx,d[0][1][0],d[0][1][1]);
  double x2 = lerp(tx,d[1][0][0],d[1][0][1]);
  double x3 = lerp(tx,d[1][1][0],d[1][1][1]);

  double y0 = lerp(ty,x0,x1);
  double y1 = lerp(ty,x2,x3);

  return lerp(tz,y0,y1);
}

double SolidTexturer::latticeNoise(int i, int j, int k) const
{
  #define PERM(x) fIndex[(x) & 255]
  #define INDEX(ix,iy,iz) PERM(ix + PERM(iy + PERM(iz)))
  return fNoiseTable[INDEX(i,j,k)];
}
