// ------------------------------
// Filename      : terrainmap.cpp
// Author        : Sven Maerivoet
// Last modified : 05/05/2001
// Target        : C++ (OpenGL)
//
// All rights reserved.
// ------------------------------

#include "constants.h"
#include "math.h"
#include "rng.h"
#include "terrainmap.h"
#include <iostream>
#include <fstream>
#include <math.h>
#include <stdlib.h>

using namespace std;

// ----------------
// class TerrainMap
// ----------------

// --------------------------------
// initialization of static members
// --------------------------------
const bool TerrainMap::kFlattenNormals  = false;
const bool TerrainMap::kSmoothenNormals = true;

// -----------
// constructor
// -----------
TerrainMap::TerrainMap(void) : fHeightfield(0),
                               fUSize(0),
                               fVSize(0),
                               fBounds(0.0,0.0,0.0,0.0),
                               fRoughness(0.0),
                               fSeaLevel(0.0),
                               fSnowLevel(0.0),
                               fMinimumHeight(0.0),
                               fMaximumHeight(0.0),
                               fLowestSeaColor(Color(0.0,0.0,0.2)),
                               fHighestSeaColor(Color(0.0,0.0,0.9)),
                               fLowestMountainColor(Color(0.0,0.4,0.0)),
                               fHighestMountainColor(Color(0.0,0.8,0.0)),
                               fLowestSnowColor(Color(0.8,0.8,0.8)),
                               fHighestSnowColor(Color(1.0,1.0,1.0)),
                               fSeaTexture(0),
                               fMountainTexture(0),
                               fSnowTexture(0),
                               fCurrentTime(0),
                               fWavePeriod(10),
                               fWaveHeight(5.0),
                               fSeaFrequency(4),
                               fUseRealColors(false)
{
}

// ----------
// destructor
// ----------
TerrainMap::~TerrainMap(void)
{
  freeTerrainMap();
}

// -------
// methods
// -------
void TerrainMap::createFractalTerrain(int size,
                                      double upperLeftHeight,
                                      double lowerLeftHeight,
                                      double upperRightHeight,
                                      double lowerRightHeight,
                                      double initialPerturbation,
                                      double roughness,
                                      const Rectangle2D& bounds)
{
  // determine the least number of subdivisions needed
  if (size < 2) {
    size = 2;
  }
  int nrOfSubdivisions = (int) ceil(log(size) / log(2.0));

  fUSize = ((int) pow(2.0,nrOfSubdivisions));
  fVSize = fUSize;
  createTerrainMap();

  // set the initial heights at the corners
  (*fHeightfield)[0][0].fHeight = upperLeftHeight;
  (*fHeightfield)[fUSize - 1][0].fHeight = upperRightHeight;
  (*fHeightfield)[0][fVSize - 1].fHeight = lowerLeftHeight;
  (*fHeightfield)[fUSize - 1][fVSize - 1].fHeight = lowerRightHeight;

  fRoughness = roughness;

  generateRecursiveFractalTerrain(nrOfSubdivisions,
                                  initialPerturbation,
                                  0,
                                  0,
                                  fUSize - 1,
                                  fVSize - 1);

  disableSea();
  calcMinMaxHeights();

  setBounds(bounds);
}

void TerrainMap::constructFromBMPFile(string filename,
                                      double maxHeight,
                                      const Rectangle2D& bounds)
{
  RGBApixmap* rgbapixmap = new RGBApixmap();

  if (!rgbapixmap->loadFromBMPFile(filename)) {
    exit(0);
  }

  fUSize = rgbapixmap->getNrOfColumns();
  fVSize = rgbapixmap->getNrOfRows();
  createTerrainMap();

  // convert the RGBA-triplets to heights (using a gray-scale transformation)
  for (int v = 0; v < fVSize; ++v) {
    for (int u = 0; u < fUSize; ++u) {

      // invert the u-coordinate (otherwise the image is mirrored)
      mRGBA mrgba = rgbapixmap->getRGBATriplet((fUSize - 1) - u,v);
      mrgba.convert((*fHeightfield)[u][v].fRed,
                    (*fHeightfield)[u][v].fGreen,
                    (*fHeightfield)[u][v].fBlue);

      Color color((*fHeightfield)[u][v].fRed,
                  (*fHeightfield)[u][v].fGreen,
                  (*fHeightfield)[u][v].fBlue);

      (*fHeightfield)[u][v].fHeight = maxHeight * color.calcGrayLevel(color);
      (*fHeightfield)[u][v].fIsAtSeaLevel = false;
    }
  }

  delete rgbapixmap;

  calcMinMaxHeights();

  setBounds(bounds);
}

void TerrainMap::setBounds(const Rectangle2D& bounds)
{
  fBounds = bounds;
  averageVertexNormals();
}

Rectangle2D TerrainMap::getBounds(void) const
{
  return fBounds;
}

void TerrainMap::setHeight(int u, int v, double h)
{
  (*fHeightfield)[u][v].fHeight = h;
}

double TerrainMap::getHeight(int u, int v) const
{
  if ((*fHeightfield)[u][v].fIsAtSeaLevel) {

    double gamma = (2.0 * Math::kPi * fSeaFrequency * u) / (fUSize - 1);
    double delta = (2.0 * Math::kPi * fSeaFrequency * v) / (fVSize - 1);
    double alpha = ((2.0 * Math::kPi * fCurrentTime) / fWavePeriod) + gamma;
    double beta  = ((2.0 * Math::kPi * fCurrentTime) / fWavePeriod) + delta;

    double height = fSeaLevel + (fWaveHeight * (sin(alpha) + cos(beta)) / 2.0);

    // clip height if needed
    if (height < fMinimumHeight) {
      return fMinimumHeight;
    }
    else if (height > fMaximumHeight) {
      return fMaximumHeight;
    }
    else {
      return height;
    }
  }
  else {
    return (*fHeightfield)[u][v].fHeight;
  }
}

void TerrainMap::setSeaLevel(double seaLevel)
{
  fSeaLevel = seaLevel;

  // level the terrain
  for (int v = 1; v < (fVSize - 1); ++v) {
    for (int u = 1; u < (fUSize - 1); ++u) {

      if ((*fHeightfield)[u][v].fHeight <= fSeaLevel) {

        // handle the location's square neighbourhood
        (*fHeightfield)[u - 1][v - 1].fIsAtSeaLevel = true;
        (*fHeightfield)[u    ][v - 1].fIsAtSeaLevel = true;
        (*fHeightfield)[u + 1][v - 1].fIsAtSeaLevel = true;
        (*fHeightfield)[u - 1][v    ].fIsAtSeaLevel = true;
        (*fHeightfield)[u    ][v    ].fIsAtSeaLevel = true;
        (*fHeightfield)[u + 1][v    ].fIsAtSeaLevel = true;
        (*fHeightfield)[u - 1][v + 1].fIsAtSeaLevel = true;
        (*fHeightfield)[u    ][v + 1].fIsAtSeaLevel = true;
        (*fHeightfield)[u + 1][v + 1].fIsAtSeaLevel = true;
      }
      else {
        (*fHeightfield)[u][v].fIsAtSeaLevel = false;
      }
    }
  }

  calcMinMaxHeights();
}

double TerrainMap::getSeaLevel(void) const
{
  return fSeaLevel;
}

void TerrainMap::setSeaFrequency(int seaFrequency)
{
  fSeaFrequency = seaFrequency;
}

int TerrainMap::getSeaFrequency(void) const
{
  return fSeaFrequency;
}

void TerrainMap::setWavePeriod(int wavePeriod)
{
  fWavePeriod = wavePeriod;
}

int TerrainMap::getWavePeriod(void) const
{
  return fWavePeriod;
}

void TerrainMap::setWaveHeight(double waveHeight)
{
  fWaveHeight = waveHeight;
}

double TerrainMap::getWaveHeight(void) const
{
  return fWaveHeight;
}

void TerrainMap::disableSea(void)
{
  for (int v = 0; v < fVSize; ++v) {
    for (int u = 0; u < fUSize; ++u) {
      (*fHeightfield)[u][v].fIsAtSeaLevel = false;
    }
  }
}

void TerrainMap::setSnowLevel(double snowLevel)
{
  fSnowLevel = snowLevel;
}

double TerrainMap::getSnowLevel(void) const
{
  return fSnowLevel;
}

void TerrainMap::setColors(const Color& lowestSeaColor,
                           const Color& highestSeaColor,
                           const Color& lowestMountainColor,
                           const Color& highestMountainColor,
                           const Color& lowestSnowColor,
                           const Color& highestSnowColor)
{
  fLowestSeaColor = lowestSeaColor;
  fHighestSeaColor = highestSeaColor;
  fLowestMountainColor = lowestMountainColor;
  fHighestMountainColor = highestMountainColor;
  fLowestSnowColor = lowestSnowColor;
  fHighestSnowColor = highestSnowColor;
}

Color TerrainMap::getColor(int u, int v) const
{
  if (fUseRealColors) {
    return Color((*fHeightfield)[u][v].fRed,
                   (*fHeightfield)[u][v].fGreen,
                   (*fHeightfield)[u][v].fBlue);
  }
  else {

    double height = getHeight(u,v);

    if ((*fHeightfield)[u][v].fIsAtSeaLevel) {
      return linearInterpolatedColor(fLowestSeaColor,
                                     fHighestSeaColor,
                                     fMinimumHeight,
                                     fSeaLevel,
                                     (*fHeightfield)[u][v].fHeight); // return the real depth !
    }
    else if (height >= fSnowLevel) {
      return linearInterpolatedColor(fLowestSnowColor,
                                     fHighestSnowColor,
                                     fSnowLevel,
                                     fMaximumHeight,
                                     height);
    }
    else {
      return linearInterpolatedColor(fLowestMountainColor,
                                     fHighestMountainColor,
                                     fSeaLevel,
                                     fSnowLevel,
                                     height);
    }
  }
}

void TerrainMap::useSpecifiedColors(void)
{
  fUseRealColors = false;
}

void TerrainMap::useRealColors(void)
{
  fUseRealColors = true;
}

void TerrainMap::setTextures(RGBApixmap* seaTexture,
                             RGBApixmap* mountainTexture,
                             RGBApixmap* snowTexture)
{
  fSeaTexture = seaTexture;
  fMountainTexture = mountainTexture;
  fSnowTexture = snowTexture;
}

RGBApixmap* TerrainMap::getTexture(int u, int v) const
{
  double height = getHeight(u,v);

  if ((*fHeightfield)[u][v].fIsAtSeaLevel) {
    return fSeaTexture;
  }
  else if (height >= fSnowLevel) {
    return fSnowTexture;
  }
  else {
    return fMountainTexture;
  }
}

void TerrainMap::setCurrentTime(int currentTime)
{
  fCurrentTime = currentTime;
}

int TerrainMap::getUSize(void) const
{
  return fUSize;
}

int TerrainMap::getVSize(void) const
{
  return fVSize;
}

Vector3D TerrainMap::getVertexNormal(int u, int v) const
{
  return (*fHeightfield)[u][v].fNormal;
}

void TerrainMap::writeTo3VNMeshFile(string filename) const
{
  fstream outStream;
  outStream.open(filename.c_str(),ios::out);

  if (outStream.fail()) {
    cerr << "TerrainMap::writeTo3VNMeshFile() -> can't open outputfile [" << filename
         << "] for writing." << endl;
    return;
  }

  int uSize = getUSize();
  int vSize = getVSize();

  int nrOfVertices = uSize * vSize;
  int nrOfNormals = nrOfVertices;
  int nrOfFaces = (2 * (uSize - 1)) * (vSize - 1);

  outStream << nrOfVertices << endl;
  outStream << nrOfNormals << endl;
  outStream << nrOfFaces << endl;

  Rectangle2D bounds = getBounds();
  Point2D p1(bounds.fLeft,bounds.fTop);
  Point2D p2(bounds.fRight,bounds.fBottom);

  int u, v;

  // save the vertices
  for (v = 0; v < vSize; ++v) {
    for (u = 0; u < uSize; ++u) {

      double x = p1.fX + ((p2.fX - p1.fX) * (((double) u) / (uSize - 1)));
      double y = getHeight(u,v);
      double z = p1.fY + ((p2.fY - p1.fY) * (((double) v) / (vSize - 1)));

      outStream << x << " "
                << y << " "
                << z << endl;
    }
  }

  // save the normals
  for (v = 0; v < vSize; ++v) {
    for (u = 0; u < uSize; ++u) {

      Vector3D vertexNormal(getVertexNormal(u,v));

      outStream << vertexNormal.fX << " "
                << vertexNormal.fY << " "
                << vertexNormal.fZ << endl;
    }
  }

  // save the faces
  for (v = 0; v < (vSize - 1); ++v) {
    for (u = 0; u < (uSize - 1); ++u) {

      // save upper-left triangle of face
      outStream << "3" << endl;

      int index1 = (v * uSize) + u;
      int index2 = ((v + 1) * uSize) + u;
      int index3 = index1 + 1;

      // save the vertex-indices
      outStream << index1 << " "
                << index2 << " "
                << index3 << endl;

      // save the normal-indices
      outStream << index1 << " "
                << index2 << " "
                << index3 << endl;

      // save lower-right triangle of face
      outStream << "3" << endl;

      index1 = index1 + 1;
      index2 = ((v + 1) * uSize) + u;
      index3 = index2 + 1;

      // save the vertex-indices
      outStream << index1 << " "
                << index2 << " "
                << index3 << endl;

      // save the normal-indices
      outStream << index1 << " "
                << index2 << " "
                << index3 << endl;

    }
  }

  outStream.close();
}

// ---------------
// utility methods
// ---------------
void TerrainMap::freeTerrainMap(void)
{
  delete fHeightfield;
  fHeightfield = 0;
}

void TerrainMap::createTerrainMap(void)
{
  freeTerrainMap();

  // create the two-dimensional array
  fHeightfield = new Heightfield;

  fHeightfield->resize(fUSize);
  for (int i = 0; i < fUSize; ++i) {
    (*fHeightfield)[i].resize(fVSize);
  }
}

void TerrainMap::generateRecursiveFractalTerrain(int subdivisionsLeft,
                                                 double perturbation,
                                                 int u1, int v1,
                                                 int u2, int v2)
{
  int middleU = (u1 + u2) / 2;
  int middleV = (v1 + v2) / 2;

  (*fHeightfield)[middleU][v1].fHeight      = ((*fHeightfield)[u1][v1].fHeight +
                                               (*fHeightfield)[u2][v1].fHeight) / 2.0;
  (*fHeightfield)[middleU][v2].fHeight      = ((*fHeightfield)[u1][v2].fHeight +
                                               (*fHeightfield)[u2][v2].fHeight) / 2.0;
  (*fHeightfield)[u1][middleV].fHeight      = ((*fHeightfield)[u1][v1].fHeight +
                                               (*fHeightfield)[u1][v2].fHeight) / 2.0;
  (*fHeightfield)[u2][middleV].fHeight      = ((*fHeightfield)[u2][v1].fHeight +
                                               (*fHeightfield)[u2][v2].fHeight) / 2.0;
  (*fHeightfield)[middleU][middleV].fHeight = ((*fHeightfield)[middleU][v1].fHeight +
                                               (*fHeightfield)[middleU][v2].fHeight +
                                               (*fHeightfield)[u1][middleV].fHeight +
                                               (*fHeightfield)[u2][middleV].fHeight) / 4.0;

  // perform random midpoint-displacement
  (*fHeightfield)[middleU][v1].fHeight      += -(perturbation / 2.0) + (perturbation * RNG::uniform());
  (*fHeightfield)[middleU][v2].fHeight      += -(perturbation / 2.0) + (perturbation * RNG::uniform());
  (*fHeightfield)[u1][middleV].fHeight      += -(perturbation / 2.0) + (perturbation * RNG::uniform());
  (*fHeightfield)[u2][middleV].fHeight      += -(perturbation / 2.0) + (perturbation * RNG::uniform());
  (*fHeightfield)[middleU][middleV].fHeight += -(perturbation / 2.0) + (perturbation * RNG::uniform());

  // handle the four newly created quadrants
  if (subdivisionsLeft > 1) {
    generateRecursiveFractalTerrain(subdivisionsLeft - 1,perturbation * fRoughness,u1,v1,middleU,middleV);
    generateRecursiveFractalTerrain(subdivisionsLeft - 1,perturbation * fRoughness,middleU,v1,u2,middleV);
    generateRecursiveFractalTerrain(subdivisionsLeft - 1,perturbation * fRoughness,u1,middleV,middleU,v2);
    generateRecursiveFractalTerrain(subdivisionsLeft - 1,perturbation * fRoughness,middleU,middleV,u2,v2);
  }
}

void TerrainMap::calcMinMaxHeights(void)
{
  fMinimumHeight = (*fHeightfield)[0][0].fHeight;
  fMaximumHeight = fMinimumHeight;

  for (int v = 0; v < fVSize; ++v) {
    for (int u = 0; u < fUSize; ++u) {

      if ((*fHeightfield)[u][v].fHeight < fMinimumHeight) {
        fMinimumHeight = (*fHeightfield)[u][v].fHeight;
      }
      else if ((*fHeightfield)[u][v].fHeight > fMaximumHeight) {
        fMaximumHeight = (*fHeightfield)[u][v].fHeight;
      }
    }
  }
}

void TerrainMap::setNormal(int u, int v, const Vector3D& normal)
{
  (*fHeightfield)[u][v].fNormal = normal;
}

void TerrainMap::averageVertexNormals(void)
{
  Point2D p1(fBounds.fLeft,fBounds.fTop);
  Point2D p2(fBounds.fRight,fBounds.fBottom);

  typedef vector< vector<Vector3D> > SurfaceNormals;
  SurfaceNormals surfaceNormals;

  surfaceNormals.resize(fUSize - 1);
  int i = 0;
  for (i = 0; i < (fUSize - 1); ++i) {
    surfaceNormals[i].resize(2 * (fVSize - 1));
  }

  int u = 0;
  int v = 0;

  // calculate all the face-normals
  for (v = 0; v < (fVSize - 1); ++v) {
    for (u = 0; u < (fUSize - 1); ++u) {

      double u1 = p1.fX + ((p2.fX - p1.fX) * (((double) u) / (fUSize - 1)));
      double v1 = p1.fY + ((p2.fY - p1.fY) * (((double) v) / (fVSize - 1)));

      double u2 = p1.fX + ((p2.fX - p1.fX) * (((double) (u + 1)) / (fUSize - 1)));
      double v2 = p1.fY + ((p2.fY - p1.fY) * (((double) (v + 1)) / (fVSize - 1)));

      surfaceNormals[u][v] = Vector3D::calcNewellNormal3(Point3D(u1,getHeight(u,v),v1),
                                                         Point3D(u1,getHeight(u,v + 1),v2),
                                                         Point3D(u2,getHeight(u + 1,v),v1));

      surfaceNormals[u][2 * v] = Vector3D::calcNewellNormal3(Point3D(u2,getHeight(u + 1,v),v1),
                                                             Point3D(u1,getHeight(u,v + 1),v2),
                                                             Point3D(u2,getHeight(u + 1,v + 1),v2));
    }
  }

  // calculate the normal at every vertex in the center of the heightfield as
  // the average of the six surrounding face-normals
  for (v = 1; v < (fVSize - 1); ++v) {
    for (u = 1; u < (fUSize - 1); ++u) {

      // find the normals of the six surrounding faces
      Vector3D normal1(surfaceNormals[u - 1][2 * (v - 1)]);
      Vector3D normal2(surfaceNormals[u - 1][v]);
      Vector3D normal3(surfaceNormals[u - 1][2 * v]);
      Vector3D normal4(surfaceNormals[u][v]);
      Vector3D normal5(surfaceNormals[u][2 * (v - 1)]);
      Vector3D normal6(surfaceNormals[u][v - 1]);

      Vector3D averagedNormal((normal1 + normal2 + normal3 +
                               normal4 + normal5 + normal6) / 6.0);

      setNormal(u,v,averagedNormal);
    }
  }

  // calculate the normals at the four corner-vertices
  for (i = 0; i < 4; ++i) {
    switch (i) {

      case 0 :
        u = 0;
        v = 0;
        break;

      case 1 :
        u = 0;
        v = fVSize - 2;
        break;

      case 2 :
        u = fUSize - 2;
        v = fVSize - 2;
        break;

      case 3 :
        u = fUSize - 2;
        v = 0;
        break;
    }

    double u1 = p1.fX + ((p2.fX - p1.fX) * (((double) u) / (fUSize - 1)));
    double v1 = p1.fY + ((p2.fY - p1.fY) * (((double) v) / (fVSize - 1)));
    double u2 = p1.fX + ((p2.fX - p1.fX) * (((double) (u + 1)) / (fUSize - 1)));
    double v2 = p1.fY + ((p2.fY - p1.fY) * (((double) (v + 1)) / (fVSize - 1)));

    setNormal(u,v,Vector3D::calcNewellNormal4(Point3D(u1,getHeight(u,v),v1),
                                              Point3D(u1,getHeight(u,v + 1),v2),
                                              Point3D(u2,getHeight(u + 1,v + 1),v2),
                                              Point3D(u2,getHeight(u + 1,v),v1)));
  }

  // calculate the normals of the vertices located completely above and below
  for (u = 1; u < (fUSize - 2); ++u) {

    // handle the vertex above
    v = 0;
    Vector3D normal1(surfaceNormals[u - 1][v]);
    Vector3D normal2(surfaceNormals[u - 1][2 * v]);
    Vector3D normal3(surfaceNormals[u][v]);

    Vector3D averagedNormal((normal1 + normal2 + normal3) / 3.0);

    setNormal(u,v,averagedNormal);

    // handle the vertex below
    v = fVSize - 1;

    normal1 = surfaceNormals[u - 1][2 * (v - 1)];
    normal2 = surfaceNormals[u][v - 1];
    normal3 = surfaceNormals[u][2 * (v - 1)];

    averagedNormal = (normal1 + normal2 + normal3) / 3.0;

    setNormal(u,v,averagedNormal);
  }

  // calculate the normals of the vertices located completely left and right
  for (v = 1; v < (fVSize - 2); ++v) {

    // handle the vertex left
    u = 0;
    Vector3D normal1(surfaceNormals[u][v - 1]);
    Vector3D normal2(surfaceNormals[u][2 * (v - 1)]);
    Vector3D normal3(surfaceNormals[u][v]);

    Vector3D averagedNormal((normal1 + normal2 + normal3) / 3.0);

    setNormal(u,v,averagedNormal);

    // handle the vertex right
    u = fUSize - 1;

    normal1 = surfaceNormals[u - 1][2 * (v - 1)];
    normal2 = surfaceNormals[u - 1][v];
    normal3 = surfaceNormals[u - 1][2 * v];

    averagedNormal = (normal1 + normal2 + normal3) / 3.0;

    setNormal(u,v,averagedNormal);
  }
}

Color TerrainMap::linearInterpolatedColor(const Color& lowestColor,
                                          const Color& highestColor,
                                          double lowestLevel,
                                          double highestLevel,
                                          double level) const
{
  Color color;

  // clip level if needed
  if (level < lowestLevel) {
    level = lowestLevel;
  }
  else if (level > highestLevel) {
    level = highestLevel;
  }

  color.fRed = lowestColor.fRed +
               (((highestColor.fRed - lowestColor.fRed) / (highestLevel - lowestLevel)) *
                (level - lowestLevel));
  color.fGreen = lowestColor.fGreen +
                 (((highestColor.fGreen - lowestColor.fGreen) / (highestLevel - lowestLevel)) *
                  (level - lowestLevel));
  color.fBlue = lowestColor.fBlue +
                (((highestColor.fBlue - lowestColor.fBlue) / (highestLevel - lowestLevel)) *
                 (level - lowestLevel));

  return color;
}
