// ------------------------------
// Filename      : terrainmap.h
// Author        : Sven Maerivoet
// Last modified : 11/04/2001
// Target        : C++ (OpenGL)
//
// All rights reserved.
// ------------------------------

#ifndef SM_TERRAINMAP_H // include-guard
#define SM_TERRAINMAP_H

#include "primitives.h"
#include "rgbapixmap.h"
#include <string>
#include <vector>

using namespace std;

class HeightInfo {
  public:
    double   fHeight;
    Vector3D fNormal;
    double   fRed;
    double   fGreen;
    double   fBlue;
    bool     fIsAtSeaLevel;
};

class TerrainMap {
  public:
    // some useful constants
    static const bool kFlattenNormals;
    static const bool kSmoothenNormals;

    // constructor
    TerrainMap(void);

    // destructor
    ~TerrainMap(void);

    // methods
    void createFractalTerrain(int size,
                              double upperLeftHeight,
                              double lowerLeftHeight,
                              double upperRightHeight,
                              double lowerRightHeight,
                              double initialPerturbation,
                              double roughness,
                              const Rectangle2D& bounds);
    void constructFromBMPFile(string filename,
                              double maxHeight,
                              const Rectangle2D& bounds);

    void setBounds(const Rectangle2D& bounds);
    Rectangle2D getBounds(void) const;

    void setHeight(int u, int v, double h);
    double getHeight(int u, int v) const;

    void setSeaLevel(double seaLevel);
    double getSeaLevel(void) const;
    void setSeaFrequency(int seaFrequency);
    int getSeaFrequency(void) const;
    void setWavePeriod(int wavePeriod);
    int getWavePeriod(void) const;
    void setWaveHeight(double waveHeight);
    double getWaveHeight(void) const;
    void disableSea();

    void setSnowLevel(double snowLevel);
    double getSnowLevel(void) const;

    void setColors(const Color& lowestSeaColor,
                   const Color& highestSeaColor,
                   const Color& lowestMountainColor,
                   const Color& highestMountainColor,
                   const Color& lowestSnowColor,
                   const Color& highestSnowColor);
    Color getColor(int u, int v) const;
    void useSpecifiedColors(void);
    void useRealColors(void);

    void setTextures(RGBApixmap* seaTexture,
                     RGBApixmap* mountainTexture,
                     RGBApixmap* snowTexture);
    RGBApixmap* getTexture(int u, int v) const;

    void setCurrentTime(int currentTime);

    int getUSize(void) const;
    int getVSize(void) const;

    Vector3D getVertexNormal(int u, int v) const;

    void writeTo3VNMeshFile(string filename) const;

  private:
    // utility methods
    void freeTerrainMap(void);
    void createTerrainMap(void);
    void generateRecursiveFractalTerrain(int subdivisionsLeft,
                                         double perturbation,
                                         int u1, int v1,
                                         int u2, int v2);
    void calcMinMaxHeights(void);
    void setNormal(int u, int v, const Vector3D& normal);
    void averageVertexNormals(void);
    Color linearInterpolatedColor(const Color& lowestColor,
                                  const Color& highestColor,
                                  double lowestLevel,
                                  double highestLevel,
                                  double level) const;

    // data-members
    typedef vector< vector<HeightInfo> > Heightfield;
    Heightfield* fHeightfield;
    int          fUSize;
    int          fVSize;
    Rectangle2D  fBounds;
    double       fRoughness;
    double       fSeaLevel;
    double       fSnowLevel;
    double       fMinimumHeight;
    double       fMaximumHeight;
    Color        fLowestSeaColor;
    Color        fHighestSeaColor;
    Color        fLowestMountainColor;
    Color        fHighestMountainColor;
    Color        fLowestSnowColor;
    Color        fHighestSnowColor;
    RGBApixmap*  fSeaTexture;
    RGBApixmap*  fMountainTexture;
    RGBApixmap*  fSnowTexture;
    int          fCurrentTime;
    int          fWavePeriod;
    double       fWaveHeight;
    int          fSeaFrequency;
    bool         fUseRealColors;
};

#endif
