// ------------------------------
// Filename      : raytracer.cpp
// Author        : Sven Maerivoet
// Last modified : 30/05/2001
// Target        : C++ (OpenGL)
//
// All rights reserved.
// ------------------------------

#ifdef _WIN32
  #include <windows.h>
#endif

#include "raytracer.h"
#include "opengl/opengl.h"
#include "opengl/primitives.h"
#include <GL/gl.h>
#include <GL/glu.h>
#include <GL/glut.h>
#include <fstream>
#include <string.h>

using namespace std;

// --------------------
// class RayTracerGLApp
// --------------------

// -----------
// constructor
// -----------
RayTracerGLApp::RayTracerGLApp(string filename,
                               int blockSize,
                               int nrOfConeSamples,
                               int shadowComputationMode,
                               double shadowFeelerJitterStrength,
                               int antiAliasMode,
                               double jitterStrength,
                               bool normalizeColors,
                               double normalizeBrightness,
                               string outputFilePrefix) : fFilename(filename),
                                                          fBlockSize(blockSize),
                                                          fNrOfConeSamples(nrOfConeSamples),
                                                          fShadowFeelerJitterStrength(shadowFeelerJitterStrength),
                                                          fJitterStrength(jitterStrength),
                                                          fNormalizeColors(normalizeColors),
                                                          fNormalizeBrightness(normalizeBrightness),
                                                          fOutputFilePrefix(outputFilePrefix)
{
  switch (shadowComputationMode) {

    default :
    case 0 :
      fComputeShadows = false;
      break;

    case 1 :
      fComputeShadows = true;
      fShadowComputationMode = TOpenGLApp::kHardShadows;
      break;

    case 2 :
      fComputeShadows = true;
      fShadowComputationMode = TOpenGLApp::kSoftShadows;
      break;
  }

  switch (antiAliasMode) {

    default :
    case 0 :
      fAntiAlias = false;
      break;

    case 1 :
      fAntiAlias = true;
      fAntiAliasMode = TOpenGLApp::kUse5Samples;
      fUseStochasticSampling = false;
      break;

    case 2 :
      fAntiAlias = true;
      fAntiAliasMode = TOpenGLApp::kUse9Samples;
      fUseStochasticSampling = false;
      break;

    case 3 :
      fAntiAlias = true;
      fAntiAliasMode = TOpenGLApp::kUse5Samples;
      fUseStochasticSampling = true;
      break;

    case 4 :
      fAntiAlias = true;
      fAntiAliasMode = TOpenGLApp::kUse9Samples;
      fUseStochasticSampling = true;
      break;
  }
}

// ----------
// destructor
// ----------
RayTracerGLApp::~RayTracerGLApp(void)
{
}

// -----------------------
// explicit initialisation
// -----------------------
void RayTracerGLApp::explicitInitialize()
{
  setCamera(Point3D(0.0,2.0,3.0),
            Point3D(0.0,0.0,0.0),
            Vector3D(0.0,1.0,0.0));
  setCameraPerspectiveShape(45.0,getWorldWindowAspectRatio(),
                            0.1,1000.0);

  loadScene(fFilename);
  setSceneViewSettings();
}

// ------------------
// callback-functions
// ------------------
void RayTracerGLApp::displayFunc(void)
{
  enableSceneLighting();
  switchLightingOn();
  drawScene();

  takeSnapshot(fOutputFilePrefix + "-opengl.bmp");

  EShadowComputationMode shadowComputationMode(fShadowComputationMode);
  if (!fComputeShadows) {
    shadowComputationMode = kNoShadows;
  }

  EAntiAliasMode antiAliasMode(fAntiAliasMode);
  if (!fAntiAlias) {
    antiAliasMode = kUse1Sample;
  }

  fChrono.start();

  rayTraceScene(fBlockSize,
                fNrOfConeSamples,
                shadowComputationMode,
                fShadowFeelerJitterStrength,
                antiAliasMode,
                fUseStochasticSampling,
                fJitterStrength,
                fNormalizeColors,
                fNormalizeBrightness,
                false,
                0.0,
                false,
                0.0);

  fChrono.stop();

  takeSnapshot(fOutputFilePrefix + "-raytraced.bmp");

  // save the information
  ofstream infoFile((fOutputFilePrefix + "-info.txt").c_str(),ios::out);
  if (infoFile) {
    infoFile << endl;
    infoFile << "Ray tracing information :" << endl;
    infoFile << "-------------------------" << endl;
    infoFile << endl;

    infoFile << "    input-file                    = " << fFilename << endl;
    infoFile << "    window-size                   = " << getClientWindowXSize() << "x"
                                                       << getClientWindowYSize() << endl;
    infoFile << "    #pixels                       = " << getClientWindowXSize() * getClientWindowYSize() << endl;
    infoFile << "    #rays shot                    = " << getNrOfRaysShot() << endl;
    infoFile << "    block-size                    = " << fBlockSize << endl;
    infoFile << "    compute shadows               = ";
    if (fComputeShadows) {
      infoFile << "yes";
      if (fShadowComputationMode == kHardShadows) {
        infoFile << " (hard shadows)" << endl;
      }
      else {
        infoFile << " (soft shadows)" << endl;
        infoFile << "    shadow-feeler jitter-strength = " << fShadowFeelerJitterStrength << endl;
      }
    }
    else {
      infoFile << "no" << endl;
    }
    infoFile << "    cone-samples                  = " << fNrOfConeSamples << endl;
    infoFile << "    anti-aliasing                 = ";
    if (!fAntiAlias) {
      infoFile << "none" << endl;
    }
    else {
      if ((fAntiAliasMode == TOpenGLApp::kUse5Samples) &&
          (!fUseStochasticSampling)) {
        infoFile << "5 rays supersampled" << endl;
      }
      else if ((fAntiAliasMode == TOpenGLApp::kUse9Samples) &&
          (!fUseStochasticSampling)) {
        infoFile << "9 rays supersampled" << endl;
      }
      else if ((fAntiAliasMode == TOpenGLApp::kUse5Samples) &&
          (fUseStochasticSampling)) {
        infoFile << "5 rays stochastically sampled" << endl;
      }
      else if ((fAntiAliasMode == TOpenGLApp::kUse9Samples) &&
          (fUseStochasticSampling)) {
        infoFile << "9 rays stochastically sampled" << endl;
      }
      if (fUseStochasticSampling) {
        infoFile << "    jitter-strength               = " << fJitterStrength << endl;
      }
    }
    infoFile << "    normalize colors              = ";
    if (fNormalizeColors) {
      infoFile << "yes" << endl;
      infoFile << "    normalize brightness          = " << fNormalizeBrightness << endl;
    }
    else {
      infoFile << "no" << endl;
    }
    long elapsedTime = fChrono.getElapsedTime();
    infoFile << "    render-time                   = " << Chrono::convertToHMS(elapsedTime)
             << " (= " << elapsedTime << " seconds)" << endl;
    infoFile << endl;
  }
  infoFile.close();

  exit(0);
}
