// ---------------------------------
// Filename      : raytracermain.cpp
// Author        : Sven Maerivoet
// Last modified : 30/05/2001
// Target        : C++ (OpenGL)
//
// All rights reserved.
// ---------------------------------

#include "raytracer.h"
#include "opengl/opengl.h"
#include <iostream>
#include <string>

// string-streams are not always compatible
// => Microsoft Visual C++ 6.0 and UNIX/Linux are treated separately
#ifdef _WIN32
  #include <sstream>
#else
  #include <strstream>
#endif

using namespace std;

// --------------------------------
// global OpenGL Application object
// --------------------------------
TOpenGLApp* fOpenGLApp;

// -------------------------------
// wrappers for callback-functions
// -------------------------------
void displayFuncWrapper(void)                              { fOpenGLApp->antiAliasedDisplayFunc(); }
void reshapeFuncWrapper(int width, int height)             { fOpenGLApp->reshapeFunc(width,height); }
void keyboardFuncWrapper(unsigned char key, int x, int y)  { fOpenGLApp->keyboardFunc(key,x,y); }
void specialFuncWrapper(int key, int x, int y)             { fOpenGLApp->specialFunc(key,x,y); }
void mouseFuncWrapper(int button, int state, int x, int y) { fOpenGLApp->mouseFunc(button,state,x,y); }
void motionFuncWrapper(int x, int y)                       { fOpenGLApp->motionFunc(x,y); }
void passiveMotionFuncWrapper(int x, int y)                { fOpenGLApp->passiveMotionFunc(x,y); }
void idleFuncWrapper(void)                                 { fOpenGLApp->idleFunc(); }

void cleanUp(void)
{
  if (fOpenGLApp != 0) {
    delete fOpenGLApp;
  }
}

double getValue(string str)
{
  stringstream tmp;
  double number;

  tmp << str;
  if (!(tmp >> number)) {
    cerr << "getValue() -> error getting value [" << str << "]." << endl;
    exit(1);
  }
  else {
    char t;
    if (tmp >> t) {
      cerr << "getValue() -> bum chars in number [" << str << "]." << endl;
      exit(1);
    } 
  }

  return number;
}

void main(int argc, char* argv[])
{
  // register the cleanup-routine
  atexit(&cleanUp);

  if (argc != (12 + 1)) {
    cout << endl;
    cout << "Usage : raytracer <sdl-file> <window-width> <window-height> <block-size>" << endl;
    cout << "                  <nr-of-cone-samples> <shadow-computation-mode> <shadow-feeler-jitter-strength>" << endl;
    cout << "                  <anti-alias-mode> <jitter-strength>" << endl;
    cout << "                  <normalize-colors> <normalize-brightness> <outputfile-prefix>" << endl;
    cout << endl;
    cout << "Shadow computation can be 0 (no shadows are computed), 1 (hard shadows are computed)" << endl;
    cout << "or 2 (soft shadows are computed)." << endl;
    cout << "The shadow-feeler-jitter-strength specifies how much the shadow-feelers are stochastically" << endl;
    cout << "repositioned (i.e. 0.05)." << endl;
    cout << "Anti-alias-mode can be 0 (no anti-aliasing), 1 (5 rays supersampled), 2 (9 rays" << endl;
    cout << "supersampled), 3 (5 rays stochastically sampled) and 4 (9 rays stochastically sampled)." << endl;
    cout << "The jitter-strength specifies how much the rays are stochastically repositioned (i.e. 0.1)." << endl;
    cout << "Use a 0 to suppress post-normalization of the colors in the render-buffer, any other value" << endl;
    cout << "enables it." << endl;
    cout << "Three files are generated, one with OpenGL's rendering (suffix -opengl.bmp), another one is" << endl;
    cout << "rendered using ray tracing (suffix -raytraced.bmp) and the third one contains rendering" << endl;
    cout << "information (suffix -info.txt)." << endl;
    cout << endl;
    exit(1);
  }

  int windowXSize = (int) getValue(argv[2]);
  int windowYSize = (int) getValue(argv[3]);

  // create and initialize the OpenGL Application object
  fOpenGLApp = new RayTracerGLApp(argv[1],
                                  (int) getValue(argv[4]),
                                  (int) getValue(argv[5]),
                                  (int) getValue(argv[6]),
                                  getValue(argv[7]),
                                  (int) getValue(argv[8]),
                                  getValue(argv[9]),
                                  (((int) getValue(argv[10])) != 0),
                                  getValue(argv[11]),
                                  argv[12]);
  fOpenGLApp->initialize(0,0,
                         windowXSize,windowYSize,
                         "Ray tracing an SDL-scene",
                         TOpenGLApp::kEnableDoubleBuffering,
                         TOpenGLApp::kEnableDepthBuffer);

  // register the necessary callback-functions
  fOpenGLApp->registerDisplayFunc(displayFuncWrapper);
  fOpenGLApp->registerReshapeFunc(reshapeFuncWrapper);

  // activate the event handling
  fOpenGLApp->activateEventHandling();
}
