// ------------------------------
// Filename      : swarm.cpp
// Author        : Sven Maerivoet
// Last modified : 30/05/2001
// Target        : C++ (OpenGL)
//
// All rights reserved.
// ------------------------------

#ifdef _WIN32
  #include <windows.h>
#endif

#include "swarm.h"
#include "opengl/math.h"
#include "opengl/opengl.h"
#include "opengl/primitives.h"
#include "opengl/rng.h"
#include <GL/gl.h>
#include <GL/glu.h>
#include <GL/glut.h>
#include <math.h>

using namespace std;

// -----------
// class Agent
// -----------

// -----------
// constructor
// -----------
Agent::Agent(TOpenGLApp* openGLApp, double size) : fPosition(0.0,0.0,0.0),
                                                   fVelocity(0.0,0.0,0.0),
                                                   fOpenGLApp(openGLApp),
                                                   fSize(size),
                                                   fTexture(0)
{
}

// ----------
// destructor
// ----------
Agent::~Agent(void)
{
}

// -------
// methods
// -------
void Agent::draw(EAgentType agentType)
{
  // make it shiny blue
  fOpenGLApp->setColor(Color(0.0,0.0,0.8));
  fOpenGLApp->setMaterial(Color(0.0,0.0,0.0),
                          Color(0.0,0.0,1.0),
                          Color(1.0,1.0,1.0),
                          100.0);

  // orient the agent to its direction-of-flight
  double angle = 0.0;
  fOpenGLApp->pushCT();
    fOpenGLApp->translate3D(fPosition);

    if ((agentType == kCone) ||
        (agentType == kTeapot)) {

      // maybe we should use Euler-angles or perhaps even quaternions ?
      angle = Math::radToDeg(Math::arctan(fVelocity.fY,fVelocity.fX));
      fOpenGLApp->rotate3D(angle,0.0,0.0,1.0);

      angle = Math::radToDeg(Math::arctan(fVelocity.fX,fVelocity.fZ));
      fOpenGLApp->rotate3D(angle,0.0,1.0,0.0);

      if (agentType == kCone) {
        glutSolidCone((GLdouble) (fSize / 2.0),
                      (GLdouble) fSize,
                      (GLint) 20,
                      (GLint) 20);
      }
      else {
        glutSolidTeapot((GLdouble) fSize);
      }
    }
    else { // Borg-cube
      Point3D p1(-fSize,+fSize,-fSize);
      Point3D p2(-fSize,-fSize,-fSize);
      Point3D p3(+fSize,-fSize,-fSize);
      Point3D p4(+fSize,+fSize,-fSize);
      Point3D q1(-fSize,+fSize,+fSize);
      Point3D q2(-fSize,-fSize,+fSize);
      Point3D q3(+fSize,-fSize,+fSize);
      Point3D q4(+fSize,+fSize,+fSize);

      fOpenGLApp->setColor(Color(1.0,1.0,1.0));
      fOpenGLApp->setMaterial(Color(0.0,0.0,0.0),
                              Color(1.0,1.0,1.0),
                              Color(1.0,1.0,1.0),
                              100.0);

      fOpenGLApp->textureQuad(fTexture,
                              Point2D(0.0,1.0),Point2D(0.0,0.0),Point2D(1.0,0.0),Point2D(1.0,1.0),
                              q1,q2,q3,q4);
      fOpenGLApp->textureQuad(fTexture,
                              Point2D(0.0,0.0),Point2D(1.0,0.0),Point2D(1.0,1.0),Point2D(0.f,1.0),
                              q4,q3,p3,p4);
      fOpenGLApp->textureQuad(fTexture,
                              Point2D(0.0,1.0),Point2D(0.0,0.0),Point2D(1.0,0.0),Point2D(1.0,1.0),
                              p4,p3,p2,p1);
      fOpenGLApp->textureQuad(fTexture,
                              Point2D(1.0,1.0),Point2D(0.0,1.0),Point2D(0.0,0.0),Point2D(1.0,0.0),
                              p1,p2,q2,q1);
      fOpenGLApp->textureQuad(fTexture,
                              Point2D(0.0,0.0),Point2D(1.0,0.0),Point2D(1.0,1.0),Point2D(0.0,1.0),
                              p1,q1,q4,p4);
      fOpenGLApp->textureQuad(fTexture,
                              Point2D(0.0,0.0),Point2D(1.0,0.0),Point2D(1.0,1.0),Point2D(0.0,1.0),
                              q2,p2,p3,q3);
    }
  fOpenGLApp->popCT();
}

// -----------
// class Swarm
// -----------

// -----------
// constructor
// -----------
Swarm::Swarm(TOpenGLApp* openGLApp,
             const Point3D& centerPosition,
             double positionConstraint,
             double speedConstraint,
             double craziness,
             double agentSize) : fOpenGLApp(openGLApp),
                                 fCenterPosition(centerPosition),
                                 fPositionConstraint(positionConstraint),
                                 fSpeedConstraint(speedConstraint),
                                 fCraziness(craziness),
                                 fAgentSize(agentSize),
                                 fAgentType(Agent::kCone)
{
  fAgents = new vector<Agent>;
  fTexture = new RGBApixmap();
  if (!fTexture->loadFromBMPFile("data/textures/other/borgcube.bmp")) {
    exit(0);
  }
  fTexture->setTextureID(fOpenGLApp->getUniqueTextureID());
}

// ----------
// destructor
// ----------
Swarm::~Swarm(void)
{
  delete fAgents;
  fAgents = 0;
}

// -------
// methods
// -------
void Swarm::moveAgents(void)
{
  Agents::iterator p = 0;

  // calculate the agents' new velocities
  for (p = fAgents->begin(); p != fAgents->end(); ++p) {

    // find this agent's nearest neighbour
    Agent neighbour = findNearestAgent(*p);

    // change this agent's velocity-vector (use half the difference between the speeds)
    p->fVelocity.fX += -(p->fVelocity.fX - neighbour.fVelocity.fX) / 2.0;
    p->fVelocity.fY += -(p->fVelocity.fY - neighbour.fVelocity.fY) / 2.0;
    p->fVelocity.fZ += -(p->fVelocity.fZ - neighbour.fVelocity.fZ) / 2.0;

    // introduce the craziness
    if (RNG::uniform() < fCraziness) {

      double maxStepSize = fSpeedConstraint / 5.0;

      // assign the agent a random velocity-vector
      // (that almost lies in the direction of the previous vector)
      p->fVelocity.fX += -maxStepSize + (maxStepSize * 2.0 * RNG::uniform());
      p->fVelocity.fY += -maxStepSize + (maxStepSize * 2.0 * RNG::uniform());
      p->fVelocity.fZ += -maxStepSize + (maxStepSize * 2.0 * RNG::uniform());

      // clip the velocity-vector
      if (p->fVelocity.fX < -fSpeedConstraint) {
        p->fVelocity.fX = -fSpeedConstraint;
      }
      else if (p->fVelocity.fX > fSpeedConstraint) {
        p->fVelocity.fX = fSpeedConstraint;
      }

      if (p->fVelocity.fY < -fSpeedConstraint) {
        p->fVelocity.fY = -fSpeedConstraint;
      }
      else if (p->fVelocity.fY > fSpeedConstraint) {
        p->fVelocity.fY = fSpeedConstraint;
      }

      if (p->fVelocity.fZ < -fSpeedConstraint) {
        p->fVelocity.fZ = -fSpeedConstraint;
      }
      else if (p->fVelocity.fZ > fSpeedConstraint) {
        p->fVelocity.fZ = fSpeedConstraint;
      }
    }
  }

  // move all the agents
  for (p = fAgents->begin(); p != fAgents->end(); ++p) {

    p->fPosition.fX += p->fVelocity.fX;
    p->fPosition.fY += p->fVelocity.fY;
    p->fPosition.fZ += p->fVelocity.fZ;

    // adjust the agent's position (it flies on the surface of something like a hypertorus)
    if (p->fPosition.fX < (fCenterPosition.fX - fPositionConstraint)) {
      p->fPosition.fX = fCenterPosition.fX + (2.0 * fPositionConstraint) - p->fPosition.fX;
    }
    else if (p->fPosition.fX > (fCenterPosition.fX + fPositionConstraint)) {
      p->fPosition.fX = fCenterPosition.fX + p->fPosition.fX - (2.0 * fPositionConstraint);
    }

    if (p->fPosition.fY < (fCenterPosition.fY - fPositionConstraint)) {
      p->fPosition.fY = fCenterPosition.fY + (2.0 * fPositionConstraint) - p->fPosition.fY;
    }
    else if (p->fPosition.fY > (fCenterPosition.fY + fPositionConstraint)) {
      p->fPosition.fY = fCenterPosition.fY + p->fPosition.fY - (2.0 * fPositionConstraint);
    }

    if (p->fPosition.fZ < (fCenterPosition.fZ - fPositionConstraint)) {
      p->fPosition.fZ = fCenterPosition.fZ + (2.0 * fPositionConstraint) - p->fPosition.fZ;
    }
    else if (p->fPosition.fZ > (fCenterPosition.fZ + fPositionConstraint)) {
      p->fPosition.fZ = fCenterPosition.fZ + p->fPosition.fZ - (2.0 * fPositionConstraint);
    }
  }
}

void Swarm::drawAgents(void)
{
  for (Agents::iterator p = fAgents->begin(); p != fAgents->end(); ++p) {
    p->draw(fAgentType);
  }  
}

void Swarm::setNrOfAgents(int nrOfAgents)
{
  if (nrOfAgents < 0) {
    nrOfAgents = 0;
  }

  fAgents->clear();

  for (int i = 0; i < nrOfAgents; ++i) {
    Agent* agent = new Agent(fOpenGLApp,fAgentSize);

    // generate a random position and velocity for this agent
    agent->fPosition.fX = -fPositionConstraint + (2.0 * fPositionConstraint * RNG::uniform());
    agent->fPosition.fY = -fPositionConstraint + (2.0 * fPositionConstraint * RNG::uniform());
    agent->fPosition.fZ = -fPositionConstraint + (2.0 * fPositionConstraint * RNG::uniform());
    agent->fPosition.fX += fCenterPosition.fX;
    agent->fPosition.fY += fCenterPosition.fY;
    agent->fPosition.fZ += fCenterPosition.fZ;
    agent->fVelocity.fX = -fSpeedConstraint + (2.0 * fSpeedConstraint * RNG::uniform());
    agent->fVelocity.fY = -fSpeedConstraint + (2.0 * fSpeedConstraint * RNG::uniform());
    agent->fVelocity.fZ = -fSpeedConstraint + (2.0 * fSpeedConstraint * RNG::uniform());

    agent->fTexture = fTexture;

    fAgents->push_back(*agent);
  }
}

int Swarm::getNrOfAgents(void) const
{
  return fAgents->size();
}

void Swarm::cycleAgentType(void)
{
  if (fAgentType == Agent::kCone) {
    fAgentType = Agent::kTeapot;
  }
  else if (fAgentType == Agent::kTeapot) {
    fAgentType = Agent::kBorgCube;
  }
  else if (fAgentType == Agent::kBorgCube) {
    fAgentType = Agent::kCone;
  }
}

// ---------------------------------
// find an agent's nearest neighbour
// ---------------------------------
Agent Swarm::findNearestAgent(const Agent& agent) const
{
  double shortestDistance = 0.0;
  Agent* nearestAgent = 0;

  for (Agents::iterator p = fAgents->begin(); p != fAgents->end(); ++p) {

    double distance = sqrt(Math::sqr(agent.fPosition.fX - p->fPosition.fX) +
                           Math::sqr(agent.fPosition.fY - p->fPosition.fY) +
                           Math::sqr(agent.fPosition.fZ - p->fPosition.fZ));

    if (shortestDistance == 0.0) {
      shortestDistance = distance;
      nearestAgent = p;
    }
    else if (distance != 0.0) {
      if (distance < shortestDistance) {
        shortestDistance = distance;
        nearestAgent = p;
      }
    }
  }

  if ((shortestDistance != 0.0) &&
      (shortestDistance < (fPositionConstraint / 5.0))) {
    return *nearestAgent;
  }
  else {
    return agent;
  }
}

// ----------------
// class SwarmGLApp
// ----------------

// --------------------------------------
// initialization of static const members
// --------------------------------------
const double SwarmGLApp::kPositionConstraint = 1000.0;

// -----------
// constructor
// -----------
SwarmGLApp::SwarmGLApp(void) : fLightMountedOnHead(false),
                               fAdvancing(false),
                               fToriDelta(5.0),
                               fXZTorusSize(100.0),
                               fYTorusSize(150.0),
                               fFogEnabled(false),
                               fSwarm(0)
{
}

// ----------
// destructor
// ----------
SwarmGLApp::~SwarmGLApp(void)
{
  delete fSwarm;
  fSwarm = 0;
}

// -----------------------
// explicit initialisation
// -----------------------
void SwarmGLApp::explicitInitialize(void)
{
  fSwarm = new Swarm(this,
                     Point3D(0.0,0.0,0.0),
                     kPositionConstraint,
                     50.0,
                     0.30,
                     100.0);
  fSwarm->setNrOfAgents(20);

  setCameraSlideStepSize(20.0);

  Point3D eyePoint(1.5 * kPositionConstraint,
                   1.5 * kPositionConstraint,
                   1.5 * kPositionConstraint);

  setCamera(eyePoint,Point3D(0.0,0.0,0.0),Vector3D(0.0,1.0,0.0));
  setCameraPerspectiveShape(45.0,getWorldWindowAspectRatio(),
                            1.0,10.0 * kPositionConstraint);
}

// ------------------
// callback-functions
// ------------------
void SwarmGLApp::displayFunc(void)
{
  switchLightOn(kLight1);
  setLightProperties(kLight1,
                     Color(1.0,1.0,1.0),
                     Color(1.0,1.0,1.0),
                     Color(1.0,1.0,1.0));
  if (fLightMountedOnHead) {
    // mount a light on our head
    setLightPosition(kLight1,getCameraEyePoint());
  }
  else {
    // mount a light at a standard position
    setLightPosition(kLight1,Point3D(kPositionConstraint,
                                     kPositionConstraint,
                                     kPositionConstraint));
  }

  if (!fFogEnabled) {
    // clear the screen with a deep sort of blue color
    setBackgroundColor(Color(0.1,0.0,0.3));
  }
  else {
    // clear the screen with the gray color of the fog
    Color fogColor(0.4,0.4,0.4);
    setBackgroundColor(fogColor);
    enableLinearFog(fogColor,100.0,1000.0);
  }
  clearScreen();

  setGlobalAmbientLight(Color(0.2,0.0,0.3));

  // show the surrounding cube
  setColor(kBlue);
  setMaterialToGold();
  glutWireCube((GLdouble) (2.0 * kPositionConstraint));

  // show two inter-locking tori at the origin
  setColor(kGreen);
  pushCT();
    setMaterialToSilver();
    glutSolidTorus((GLdouble) 50.0,
                   (GLdouble) fXZTorusSize,
                   (GLint) 20,
                   (GLint) 20);
    setMaterialToGold();
    rotate3D(90.0,1.0,0.0,0.0);
    glutSolidTorus((GLdouble) 50.0,
                   (GLdouble) fYTorusSize,
                   (GLint) 20,
                   (GLint) 20);
  popCT();

  // draw a grid
  setColor(kCyan);
  setMaterialToChrome();
  for (int x = (int) -kPositionConstraint; x < (int) kPositionConstraint; x += 100) {
    line(x,-kPositionConstraint,-kPositionConstraint,
         x,-kPositionConstraint,kPositionConstraint);
  }
  for (int z = (int) -kPositionConstraint; z < (int) kPositionConstraint; z += 100) {
    line(-kPositionConstraint,-kPositionConstraint,z,
         kPositionConstraint,-kPositionConstraint,z);
  }

  // show all the birds
  fSwarm->drawAgents();

  // show some information
  Point3D eyePoint = getCameraEyePoint();
  drawString(10,15,"eye = (" +
                   convertDoubleToString(eyePoint.fX) + "," +
                   convertDoubleToString(eyePoint.fY) + "," +
                   convertDoubleToString(eyePoint.fZ) + ")",
             kYellow);
  drawString(10,35,"#agents = " + convertDoubleToString(fSwarm->getNrOfAgents()),kYellow);
}

void SwarmGLApp::idleFunc(void)
{
  // handle default flying behaviour
  TOpenGLApp::idleFunc();

  // shrink/enlarge the tori at the origin
  fXZTorusSize += fToriDelta;
  fYTorusSize -= fToriDelta;
  if ((fXZTorusSize > 175.0) ||
      (fXZTorusSize < 75.0)) {
    fToriDelta = -fToriDelta;
  }

  // move birds
  fSwarm->moveAgents();

  if (fAdvancing) {
    if (!advanceCameraAlongPath()) {
      fAdvancing = false;
    }
  }

  redisplay();
}

void SwarmGLApp::keyboardFunc(unsigned char key, int x, int y)
{
  // handle some default keys
  TOpenGLApp::keyboardFunc(key,x,y);

  Point3D eyePoint;

  switch (key) {
    case '+' :
      fSwarm->setNrOfAgents(fSwarm->getNrOfAgents() + 5);
      break;

    case '-' :
      fSwarm->setNrOfAgents(fSwarm->getNrOfAgents() - 5);
      break;

    case 'g' :
      if (fFogEnabled) {
        fFogEnabled = false;
        disableFog();
      }
      break;

    case 'G' :
      if (!fFogEnabled) {
        fFogEnabled = true;
      }
      break;

    case 'o' :
    case 'O' :
      eyePoint.fX = 1.5 * kPositionConstraint;
      eyePoint.fY = 1.5 * kPositionConstraint;
      eyePoint.fZ = 1.5 * kPositionConstraint;
      setCamera(eyePoint,getCameraViewPoint(),getCameraUpDirection());
      break;

    case 'p' :
      fAdvancing = false;
      break;

    case 'P' :
      fAdvancing = true;
      loadCameraPath("data/path/swarm.path");
      setCameraSlideStepSize(50.0);
      break;

    case 'h' :
      fLightMountedOnHead = false;
      break;

    case 'H' :
      fLightMountedOnHead = true;
      break;

    case 'b' :
    case 'B' :
      fSwarm->cycleAgentType();
      break;
  }

  if (fScheduleRedisplay) {
    redisplay();
  }
}
