// --------------------------------
// Filename      : generate-dna.cpp
// Author        : Sven Maerivoet
// Last modified : 14/06/2001
// Target        : C++ (OpenGL)
//
// All rights reserved.
// --------------------------------

#include "opengl/rng.h"
#include <iostream>
#include <math.h>
#include <string>

// string-streams are not always compatible
// => Microsoft Visual C++ 6.0 and UNIX/Linux are treated separately
#ifdef _WIN32
  #include <sstream>
#else
  #include <strstream>
  #define  stringstream strstream
#endif

using namespace std;

int kLeftMargin;

double getValue(string str)
{
  stringstream tmp;
  double number;

  tmp << str;
  if (!(tmp >> number)) {
    cerr << "getValue() -> error getting value [" << str << "]." << endl;
    exit(1);
  }
  else {
    char t;
    if (tmp >> t) {
      cerr << "getValue() -> bum chars in number [" << str << "]." << endl;
      exit(1);
    } 
  }

  return number;
}

void makeLeftMargin(void)
{
  for (int i = 0; i < kLeftMargin; ++i) {
    cout << " ";
  }
}

void generateDNA(double helixRadius,
                 double helixLength,
                 double moleculeRadius,
                 int nrOfRevolutions,
                 int nrOfSequences)
{
  makeLeftMargin(); cout << "push" << endl;

  for (int sequence = 0; sequence < nrOfSequences; ++sequence) {

    double angle = sequence * ((360.0 * nrOfRevolutions) / nrOfSequences);
    double height = sequence * (helixLength / nrOfSequences);

    // choose an inner-double base-pair
    // 0 = A, 1 = T, 2 = C, 3 = G
    int leftBase = RNG::uniformInt() % 4;
    int rightBase;
    if (leftBase == 0) {
      rightBase = 1;
    }
    else if (leftBase == 1) {
      rightBase = 0;
    }
    else if (leftBase == 2) {
      rightBase = 3;
    }
    else if (leftBase == 3) {
      rightBase = 2;
    }

    // make inner hydrogen bonds
    if ((leftBase == 0) || (leftBase == 1)) {
      // make two weak inner bonds for A en T
      double heightOffset = (moleculeRadius / 4.0);
      makeLeftMargin(); cout << "  push" << endl;
      makeLeftMargin(); cout << "    defaultMaterials" << endl;
      makeLeftMargin(); cout << "    diffuse            1.0 1.0 1.0" << endl;
      makeLeftMargin(); cout << "    specular           0.7 0.7 0.7" << endl;
      makeLeftMargin(); cout << "    specularExponent 100.0" << endl;
      makeLeftMargin(); cout << "    rotate           " << angle << " 0.0 1.0 0.0" << endl;
      makeLeftMargin(); cout << "    translate        " << -(helixRadius / 4.0) << " " << (height + heightOffset) << " 0.0" << endl;
      makeLeftMargin(); cout << "    scale            " << (helixRadius / 2.0) << " " << (moleculeRadius / 4.0) << " " << (moleculeRadius / 4.0) << endl;
      makeLeftMargin(); cout << "    rotate            90.0 0.0 1.0 0.0" << endl;
      makeLeftMargin(); cout << "    cylinder         0" << endl;
      makeLeftMargin(); cout << "  pop" << endl;
      makeLeftMargin(); cout << "  push" << endl;
      makeLeftMargin(); cout << "    defaultMaterials" << endl;
      makeLeftMargin(); cout << "    diffuse            1.0 1.0 1.0" << endl;
      makeLeftMargin(); cout << "    specular           0.7 0.7 0.7" << endl;
      makeLeftMargin(); cout << "    specularExponent 100.0" << endl;
      makeLeftMargin(); cout << "    rotate           " << angle << " 0.0 1.0 0.0" << endl;
      makeLeftMargin(); cout << "    translate        " << -(helixRadius / 4.0) << " " << (height - heightOffset) << " 0.0" << endl;
      makeLeftMargin(); cout << "    scale            " << (helixRadius / 2.0) << " " << (moleculeRadius / 4.0) << " " << (moleculeRadius / 4.0) << endl;
      makeLeftMargin(); cout << "    rotate            90.0 0.0 1.0 0.0" << endl;
      makeLeftMargin(); cout << "    cylinder         0" << endl;
      makeLeftMargin(); cout << "  pop" << endl;
    }
    else {
      // make three weak inner bonds for C en G
      double heightOffset = (moleculeRadius / 4.0);
      makeLeftMargin(); cout << "  push" << endl;
      makeLeftMargin(); cout << "    defaultMaterials" << endl;
      makeLeftMargin(); cout << "    diffuse            1.0 1.0 1.0" << endl;
      makeLeftMargin(); cout << "    specular           0.7 0.7 0.7" << endl;
      makeLeftMargin(); cout << "    specularExponent 100.0" << endl;
      makeLeftMargin(); cout << "    rotate           " << angle << " 0.0 1.0 0.0" << endl;
      makeLeftMargin(); cout << "    translate        " << -(helixRadius / 4.0) << " " << (height + heightOffset) << " 0.0" << endl;
      makeLeftMargin(); cout << "    scale            " << (helixRadius / 2.0) << " " << (moleculeRadius / 4.0) << " " << (moleculeRadius / 4.0) << endl;
      makeLeftMargin(); cout << "    rotate            90.0 0.0 1.0 0.0" << endl;
      makeLeftMargin(); cout << "    cylinder         0" << endl;
      makeLeftMargin(); cout << "  pop" << endl;
      makeLeftMargin(); cout << "  push" << endl;
      makeLeftMargin(); cout << "    defaultMaterials" << endl;
      makeLeftMargin(); cout << "    diffuse            1.0 1.0 1.0" << endl;
      makeLeftMargin(); cout << "    specular           0.7 0.7 0.7" << endl;
      makeLeftMargin(); cout << "    specularExponent 100.0" << endl;
      makeLeftMargin(); cout << "    rotate           " << angle << " 0.0 1.0 0.0" << endl;
      makeLeftMargin(); cout << "    translate        " << -(helixRadius / 4.0) << " " << height << " 0.0" << endl;
      makeLeftMargin(); cout << "    scale            " << (helixRadius / 2.0) << " " << (moleculeRadius / 4.0) << " " << (moleculeRadius / 4.0) << endl;
      makeLeftMargin(); cout << "    rotate            90.0 0.0 1.0 0.0" << endl;
      makeLeftMargin(); cout << "    cylinder         0" << endl;
      makeLeftMargin(); cout << "  pop" << endl;
      makeLeftMargin(); cout << "  push" << endl;
      makeLeftMargin(); cout << "    defaultMaterials" << endl;
      makeLeftMargin(); cout << "    diffuse            1.0 1.0 1.0" << endl;
      makeLeftMargin(); cout << "    specular           0.7 0.7 0.7" << endl;
      makeLeftMargin(); cout << "    specularExponent 100.0" << endl;
      makeLeftMargin(); cout << "    rotate           " << angle << " 0.0 1.0 0.0" << endl;
      makeLeftMargin(); cout << "    translate        " << -(helixRadius / 4.0) << " " << (height - heightOffset) << " 0.0" << endl;
      makeLeftMargin(); cout << "    scale            " << (helixRadius / 2.0) << " " << (moleculeRadius / 4.0) << " " << (moleculeRadius / 4.0) << endl;
      makeLeftMargin(); cout << "    rotate            90.0 0.0 1.0 0.0" << endl;
      makeLeftMargin(); cout << "    cylinder         0" << endl;
      makeLeftMargin(); cout << "  pop" << endl;
    }

    // make a strong outer bond to the left
    makeLeftMargin(); cout << "  push" << endl;
    makeLeftMargin(); cout << "    defaultMaterials" << endl;
    makeLeftMargin(); cout << "    diffuse            1.0 1.0 1.0" << endl;
    makeLeftMargin(); cout << "    specular           0.7 0.7 0.7" << endl;
    makeLeftMargin(); cout << "    specularExponent 100.0" << endl;
    makeLeftMargin(); cout << "    rotate           " << angle << " 0.0 1.0 0.0" << endl;
    makeLeftMargin(); cout << "    translate        " << -helixRadius << " " << height << " 0.0" << endl;
    makeLeftMargin(); cout << "    scale            " << (3.0 * (helixRadius / 4.0)) << " " << (moleculeRadius / 2.0) << " " << (moleculeRadius / 2.0) << endl;
    makeLeftMargin(); cout << "    rotate            90.0 0.0 1.0 0.0" << endl;
    makeLeftMargin(); cout << "    cylinder         0" << endl;
    makeLeftMargin(); cout << "  pop" << endl;

    // make a strong outer bond to the right
    makeLeftMargin(); cout << "  push" << endl;
    makeLeftMargin(); cout << "    defaultMaterials" << endl;
    makeLeftMargin(); cout << "    diffuse            1.0 1.0 1.0" << endl;
    makeLeftMargin(); cout << "    specular           0.7 0.7 0.7" << endl;
    makeLeftMargin(); cout << "    specularExponent 100.0" << endl;
    makeLeftMargin(); cout << "    rotate           " << angle << " 0.0 1.0 0.0" << endl;
    makeLeftMargin(); cout << "    translate        " << (helixRadius / 4.0) << " " << height << " 0.0" << endl;
    makeLeftMargin(); cout << "    scale            " << (3.0 * (helixRadius / 4.0)) << " " << (moleculeRadius / 2.0) << " " << (moleculeRadius / 2.0) << endl;
    makeLeftMargin(); cout << "    rotate            90.0 0.0 1.0 0.0" << endl;
    makeLeftMargin(); cout << "    cylinder         0" << endl;
    makeLeftMargin(); cout << "  pop" << endl;

    // make an inner-left base
    makeLeftMargin(); cout << "  push" << endl;
    makeLeftMargin(); cout << "    defaultMaterials" << endl;
    switch (leftBase) {
      case 0:
        makeLeftMargin(); cout << "    diffuse            1.0 0.0 0.0" << endl;
        break;
      case 1:
        makeLeftMargin(); cout << "    diffuse            0.0 1.0 0.0" << endl;
        break;
      case 2:
        makeLeftMargin(); cout << "    diffuse            1.0 0.0 1.0" << endl;
        break;
      case 3:
        makeLeftMargin(); cout << "    diffuse            1.0 1.0 0.0" << endl;
        break;
    }
    makeLeftMargin(); cout << "    specular           0.7 0.7 0.7" << endl;
    makeLeftMargin(); cout << "    specularExponent 100.0" << endl;
    makeLeftMargin(); cout << "    rotate           " << angle << " 0.0 1.0 0.0" << endl;
    makeLeftMargin(); cout << "    translate        " << -(helixRadius / 4.0) << " " << height << " 0.0" << endl;
    makeLeftMargin(); cout << "    scale            " << (3.0 * (moleculeRadius / 4.0)) << " " << (3.0 * (moleculeRadius / 4.0)) << " " << (3.0 * (moleculeRadius / 4.0)) << endl;
    makeLeftMargin(); cout << "    sphere" << endl;
    makeLeftMargin(); cout << "  pop" << endl;

    // make an inner-right base
    makeLeftMargin(); cout << "  push" << endl;
    makeLeftMargin(); cout << "    defaultMaterials" << endl;
    switch (rightBase) {
      case 0:
        makeLeftMargin(); cout << "    diffuse            1.0 0.0 0.0" << endl;
        break;
      case 1:
        makeLeftMargin(); cout << "    diffuse            0.0 1.0 0.0" << endl;
        break;
      case 2:
        makeLeftMargin(); cout << "    diffuse            1.0 0.0 1.0" << endl;
        break;
      case 3:
        makeLeftMargin(); cout << "    diffuse            1.0 1.0 0.0" << endl;
        break;
    }
    makeLeftMargin(); cout << "    specular           0.7 0.7 0.7" << endl;
    makeLeftMargin(); cout << "    specularExponent 100.0" << endl;
    makeLeftMargin(); cout << "    rotate           " << angle << " 0.0 1.0 0.0" << endl;
    makeLeftMargin(); cout << "    translate        " << (helixRadius / 4.0) << " " << height << " 0.0" << endl;
    makeLeftMargin(); cout << "    scale            " << (3.0 * (moleculeRadius / 4.0)) << " " << (3.0 * (moleculeRadius / 4.0)) << " " << (3.0 * (moleculeRadius / 4.0)) << endl;
    makeLeftMargin(); cout << "    sphere" << endl;
    makeLeftMargin(); cout << "  pop" << endl;

    // make a phosphate to the left
    makeLeftMargin(); cout << "  push" << endl;
    makeLeftMargin(); cout << "    defaultMaterials" << endl;
    makeLeftMargin(); cout << "    diffuse            0.0 0.0 1.0" << endl;
    makeLeftMargin(); cout << "    specular           0.7 0.7 0.7" << endl;
    makeLeftMargin(); cout << "    specularExponent 100.0" << endl;
    makeLeftMargin(); cout << "    rotate           " << angle << " 0.0 1.0 0.0" << endl;
    makeLeftMargin(); cout << "    translate        " << -helixRadius << " " << height << " 0.0" << endl;
    makeLeftMargin(); cout << "    scale            " << moleculeRadius << " " << moleculeRadius << " " << moleculeRadius << endl;
    makeLeftMargin(); cout << "    sphere" << endl;
    makeLeftMargin(); cout << "  pop" << endl;

    // make a phosphate to the right
    makeLeftMargin(); cout << "  push" << endl;
    makeLeftMargin(); cout << "    defaultMaterials" << endl;
    makeLeftMargin(); cout << "    diffuse            0.0 0.0 1.0" << endl;
    makeLeftMargin(); cout << "    specular           0.7 0.7 0.7" << endl;
    makeLeftMargin(); cout << "    specularExponent 100.0" << endl;
    makeLeftMargin(); cout << "    rotate           " << angle << " 0.0 1.0 0.0" << endl;
    makeLeftMargin(); cout << "    translate        " << helixRadius << " " << height << " 0.0" << endl;
    makeLeftMargin(); cout << "    scale            " << moleculeRadius << " " << moleculeRadius << " " << moleculeRadius << endl;
    makeLeftMargin(); cout << "    sphere" << endl;
    makeLeftMargin(); cout << "  pop" << endl;
  }
  makeLeftMargin(); cout << "pop" << endl;
}

void main(int argc, char* argv[])
{
  if (argc != (6 + 1)) {
    cout << endl;
    cout << "Usage : generate-dna <left-margin> <helix-radius> <helix-length>" << endl;
    cout << "                     <molecule-radius> <#revolutions> <#sequences>" << endl;
    cout << endl;
    exit(0);
  }

  kLeftMargin = (int) getValue(argv[1]);

  RNG::setSeedWithTime();

  generateDNA(getValue(argv[2]),
              getValue(argv[3]),
              getValue(argv[4]),
              (int) getValue(argv[5]),
              (int) getValue(argv[6]));
}
